/****************************************************
 *  File UFSD/include/fsbase.h
 *  Copyright (C) 2000 Paragon Software,
 *
 *  Here are the definitions for Base classes
 *  of file, directory and file system
 *  in Universal File System Driver (UFSD).
 *  For every new file system supported in
 *  UFSD developer of this file system must derive
 *  his own File, Dir, FileSystem classes from
 *  classes described here

 Since 2001 this file is supported by Alexander Mamaev

 ****************************************************/

#ifndef __UFSD_FSBASE_H
#define __UFSD_FSBASE_H

namespace UFSD{

//Known file systems
#define FS_NTFS     0x00000002
#define FS_FAT16    0x00000004
#define FS_FAT32    0x00000008
#define FS_EXFAT    0x00000010
#define FS_HFSPLUS  0x00000020
#define FS_EXT2     0x00000040

#define FS_LSWAP    0x00000080
#define FS_REISERFS 0x00000100
#define FS_FAT12    0x00000200
#define FS_SYSV     0x00000400
#define FS_HTFS     0x00000800
#define FS_REFS     0x00001000
#define FS_XFS      0x00002000

//virtual FS
#define FS_REG      0x00100000
#define FS_ISO      0x00200000
#define FS_ZIP      0x00400000
#define FS_FB       0x00800000

//
// File systems mask (exclude reg,iso,zip,fb,...)
//
#define FS_MASK     (FS_NTFS|FS_FAT12|FS_FAT16|FS_FAT32|FS_EXFAT|FS_HFSPLUS|FS_EXT2|FS_LSWAP|FS_REFS|FS_XFS)


enum VolumeState {
    NoChange, // Used in SetVolumeInfo function to skip dirty flag changing
    Dirty,    // Used in (Get/Set)VolumeInfo
    Clean     // Used in (Get/Set)VolumeInfo
};


//Seek origins
enum SEEKOrig {
    UFSD_SEEKCUR=0,
    UFSD_SEEKEND,
    UFSD_SEEKSET
};

//
// File attributes
//
#define UFSD_NORMAL       0x00000000  //  Normal file - read/write permitted
#define UFSD_RDONLY       0x00000001  //  Read-only file
#define UFSD_HIDDEN       0x00000002  //  Hidden file
#define UFSD_SYSTEM       0x00000004  //  System file
#define UFSD_VOLID        0x00000008  //  Volume-ID entry
#define UFSD_SUBDIR       0x00000010  //  Subdirectory
#define UFSD_ARCH         0x00000020  //  Archive file
#define UFSD_LINK         0x00000400  //  File is link (Reparse point)
#define UFSD_COMPRESSED   0x00000800  //  NTFS Compressed file
#define UFSD_ENCRYPTED    0x00004000  //  NTFS Encrypted file
#define UFSD_ATTRIB_MASK  0x00004fff

//
// Virtual attributes
//
#define UFSD_TASIZE       0x00800000  //  TotalAllocSize field is valid
#define UFSD_NSUBDIRS     0x01000000  //  Subdirs count is valid
#define UFSD_OBJID        0x02000000  //  File contains ObjId attributes. For NTFS use IOCTL_GET_OBJECT_ID2
#define UFSD_VSIZE        0x04000000  //  FileInfo contains valid field 'ValidSize'
#define UFSD_UGM          0x08000000  //  FileInfo contains valid fields 'Mode','Uid','Gid'
#define UFSD_STREAMS      0x10000000  //  File contains additional data streams. For NTFS use CFSObject::GetObjectInfo
#define UFSD_EA           0x20000000  //  File contains Extended attributes. For NTFS use CFSObject::GetObjectInfo
#define UFSD_RESIDENT     0x40000000  //  NTFS resident default data stream. Use CFSObject::GetObjectInfo
#define UFSD_FSSPEC       0x80000000  //  FileInfo contains valid 'FSAttrib'


//
// FileInfo::Mode
//
#define U_IFMT   0170000
#define U_IFSOCK 0140000
#define U_IFLNK  0120000
#define U_IFREG  0100000
#define U_IFBLK  0060000
#define U_IFDIR  0040000
#define U_IFCHR  0020000
#define U_IFIFO  0010000
#define U_ISUID  0004000
#define U_ISGID  0002000
#define U_ISVTX  0001000

#define U_ISLNK(m)  (((m) & U_IFMT) == U_IFLNK)
#define U_ISREG(m)  (((m) & U_IFMT) == U_IFREG)
#define U_ISDIR(m)  (((m) & U_IFMT) == U_IFDIR)
#define U_ISCHR(m)  (((m) & U_IFMT) == U_IFCHR)
#define U_ISBLK(m)  (((m) & U_IFMT) == U_IFBLK)
#define U_ISFIFO(m) (((m) & U_IFMT) == U_IFIFO)
#define U_ISSOCK(m) (((m) & U_IFMT) == U_IFSOCK)

#define U_IRWXU 00700
#define U_IRUSR 00400
#define U_IWUSR 00200
#define U_IXUSR 00100

#define U_IRWXG 00070
#define U_IRGRP 00040
#define U_IWGRP 00020
#define U_IXGRP 00010

#define U_IRWXO 00007
#define U_IROTH 00004
#define U_IWOTH 00002
#define U_IXOTH 00001

#define U_IRWXUGO (U_IRWXU|U_IRWXG|U_IRWXO)
#define U_IALLUGO (U_ISUID|U_ISGID|U_ISVTX|U_IRWXUGO)
#define U_IRUGO   (U_IRUSR|U_IRGRP|U_IROTH)
#define U_IWUGO   (U_IWUSR|U_IWGRP|U_IWOTH)
#define U_IXUGO   (U_IXUSR|U_IXGRP|U_IXOTH)


//File name length definitions
#define MAX_FILENAME    256           // FAT can store 256 symbols per file name
                                      // NTFS can store 255 symbols per file name
#define MAX_DOSFILENAME 13

#ifndef NO_PRAGMA_PACK
#pragma pack( push, 1 )
#endif

//
// File information structure
//
struct FileInfo{

  UINT64          Id;               // Volume unique file/folder identificator

  // All times are 64 bit number of 100 nanoseconds  seconds since 1601 year
  UINT64          CrTime;           // utc time of file creation
  UINT64          ReffTime;         // utc time of last access to file (atime)
  UINT64          ModiffTime;       // utc time of last file modification (mtime)
  UINT64          ChangeTime;       // utc time of last attribute modification (ctime)
  UINT64          AllocSize;        // Allocated bytes for file
  UINT64          FileSize;         // Length of file in bytes
  UINT64          ValidSize;        // Valid size. Valid if FlagOn(Attrib,UFSD_VSIZE)
  UINT64          TotalAllocSize;   // The sum of the allocated clusters for a file (usefull for sparse/compressed files) Valid if FlagOn(Attrib,UFSD_TASIZE)

  unsigned int    Attrib;           // File attributes. See UFSD_SUBDIR, UFSD_SYSTEM and so on
  unsigned int    FSAttrib;         // File system specific attributes. Valid if FlagOn(Attrib,UFSD_FSSPEC)

  unsigned short  nSubDirs;         // Subdirs count. valid if FlagOn(Attrib,UFSD_NSUBDIRS)
  unsigned short  Mode;             // mode       valid if FlagOn(Attrib,UFSD_UGM)
  unsigned int    Uid;              // user ID    valid if FlagOn(Attrib,UFSD_UGM)
  unsigned int    Gid;              // group ID   valid if FlagOn(Attrib,UFSD_UGM)
  unsigned int    Dev;              // if U_IFBLK == mode || U_IFCHR == mode  valid if FlagOn(Attrib,UFSD_UGM)

  unsigned short  HardLinks;        // Count of hardlinks. NOTE: NTFS does not fill this member in FindFirst/FindNext. Use CFSObject::GetObjectInfo instead.
  unsigned short  Gen;              // Generation of file

  unsigned short  NameLen;          // Length of Name in characters

#ifdef UFSD_DRIVER_LINUX
  // Driver does not uses the bullshit below
  unsigned short  Name[MAX_FILENAME+1];     // Zero terminated file name (always unicode)
#else
  unsigned short  AltNameLen;               // Length of AltName (if exists) in characters

  unsigned short  Name[MAX_FILENAME+1];     // Zero terminated long file name
  unsigned short  AltName[MAX_DOSFILENAME]; // Alternative DOS name

  STRType         NameType;         // Name type (UNICODE or ASCII)
  STRType         AltNameType;      // Alternative name type (UNICODE or ASCII)
  char            Res[2];           // To align struct
#endif

}  __attribute__ ((packed));

typedef FileInfo  t_FileInfo;
//C_ASSERT( 0 == (sizeof(FileInfo)%8) );

#ifndef NO_PRAGMA_PACK
#pragma pack( pop )
#endif


//Flags for file info modification
#define SET_ATTRIBUTES        0x00000001    // Set file attributes
#define SET_FSATTRIBUTES      0x00000002
#define SET_MODTIME           0x00000004    // Set modification time
#define SET_CRTIME            0x00000008    // Set creation time
#define SET_REFFTIME          0x00000010    // Set reference time
#define SET_CHTIME            0x00000020    // Set last time any attribute was modified
#define SET_ADD_ATTRIBUTES    0x00000080    // Add specified attributes
#define SET_REM_ATTRIBUTES    0x00000100    // Remove specified attributes
#define SET_UGM               0x00000200    // Set uid/gid/mode
#define SET_VSIZE_SIZE        0x00000400    // Set sizes

// Possible flags for SetEa
#define EXATTR_CREATE    0x1     // set value, fail if attr already exists
#define EXATTR_REPLACE   0x2     // set value, fail if attr does not exist
#define EXATTR_NEED_EA   0x4     // file to which the EA belongs cannot be interpreted without understanding the associated extended attributes.

// Possible flags for CFile::Allocate
#define UFSD_ALLOCATE_KEEP_SIZE   0x01

// Possible flags for CFile::GetMap
// Allocate clusters if requested range is not allocated yet
#define UFSD_MAP_VBO_CREATE             0x0001
// Allow CFile::GetMap( Vbo, Bytes, ... ) to allocate only 'continues' fragment
// Must be combined with UFSD_MAP_VBO_CREATE
#define UFSD_MAP_VBO_CREATE_CONTINUES   0x0002
// Allow CFile::GetMap( Vbo, Bytes, ... ) to allocate less than 'Bytes'
// Must be combined with UFSD_MAP_VBO_CREATE
#define UFSD_MAP_VBO_CREATE_PARTITIAL   0x0004
// Do not call any lock operations
#define UFSD_MAP_VBO_NOLOCK             0x0008

// Possible flags for MapInfo::Flags
#define UFSD_MAP_LBO_NEW                0x0001

// Special values for MapInfo::Lbo
// Fragment of file is not allocated (sparsed)
#define UFSD_VBO_LBO_HOLE         ((UINT64)-1)
// File is resident
#define UFSD_VBO_LBO_RESIDENT     ((UINT64)-2)
// Fragment is really compressed (need decompress)
#define UFSD_VBO_LBO_COMPRESSED   ((UINT64)-3)
// File is encrypted
#define UFSD_VBO_LBO_ENCRYPTED    ((UINT64)-4)


//
// Layout of file fragment
//
struct MapInfo{
  UINT64  Lbo;        // Logical byte offset
  UINT64  Len;        // Length of map in bytes
  UINT64  Alloc;      // The allocated size for file
  UINT64  TotalAlloc; // The sum of the allocated clusters for a file (used for sparsed and compressed files)
  size_t  Flags;      // Properties of fragment
};


//////////////////////////////////////////////
//Class enumerator for file search procedures
//////////////////////////////////////////////
class UFSD_ABSTRACT_CLASS CEntryNumerator : public CMemBased<CEntryNumerator>
{
protected:
  UINT64  m_Position;
  CEntryNumerator(CMemoryManager * mm) : CMemBased<CEntryNumerator>(mm), m_Position(0){}
  ~CEntryNumerator(){}; // non virtual!

public:
  virtual void Destroy() = 0;
#ifdef UFSD_DRIVER_LINUX
  UINT64 GetPosition(){ return m_Position; }
  //Set next entry position
  virtual void SetPosition( const UINT64& ) = 0;
#else
  virtual UINT64 GetPosition(){ return m_Position; }
  //Set next entry position
  virtual void SetPosition( const UINT64& Position){ m_Position = Position; }
  //Reset entry position to Zero
  virtual void Reset(){ m_Position = 0; }
#endif
};

class CDir;
///////////////////////////////////////////
//Base class for File and Directory classes
///////////////////////////////////////////
class UFSD_ABSTRACT_CLASS CFSObject : public CMemBased<CFSObject>
{
protected:
    CFSObject( CMemoryManager * mm )
      : CMemBased<CFSObject>( mm )
    {
      m_Entry.init();
    }

    ~CFSObject() {
#ifndef UFSD_DRIVER_LINUX
      free( m_aName );
      free( m_aAltName );
#endif
      m_Entry.remove();
    }

public:
    list_head         m_Entry;

    enum FSObjectType{FSObjUnknown = 0, FSObjDir = 1, FSObjFile = 2};

    CDir*             m_Parent;

#if !defined UFSD_DRIVER_LINUX || defined APPLE
    size_t            m_Cookie;                   // Extra information to be used by consumer.
#endif

#if defined UFSD_FUSE
    struct {
      hlist_node        hash;
      UINT64            isize, asize;
      unsigned          uid;
      unsigned          gid;
      unsigned          mode;
      unsigned          atime;    // ReffTime
      unsigned          ctime;    // ChangeTime
      unsigned          mtime;    // ModiffTime
      unsigned          dev;
      unsigned short    nlink;
      unsigned char     set_mode;
      unsigned char     set_time;
    } inode;
#endif

    int               m_Reff_count;               // Reference count

#ifdef UFSD_DRIVER_LINUX
    char              m_tObjectType;              // Object type - directory or file
    char              Res;                        // Reserved
#else
    unsigned short*   m_aName;                    // Name for file system object
    unsigned short*   m_aAltName;                 // Alternative DOS name

    unsigned short    m_NameLen;                  // The length of name in symbols
    unsigned short    m_AltNameLen;               // The length of alternative name in symbols

    char              m_tObjectType;              // Object type - directory or file
    char              m_tNameType;                // Name type - ASCII or UNICODE
    char              m_tAltNameType;             // Alternative name type (UNICODE or ASCII)
    char              Res;                        // Reserved
#endif

    //Creates object of specified type with name and name_type
    //with parent dir parent.
    //Returns:
    //  0 if ok
    //  error code otherwise
    unsigned Init(
        IN CDir*        parent,
        IN STRType      name_type,
        IN const void*  name,
        IN FSObjectType obj_type,
        IN size_t       name_len,
        IN STRType      AltName_type  = StrUnknown,
        IN const void*  AltName       = NULL,
        IN size_t       AltNameLen    = 0
        );

    // Returns the length of filled buffer
    // If FullName is NULL then the returned value is the length
    // (not including last zero!) required for full name
    // chSize         - Maximum number of symbols in buffer (including last zero)
    size_t GetFullName(
        IN  CStrings* Str,
        OUT char*     FullName,
        IN  size_t    chSize
        ) const;

    //Sets 'Name' and 'NameType'
    //Returns:
    //  0 if ok
    //  error code otherwise
    unsigned SetName(
        IN STRType      type,
        IN const void*  name,
        IN size_t       name_len
        );

    //Sets 'AltName' and 'AltNameType'
    //Returns error code
    unsigned SetAltName(
        IN STRType      type,
        IN const void*  name,
        IN size_t       name_len
        );

    // Returns pointer to 'name'
    // and sets type value into NameType value
    const void* GetName(
        OUT STRType&  type,
        IN  bool      bAltName = false,
        OUT size_t*   NameLen  = NULL
        ) const;

    //Compares name with the name of the FSObject
    //  type -  string type (UNICODE / ASCII)
    //  name -  pointer to the name to compare
    //Returns:
    //  0   - if equal
    //  < 0 - if less
    //  > 0 - if greater
    int CompareName(
        IN CStrings*    str,
        IN STRType      type,
        IN const void * name,
        IN size_t       name_len
        );

#ifndef UFSD_DRIVER_LINUX
    //Returns pointer to reference Count
    int GetReffCount()const{ return m_Reff_count; }

    //Increment reference count
    void IncReffCount( int Refs=1 ) { m_Reff_count += Refs; }

    //Decrement reference count
    void DecReffCount( int Refs=1 ) { m_Reff_count -= Refs; }

    //Returns object type for current object (File or Directory)
    FSObjectType GetObjectType()const{ return static_cast<FSObjectType>(m_tObjectType); }
#endif

    //=================================================================
    //          CFSObject virtual functions
    //=================================================================

    virtual void Destroy() = 0;

    // Gets file/directory information
    // fi - file information to get
    // Returns:
    //   UFSD error code
    virtual unsigned GetObjectInfo(
        OUT FileInfo& fi
        ) UFSD_DRIVER_PURE_VIRTUAL;

    // Sets file/directory information
    // fi - file information to set
    // Flags     - bit mask of fields to set. See SET_XXX
    // Returns:
    //   UFSD error code
    virtual unsigned SetObjectInfo(
        IN const FileInfo&  fi,
        IN size_t           Flags
        ) = 0;

    // Returns unique ID
    virtual UINT64 GetObjectID() const = 0;

    // Returns parent unique ID
    // -1 means error
    virtual UINT64 GetParentID(
        IN unsigned int Pos = 0
        ) const UFSD_DRIVER_PURE_VIRTUAL;

    // Returns name
    // NULL means error
    virtual const void* GetNameByPos(
        IN  unsigned int Pos,
        OUT STRType&     Type,
        OUT size_t&      NameLen
        ) const UFSD_DRIVER_PURE_VIRTUAL;

    // Flush file/directory
    virtual unsigned Flush(
        IN bool bDelete = false
        ) = 0;

    //=================================================================
    // Functions to work with named streams
    //=================================================================

    // Copy a list of available stream names into the buffer
    // provided, or compute the buffer size required
    // 'Buffer' - is a set of zero terminated strings of type 'NameType'
    // If ERR_MORE_DATA - *Bytes will contain the necessary size in bytes
    virtual unsigned ListStreams(
        OUT void*     Buffer,
        IN  size_t    BytesPerBuffer,
        OUT size_t*   Bytes,
        OUT STRType*  NameType
        );

    // Reads named stream
    virtual unsigned ReadStream(
        IN  STRType       NameType,
        IN  const void*   Name,
        IN  size_t        NameLen,
        IN  const UINT64& Offset,
        OUT void*         Buffer,
        IN  size_t        BytesPerBuffer,
        OUT size_t*       Read
        );

    // Creates/Writes named stream
    virtual unsigned WriteStream(
        IN  STRType       NameType,
        IN  const void*   Name,
        IN  size_t        NameLen,
        IN  const UINT64& Offset,
        IN  const void*   Buffer,
        IN  size_t        BytesPerBuffer,
        OUT size_t*       Written
        );

    // Delete named stream
    virtual unsigned RemoveStream(
        IN  STRType     NameType,
        IN  const void* Name,
        IN  size_t      NameLen
        );

    //=================================================================
    // Functions to work with extended attributes (looks like named streams)
    //=================================================================

    // Copy a list of extended attribute names into the buffer
    // provided, or compute the buffer size required
    virtual unsigned ListEa(
        OUT void*     Buffer,
        IN  size_t    BytesPerBuffer,
        OUT size_t*   Bytes
        ) UFSD_DRIVER_PURE_VIRTUAL;

    // Reads extended attribute
    virtual unsigned GetEa(
        IN  const char* Name,
        IN  size_t      NameLen,
        OUT void*       Buffer,
        IN  size_t      BytesPerBuffer,
        OUT size_t*     Len
        ) UFSD_DRIVER_PURE_VIRTUAL;

    // Creates/Writes extended attribute
    virtual unsigned SetEa(
        IN  const char* Name,
        IN  size_t      NameLen,
        IN  const void* Buffer,
        IN  size_t      BytesPerBuffer,
        IN  size_t      Flags // EXATTR_XXX
        ) UFSD_DRIVER_PURE_VIRTUAL;

    // Reads symbolic link
    virtual unsigned ReadSymLink(
        OUT char*       Buffer,
        IN  size_t      BufLen,
        OUT size_t*     Read
        ) UFSD_DRIVER_PURE_VIRTUAL;

    //=================================================================
    //          END OF CFSObject virtual functions
    //=================================================================

};


/////////////////////////////////
// Base class for file objects
/////////////////////////////////
class UFSD_ABSTRACT_CLASS CFile : public CFSObject
{
protected:
    CFile( CMemoryManager * mm ) : CFSObject(mm){}
    ~CFile(){}
public:

    //=================================================================
    //          CFile  virtual functions
    //=================================================================

    //Retrieves file size
    //Returns:
    //  0 - if ok
    //  error code otherwise
    virtual unsigned GetSize(
        OUT UINT64& BytesPerFile,
        OUT UINT64* ValidSize = NULL,
        OUT UINT64* AllocSize = NULL
        ) = 0;

    //Truncates or Expands file
    //Returns:
    //  0 if everything is fine
    //  error code - otherwise
    // if ValidSize is not NULL then update 'valid' size too (if FS supports it)
    virtual unsigned SetSize(
        IN const UINT64&  BytesPerFile,
        IN const UINT64*  ValidSize = NULL,
        OUT UINT64*       AllocSize = NULL
        ) = 0;

    //Read len bytes from file from specified position
    //  buffer  - output buffer
    //  len - number of bytes to read
    //  read - number of bytes read from file
    //Returns:
    // 0 if all ok
    // error code otherwise
    virtual unsigned Read(
        IN const UINT64&  Offset,
        OUT size_t        &read,
        OUT void *        buffer,
        IN size_t         len
        ) = 0;

    //Write len bytes to file from buffer
    //to file from specified position
    //  buffer  - output buffer
    //  len - number of bytes to read
    //  written - number of bytes written
    //Returns:
    // 0 if all ok
    // error code otherwise
    virtual unsigned Write(
        IN  const UINT64&   Offset,
        OUT size_t&         written,
        IN  const void*     buffer,
        IN  size_t          len
        ) = 0;

    //
    // Translates Virtual Byte Offset into Logical Byte Offset
    //
    // if ( FlagOn( Flags, UFSD_MAP_VBO_CREATE ) ) {
    //   host requires [Vbo, Vbo+Bytes) for writing
    // } else {
    //   host requires [Vbo, Vbo+Bytes) for reading
    // }
    //
    // If success then:
    //
    // if ( 0 == Map->Len ) {
    //   fragment has no map
    // } else {
    //   fragment has map
    //   assert( Map->Vbo <= Vbo && Vbo < Map->Vbo + Map->Len );
    //   if ( -1 != Map->Lbo ) {
    //     fragment has mapping [Vbo, Vbo+Len) => [Map->Lbo, Map->Lbo + Map->Len)
    //     if ( FlagOn( Map->Flags, UFSD_MAP_LBO_NEW ) ) {
    //       fragment is not initialized. (not zeroed)
    //     }
    //   } else {
    //     fragment is sparsed
    //   }
    // }
    //
    virtual unsigned GetMap(
        IN  const UINT64& Vbo,
        IN  const UINT64& Bytes,
        IN  size_t        Flags,
        OUT MapInfo*      Map
        ) UFSD_DRIVER_PURE_VIRTUAL;

#ifndef UFSD_DRIVER_LINUX
    // Allocates the disk space within the range [Offset,Offset+Bytes)
    virtual unsigned Allocate(
        IN const UINT64&  Offset,
        IN const UINT64&  Bytes,
        IN size_t         , // Flags = 0,   // see UFSD_ALLOCATE_XXX flags
        OUT MapInfo*      Map
        )
    {
      return GetMap( Offset, Bytes, UFSD_MAP_VBO_CREATE, Map );
    }
#endif

#if defined UFSD_NTFS_DELETE_SUBFILE | defined UFSD_HFS_DELETE_SUBFILE
    // Deletes the range of file
    virtual unsigned Delete(
        IN const UINT64&  Offset,
        IN const UINT64&  Bytes
        ) UFSD_DRIVER_PURE_VIRTUAL;
#endif

    //=================================================================
    //          END OF CFile  virtual functions
    //=================================================================
};


/////////////////////////////////////////
//Base class for directory objects
/////////////////////////////////////////
class UFSD_ABSTRACT_CLASS CDir : public CFSObject
{
protected:
    CDir( CMemoryManager * mm )
      : CFSObject( mm )
    {
      m_OpenFiles.init();
      m_OpenDirs.init();
    }
    ~CDir() { ClearAll(); }

    void ClearAll(); // gcc bug (two identical dtors)

public:
    list_head   m_OpenFiles;      // List of open files
    list_head   m_OpenDirs;       // List of open dirs


    // Helper function for Rename
    void  Relink(
        IN CFSObject* Fso
        );

    //
    // This function closes file or directory
    //
    unsigned Close(
        IN CFSObject* Fso
        );

    // This function looks opened object by name
    CFSObject* LookupOpenObject(
        IN CStrings*    str,
        IN STRType      type,
        IN const void*  name,
        IN size_t       name_len,
        IN int          IsDir // < 0 if any (file or dir)
        );

    // This function enumerates opened directories
    size_t EnumOpenedDirs(
        IN bool (*fn)(
            IN CDir*  dir,
            IN void*  Param
            ),
        IN void*  Param,
        OUT bool* bBreak
        ) const __attribute__ ((noinline));

    // This function enumerates opened files
    size_t  EnumOpenedFiles(
        IN bool (*fn)(
            IN CFile* file,
            IN void*  Param
            ),
        IN void*  Param,
        OUT bool* bBreak
        ) const __attribute__ ((noinline));

    //=================================================================
    //          CDir  virtual functions
    //=================================================================

    //
    // This function opens file or directory
    //
    virtual unsigned Open(
        IN  STRType     type,
        IN  const void* name,
        IN  size_t      name_len,
        OUT CFSObject** Fso,
        OUT FileInfo*   fi
        ) UFSD_DRIVER_PURE_VIRTUAL;

    //
    // Create file/dir/node
    //
    virtual unsigned Create(
        IN STRType        type,
        IN const void*    name,
        IN size_t         name_len,
        IN unsigned short mode,
        IN unsigned int   uid,
        IN unsigned int   gid,
        IN const void*    data,
        IN size_t         data_len,
        OUT CFSObject**   fso
        ) UFSD_DRIVER_PURE_VIRTUAL;

#ifndef UFSD_DRIVER_LINUX
    //Creates sub directory with the specified name
    //  type - defines name type (UNICODE or ASCII)
    //Returns:
    //  0 - if successfully
    //  error code otherwise
    virtual unsigned CreateDir(
        IN STRType      type,
        IN const void * name,
        IN size_t       name_len,
        OUT CDir**      dir = NULL // OPTIONAL
        );

    //Opens directory with specified name and places it to the list of
    //open directories.
    //  type - defines type of name (UNICODE or ASCII)
    //  dir  - pointer to the opened directory
    //Returns:
    //  0 - on success
    //  error code otherwise
    virtual unsigned OpenDir(
        IN  STRType     type,
        IN  const void* name,
        IN  size_t      name_len,
        OUT CDir *&     dir
        );

    //Closes opened directory 'dir'
    //Returns:
    //  0 - if successfully
    //  error code otherwise
    virtual unsigned CloseDir(
        IN CDir * dir
        );

    //Creates file in this dir
    //  type - defines name type (UNICODE or ASCII)
    //Returns:
    //  0 - if successfully
    //  error code otherwise
    virtual unsigned CreateFile(
        IN STRType      type,
        IN const void * name,
        IN size_t       name_len,
        OUT CFile**     file = NULL // OPTIONAL
        );

    //Opens file in this directory
    //  type - defines type of 'name' (UNICODE or ASCII)
    //  file - pointer to the opened file
    //Returns:
    //  0 - if successfully
    //  error code otherwise
    virtual unsigned OpenFile(
        IN  STRType     type,
        IN  const void* name,
        IN  size_t      name_len,
        OUT CFile *&    file
        );

    //Closes opened file 'file'
    //Return:
    //  0 - if successful
    //  error code otherwise
    virtual unsigned CloseFile(
        IN CFile * file
        );

    //Set file information
    //  type - file name type (ASCII or UNCODE)
    //  name - file name
    //  fi   - file information
    //Returns:
    //  0 - if successfully
    //  error code otherwise
    virtual unsigned SetFileInfo(
        IN STRType            type,
        IN const void *       name,
        IN size_t             name_len,
        IN const FileInfo&    fi,
        IN size_t             flags
        );

    // This function gets unique ID of FS object by its' path
    virtual unsigned GetFSObjID(
      IN  STRType     type,
      IN  const void* name,
      IN  size_t      name_len,
      OUT UINT64&     ID,
      IN  int         fFollowSymlink = 1
      );

    //This function creates symbolic link 'name' to symlink
    virtual unsigned CreateSymLink(
        IN STRType      type,
        IN const void*  name,
        IN size_t       name_len,
        IN const char*  symlink,
        IN size_t       sym_len,
        OUT UINT64*     pId
        );

    //This function reads symbolic link name
    virtual unsigned ReadSymLink(
        IN  STRType     type,
        IN  const void* name,
        IN  size_t      name_len,
        OUT char*       buffer,
        IN  size_t      buflen,
        OUT size_t*     read
        );

    //This function checks is link name referenced to existing file object
    virtual unsigned CheckSymLink(
        IN  STRType     type,
        IN  const void* name,
        IN  size_t      name_len
        );
#endif // #ifndef UFSD_DRIVER_LINUX

    //Deletes file or empty directory with name 'name' in dir;
    //Returns:
    //  0 - if successfully
    //  error code otherwise
    virtual unsigned Unlink(
        IN STRType      type,
        IN const void * name,
        IN size_t       name_len,
        IN CFSObject*   Fso = NULL
        ) = 0;

    // Creates a hard link to the file/directory
    // NOTE: hard link to directory usually is not allowed
    virtual unsigned Link(
        IN CFSObject*   Fso,
        IN STRType      type,
        IN const void * name,
        IN size_t       name_len
        ) UFSD_DRIVER_PURE_VIRTUAL;

    //Finds file or directory corresponding to name_mask specified
    //  type - defines ASCII or UNICODE
    //  name_maks - name mask informational, may be ignored
    //Returns:
    //  0 - if successfully
    //  error code otherwise
    virtual unsigned StartFind(
        IN  STRType           type,
        IN  const void*       mask,
        IN  size_t            len,
        OUT CEntryNumerator*& dentry_num,
        IN  size_t            options = 0 // See UFSD_ENUM_XXX
        ) = 0;

    //Finds file or directory corresponding to name_mask specified
    //  type - defines ASCII or UNICODE
    //  name_maks - name mask informational, may be ignored
    //  fi - file information found
    //Returns:
    //  0 - if successfully
    //  error code otherwise
    virtual unsigned FindNext(
        IN  CEntryNumerator*  dentry_num,
        OUT FileInfo&         fi
        ) = 0;

    //Retrieve file information
    //Returns:
    //  0 - if successfully
    //  error code otherwise
    virtual unsigned GetFileInfo(
        IN  STRType       type,
        IN  const void*   name,
        IN  size_t        name_len,
        OUT FileInfo&     fi
        ) = 0;

    //Rename file or directory
    //  src_name -  source name
    //  dst_name -  target name
    //  src      -  ?????
    //  tgt_dir  -  target directory
    //Returns:
    //  0 - if successfully
    //  error code otherwise
    virtual unsigned Rename(
        IN STRType      type,
        IN const void*  src_name,
        IN size_t       src_len,
        IN CFSObject*   src,
        IN CDir*        tgt_dir,
        IN const void*  dst_name,
        IN size_t       dst_len
        ) = 0;

    // Counts subdirectories and/or files
    virtual unsigned GetSubDirAndFilesCount(
        OUT bool*   IsEmpty,        // may be NULL
        OUT size_t* SubDirsCount,   // may be NULL
        OUT size_t* FilesCount      // may be NULL
        ) UFSD_DRIVER_PURE_VIRTUAL;

    // Retrieves directory size
    virtual UINT64 GetSize() UFSD_DRIVER_PURE_VIRTUAL;

    //=================================================================
    //          END OF CDir  virtual functions
    //=================================================================

};


class CTime;

//Base class for all file systems
class UFSD_ABSTRACT_CLASS CFileSystem : public CMemBased<CFileSystem>
{
protected:
    char*         m_TraceBuffer;

public:
    CDir*         m_RootDir;
    CDir*         m_CurrentDir;
    CStrings*     m_Strings;
    CTime*        m_Time;
    CRWBlock*     m_Rw;
    size_t        m_Options;        // See UFSD_OPTIONS_XXX
#ifdef UFSD_DRIVER_LINUX
    void (UFSDAPI_CALL *m_OnSetDirty)( super_block* Sb );
    super_block*  m_Sb;
#endif
    unsigned char m_Log2OfClump;    // Log2( clump size in bytes )
    bool          m_bDirty;         // Simple flag for internal use

protected:
    ////////////////
    // NOTE: Every instance of CFileSystem derived class
    // must be created as new( mm ) CFileSystemXXX( mm, ss, tt );
    // operator CMemBased::operator new() zeroes allocated memory
    // so we can skip zeroing any members
    ////////////////
    CFileSystem( CMemoryManager * mm, CStrings * ss, CTime * tt )
      : CMemBased<CFileSystem>(mm)
      , m_Strings(ss)
      , m_Time(tt)
    {
    }

    ~CFileSystem();  // non virtual!
public:

    const char* GetFullNameA( IN const CFSObject* obj );
    const char* U2A( IN const unsigned short* szUnicode, IN size_t Len, OUT char szBuffer[256] = NULL );

    // Helper function for exfat/ntfs
    unsigned CreateIntxLink(
        IN  const char* Link,
        IN  size_t      LinkLen,
        OUT void**      IntxLnk,
        OUT size_t*     IntxLen
        );

    //=================================================================
    //          CFileSystem virtual functions
    //=================================================================

    virtual void Destroy() = 0;

    //Initializes File System
    //  RwBlock - Read/Write object
    //  Option  - FS specific options, see UFSD_OPTIONS_XXX
    //  Flags   - Events while init, see UFSD_FLAGS_XXX
    //Returns:
    //  0 - if success;desc non zero;
    //  error code otherwise.
    virtual unsigned Init(
        IN CRWBlock*  RwBlock,
        IN size_t     Options = 0,
        OUT size_t*   Flags   = NULL
        ) = 0;

    // This function is used to reinit FileSystem by new options
    virtual unsigned ReInit(
        IN size_t     Options = 0,
        OUT size_t*   Flags   = NULL
        ) UFSD_DRIVER_PURE_VIRTUAL;

    //Opens Root directory and assigns m_RootDir a value.
    //Called when File System is mounted
    //Returns:
    //  not NULL if dir opened successfully
    virtual unsigned OpenRoot()=0;

    //Closes Root directory, closes all opened files.
    //Called when File System is unmounted
    //Returns:
    //  0 - if successfully
    //  error code otherwise
    virtual unsigned CloseRoot();

    //Flushes directory tree
    virtual unsigned Flush( bool bWait = false ) UFSD_DRIVER_PURE_VIRTUAL;

    //Flushes all data to disk
    virtual unsigned Sync() UFSD_DRIVER_PURE_VIRTUAL;

    //Gets allocation information
    virtual unsigned GetVolumeInfo(
         OUT UINT64*        FreeClusters,
         OUT UINT64*        TotalClusters,
         OUT size_t*        BytesPerCluster,
         OUT unsigned char  VolumeSerial[16]        = NULL,       // OPTIONAL
         IN  STRType        VolumeNameType          = StrUnknown, // OPTIONAL
         OUT void*          VolumeNameBuffer        = NULL,       // OPTIONAL
         IN  size_t         VolumeNameBufferLength  = 0,          // OPTIONAL, in (wide)chars
         OUT VolumeState*   VolState                = NULL,       // OPTIONAL
         OUT size_t*        BytesPerSector          = NULL        // OPTIONAL
        ) = 0;

    // Set volume information.
    virtual unsigned SetVolumeInfo(
         IN const unsigned char  VolumeSerial[16],  // OPTIONAL
         IN STRType              VolumeNameType,    // OPTIONAL
         IN const void*          VolumeNameBuffer,  // OPTIONAL
         IN VolumeState          NewState           // OPTIONAL (=NoChange)
        ) = 0;

    // Function a-la DeviceIoControl
    virtual unsigned IoControl(
        IN  size_t          FsIoControlCode,
        IN  const void*     InBuffer       = NULL, // OPTIONAL
        IN  size_t          InBufferSize   = 0,    // OPTIONAL
        OUT void*           OutBuffer      = NULL, // OPTIONAL
        IN  size_t          OutBufferSize  = 0,    // OPTIONAL
        OUT size_t*         BytesReturned  = NULL  // OPTIONAL
        ) = 0;

#ifdef UFSD_DRIVER_LINUX
    // Returns 0 if names are not equal
    //        +1 if names are equal
    //        -1 if some error occurs
    virtual int IsNamesEqual(
        IN const unsigned short*  Name1,
        IN size_t                 Len1,
        IN const unsigned short*  Name2,
        IN size_t                 Len2
        ) = 0;

    // Returns hash of name
    virtual unsigned int NameHash(
        IN const unsigned short*  Name,
        IN size_t                 Len
        ) = 0;
#endif

    // Opens file or dir by its id
    virtual unsigned OpenByID(
        IN  UINT64      ID,
        OUT CFSObject** Fso,
        IN  int         fFollowSymlink = 1
        ) UFSD_DRIVER_PURE_VIRTUAL;

    // Returns name of file system
    virtual const char* GetFsName() const = 0;

    // Returns any parent for given file/dir
    virtual unsigned GetParent(
        IN  CFSObject*  Fso,
        OUT CFSObject** Parent
        ) UFSD_DRIVER_PURE_VIRTUAL;

#ifdef UFSD_DRIVER_LINUX
    // Set callback function which is called for every change of free block count
    virtual void SetFreeSpaceCallBack(
        IN  void (*FreeSpaceCallBack)(
            IN size_t Lcn,
            IN size_t Len,
            IN int    AsUsed,
            IN void*  Arg
            ),
        IN void* Arg
        ) = 0;
#endif

    //=================================================================
    //          END OF CFileSystem virtual functions
    //=================================================================

};


// This function returns string like "UFSD version 3.3 (Mar 11 2003 , 20:05:38)"
UFSD_API const char*
__stdcall
GetUFSDVersion();


// This function returns string like "Warning: little endian library, big endian platform"
// or NULL if OK
UFSD_API const char*
__stdcall
CheckUFSDEndian();


///////////////////////////////////////////////////////////
// Log2
//
// Simple function that returns log2 of MSB (most significant bit)
///////////////////////////////////////////////////////////
static inline unsigned char
Log2( IN size_t Num )
{
  unsigned Ret = 0;
  while( 0 != ( Num >>= 1 ) )
    Ret += 1;

  return (unsigned char)Ret;
}

//===================================
// Functions to work with names
//===================================
// This function translates long name to short one
unsigned
GetShortName(
    IN CStrings*    Strings,
    IN STRType      Type,
    IN const void*  LongName,
    IN size_t       LongNameLen,
    OUT void*       ShortName,
    IN size_t       MaxShortNameLen,
    IN bool         bDoUpCase,
    OUT size_t*     ShortNameLen
    );

// This function creates first "~n" name
unsigned
GetFirstUniqueName(
    IN STRType    Type,
    IN OUT void*  ShortName,
    OUT size_t*   ShortNameLen,
    IN CMemoryManager* Mm
    );

// This function creates next "~n" name
unsigned
GetNextUniqueName(
    IN STRType    Type,
    IN OUT void*  ShortName,
    OUT size_t*   ShortNameLen,
    IN CMemoryManager* Mm
    );

// This function returns number from "~n" name
unsigned
GetUniqueNameNumber(
    IN STRType    Type,
    IN void*      ShortName,
    OUT size_t*   Number
    );

// This function creates pseudo short name
unsigned
GetPseudoShortName(
    IN CStrings*      Strings,
    IN STRType        Type,
    IN const void*    LongName,
    IN size_t         LongNameLen,
    OUT void*         ShortName,
    IN size_t         MaxShortNameLen,
    IN size_t         UniqueNum,
    IN bool           bDoUpCase,
    OUT size_t*       ShortNameLen,
    IN CMemoryManager* Mm
    );

//===================================
// End of functions to work with names
//===================================


// Returns Gb,Mb to print with "%u.%02u Gb"
UFSD_API unsigned int
FormatSizeGb(
    IN  const UINT64& Bytes,
    OUT unsigned int* Mb
    );


#ifdef UFSD_TRACE
// This function Dumps memory into trace
extern "C" UFSD_API void
UFSD_DumpMemory(
    IN const void*  Mem,
    IN size_t       nBytes
    );
#endif


// This struct is used in CUFSD::Format functions
// The last bad block has size 0
struct t_BadBlock{
  UINT64  Sector;
  UINT64  nSectors;
};

// This struct is used to get available format parameters
struct t_FormatParams{
  unsigned int  MinBytesPerCluster;
  unsigned int  MaxBytesPerCluster;
  unsigned int  DefBytesPerCluster;
};


// Forward declaration
struct IMessage;
///////////////////////////////////////////////////////////
//  VerifySectors
//
//  Start         - Start sector to verify
//  TotalSectors  - Total sectors to verify (relative Start)
//  BadBlocks     - The array of bad sectors
//  Msg           - Outlet for messages and progress bar
//
// Verifies sectors for bad blocks
//
//  NOTE: If the function returns ERR_NOERROR
//        then caller should free the returned array BadBlocks
///////////////////////////////////////////////////////////
extern "C" UFSD_API unsigned
VerifySectors(
    IN const UINT64&    StartSector,
    IN const UINT64&    TotalSectors,
    IN CRWBlock*        RWBlock,
    IN CMemoryManager*  MemMngr,
    IN IMessage*        Msg,
    IN const char*      Hint,
    OUT t_BadBlock**    BadBlocks
    );


///////////////////////////////////////////////////////////
// DiscardRange
//
// Discards (not necessary zeroing) range of bytes (at least 512 bytes aligned)
///////////////////////////////////////////////////////////
extern "C" UFSD_API unsigned
DiscardRange(
    IN UINT64       Offset,
    IN UINT64       Bytes,
    IN CRWBlock*    Rw,
    IN IMessage*    Msg,
    IN const char*  Hint
    );


///////////////////////////////////////////////////////////
// ComputeVolId
//
// This function calculates unique volume id
///////////////////////////////////////////////////////////
unsigned int
ComputeVolId(
    IN unsigned int Seed,
    IN CTime*       Time
    );


// A-la _snwprintf( Buf, MaxChars, L"%u", N )
int
ItoW(
    IN  unsigned int    N,
    OUT unsigned short* Buf,
    IN  int             MaxChars
    );

// A-la swscanf( Buf, L"%u", &N )
unsigned int
WtoI(
    IN  const unsigned short* Buf,
    IN  int                   Len,
    OUT bool*                 Error
    );


// A-la _snprintf( Buf, MaxChars, "%u", N )
int
ItoA(
    IN  unsigned int    N,
    OUT unsigned char*  Buf,
    IN  int             MaxChars
    );

// A-la sscanf( Buf, "%u", &N )
unsigned int
AtoI(
    IN  const unsigned char*  Buf,
    IN  int                   Len,
    OUT bool*                 Error
    );

// A-la _sntprintf( Buf, "%llx", N )
char*
U64toAh(
    IN  UINT64  N,
    OUT char*   Buf = NULL
    );


#ifdef UFSD_BIGENDIAN
  #define INTX_SYMBOLIC_LINK      PU64(0x496E74784C4E4B01)  // "IntxLNK\1"
  #define INTX_CHARACTER_DEVICE   PU64(0x496E747843485200)  // "IntxCHR\0"
  #define INTX_BLOCK_DEVICE       PU64(0x496E7478424C4B00)  // "IntxBLK\0"
#else
  #define INTX_SYMBOLIC_LINK      PU64(0x014B4E4C78746E49)  // "IntxLNK\1"
  #define INTX_CHARACTER_DEVICE   PU64(0x0052484378746E49)  // "IntxCHR\0"
  #define INTX_BLOCK_DEVICE       PU64(0x004B4C4278746E49)  // "IntxBLK\0"
#endif

///////////////////////////////////////////////////////////
// minor_dev
//
// Retrieves minor from encoded device
///////////////////////////////////////////////////////////
static inline unsigned int
minor_dev(
    IN unsigned int dev
    )
{
  return (dev & 0xff) | ((dev >> 12) & 0xfff00);
}


///////////////////////////////////////////////////////////
// major_dev
//
// Retrieves major from encoded device
///////////////////////////////////////////////////////////
static inline unsigned int
major_dev(
    IN unsigned int dev
    )
{
  return (dev & 0xfff00) >> 8;
}


///////////////////////////////////////////////////////////
// encode_dev
//
// Retrieves encoded dev from major/minor
///////////////////////////////////////////////////////////
static inline unsigned int
encode_dev(
    IN unsigned int minor,
    IN unsigned int major
    )
{
  return (minor & 0xff) | (major << 8) | ((minor & ~0xff) << 12);
}

//
// Environment hint
//
extern const char szUFSDVersionEnv[];

//
// Possible options for CDir::StartFind
//

#define UFSD_ENUM_SKIP_SHORT_NAMES    0x00000001    // Do not find short names while enumerating directory
#define UFSD_ENUM_SHOW_META           0x00000002    // Show meta files
#define UFSD_ENUM_SKIP_HIDDEN         0x00000004    // Skip hidden files
#define UFSD_ENUM_NAME_ID_ATTR_ONLY   0x00000008    // collect only name, Id and base attrib (file, dir, symlink)

//
// Possible options for CFileSystem::Init()
//

//
// NTFS specific:
//
#define NTFS_OPTIONS_ENABLE_SHORTNAME   0x00000001    // Generate short names when create/rename files
#define NTFS_OPTIONS_ENABLE_PSEUDONAME  0x00000002    // Use pseudo names to search in directories
#define NTFS_OPTIONS_CASE_SENSITIVE     0x00000004    // Use case sensitive names
#define NTFS_OPTIONS_BEST_COMPRESSION   0x00000008    // Use best compression instead of standard
#define NTFS_OPTIONS_INCLUDE_BCFS       0x00000010    // Accept 'NTFS' and 'BCFS' signature in boot
#define NTFS_OPTIONS_CREATE_SPARSE      0x00000020    // By default create sparse file

//
// ExFAT specific:
//
#define EXFAT_OPTIONS_CASE_SENSITIVE   0x00000004     // Use case sensitive names
#define EXFAT_OPTIONS_INCLUDE_BCFS     0x00000010     // Accept 'exFAT' and 'BCFAT' signature in boot

//
// ZIP specific:
//

//  0x00000001    // Bit 1
//  0x00000002    // Bit 2
//  0x00000004    // Bit 3
//  0x00000008    // Bit 4

#define ZIP_OPTIONS_WRITE               0x00000010    // 0 read only, 1 write only
#define ZIP_OPTIONS_WRITE_DEFAULT_COMPRESSION   (ZIP_OPTIONS_WRITE | 0x0000000F)

//
// Common
//
#define UFSD_CHECKFS_MODE               0x10000000    // Init FS from checkfs process
#define UFSD_OPTIONS_JNL                0x20000000    // Activate journal
#define UFSD_OPTIONS_RECOGNIZE_FS       0x40000000    // Just recognize file system
#define UFSD_OPTIONS_NOBUFFERING        0x80000000    // Do not buffer not aligned read/write

//
// Events while volume initialization/reinitialization
//
#define UFSD_FLAGS_JOURNAL              0x00000001    // Volume with journal
#define UFSD_FLAGS_REPLAYED             0x00000002    // Volume was replayed
#define UFSD_FLAGS_NEED_REPLAY          0x00000004    // Read-only volume requires replay
#define UFSD_FLAGS_LOGICAL_SECTOR       0x00000008    // Logical sector size does not match physical sector size ( HFSJ and more then 2^31 of sectors )
#define UFSD_FLAGS_NEED_REPLAY_NATIVE   0x00000010    // Native journal is not empty and not possible to replay
#define UFSD_FLAGS_INVALID_JOURNAL      0x00000020    // Native journal is not valid or not supported

} // namespace UFSD

#endif // __UFSD_FSBASE_H

