////////////////////////////////////////////////////////////////
// April 2012. (C) Paragon Software
// Valery Kovalev
//
// This file contain declaration and definitions of
// Interface container
//
////////////////////////////////////////////////////////////////


#ifndef __U_INTERFACE_H
#define __U_INTERFACE_H

//#include "ufsd.h"
#include "u_types.h"
#include "u_errors.h"
#include "u_mmngr.h"
#include "u_rwblck.h"
#include "u_string.h"
#include "u_time.h"
#include "u_list.h"
#include "u_msg.h"

#include "u_thread.h"
#include "u_rsaver.h"
#include "u_dedup.h"
#include "u_fileio.h"
#include "u_hash.h"
#include "u_cmpres.h"
#include "u_crypt.h"
#include "u_indexf.h"

namespace UFSD {


////////////////////////////////////////////////////////////////
#define INTERFACE_CMemoryManager        0xfa01 /*CMemoryManager*/
#define INTERFACE_CStrings              0xfa02 /*CStrings*/
#define INTERFACE_CTime                 0xfa03 /*CTime*/
#define INTERFACE_IMessage              0xfa04 /*IMessage*/
#define INTERFACE_CLog                  0xfa05 /*CLog*/

#define INTERFACE_IUThread              0xfa06 /*IUThread*/
#define INTERFACE_IRestartSaver         0xfa07 /*IRestartSaver*/
#define INTERFACE_IFileIo               0xfa08 /*IFileIo*/
#define INTERFACE_ICompressIo           0xfa09 /*ICompressIo*/
#define INTERFACE_ICryptIo              0xfa0a /*ICryptIo*/
#define INTERFACE_IHash                 0xfa0b /*IHash*/

#define INTERFACE_IFileBlockDedupServer 0xfa0c /*IFileBlockDedupServer*/
#define INTERFACE_VimMessage            0xfa0d /*VimMessage*/


//class that is used as container for different interface object
//like CMemoryManager IMessage and other
//where interface is macros INTERFACE_XXX
class UFSD_ABSTRACT_CLASS IInterface : public CMemBased<IInterface>
{
public:
  IInterface(
    IN CMemoryManager*  mm
  )
  : CMemBased<IInterface>(mm)
  {
  }

  //destruct object
  virtual void     Destroy() = 0;

  //get pointer at interface object
  virtual void*    QueryInterface(unsigned int InterfaceId) = 0;
  //save pointer at interface object
  virtual unsigned SaveInterface(unsigned int InterfaceId, void* pIObject) = 0;


  //for namespace UFSD::
  //use type without namespace prefix
  #define SaveI(type, obj) SaveInterface(INTERFACE_##type, (obj))
  #define QueryI(type)     __Query((UFSD::type*)NULL, INTERFACE_##type)

  //example:
  //UFSD::IInterface*     ii = ...;
  //UFSD::CMemoryManager* mm = ...;
  //...
  //ii->Save(CMemoryManager, mm);
  //...
  //mm = ii->Query(CMemoryManager);

  template<typename T>
  T* __Query(T* /*pNULL*/, int id)
  {
    return reinterpret_cast<T*>(QueryInterface(id));
  }
};


////////////////////////////////////////////////////////////////
UFSD_API IInterface* __stdcall UFSD_CreateIInterface(
    IN CMemoryManager*  mm
);


////////////////////////////////////////////////////////////////
#ifdef IINTERFACE_IMPLEMENTATION

////////////////////////////////////////////////////////////////
struct InterfaceList : public list_head
{
  unsigned int  InterfaceId;
  void*         pInterface;
};


class UInterface : public IInterface
{
  list_head        m_List; //InterfaceList

public:
  UInterface(IN CMemoryManager* mm)
  : IInterface(mm)
  {
    m_List.init();
  }

  ~UInterface();

  //destruct object
  virtual void     Destroy() {delete this;}
  //get pointer at interface object
  virtual void*    QueryInterface(unsigned int InterfaceId);
  //save pointer at interface object
  virtual unsigned SaveInterface(unsigned int InterfaceId, void* pIObject);
};


////////////////////////////////////////////////////////////////
UInterface::~UInterface()
{
  list_head* pos, * n;
  list_for_each_safe(pos, n, &m_List)
  {
    InterfaceList* e = list_entry(pos, InterfaceList, next);
    m_Mm->Free(e);
  }
}


////////////////////////////////////////////////////////////////
//get pointer at interface object
void*    UInterface::QueryInterface(unsigned int InterfaceId)
{
  list_head* pos;
  list_for_each(pos, &m_List)
  {
    InterfaceList* e = list_entry(pos, InterfaceList, next);
    if(e->InterfaceId == InterfaceId)
      return e->pInterface;
  }
  return NULL;
}


////////////////////////////////////////////////////////////////
//save pointer at interface object
unsigned UInterface::SaveInterface(unsigned int InterfaceId, void* pIObject)
{
  list_head* pos;
  list_for_each(pos, &m_List)
  {
    InterfaceList* e = list_entry(pos, InterfaceList, next);
    if(e->InterfaceId == InterfaceId)
    {
      //change interface
      e->pInterface  = pIObject;
      return ERR_NOERROR;
    }
  }
  InterfaceList* pInterface = (InterfaceList*)m_Mm->Malloc(sizeof(InterfaceList));
  if(!pInterface)
    return ERR_NOMEMORY;

  pInterface->InterfaceId = InterfaceId;
  pInterface->pInterface  = pIObject;

  pInterface->insert_after(&m_List);

  return ERR_NOERROR;
}


////////////////////////////////////////////////////////////////
UFSD_API IInterface* __stdcall UFSD_CreateIInterface(
    IN CMemoryManager*  mm
)
{
  if(!mm)
    return NULL;

  return new(mm) UInterface(mm);
}

#endif //IINTERFACE_IMPLEMENTATION


} //namespace UFSD


#endif //__U_INTERFACE_H
