/****************************************************
 *  File UFSD/include/memmngr.h
 *  Copyright (C) 2000 Paragon Software,
 *
 *  Here are two classes defined CMemoryManager and CMemoryBased.
 *
 *  All classes used in UFSD are derived from CMemBased.
 *  CMemoryManager is used for basic memory operations.
 ****************************************************/
#ifndef __UFSD_MMNGR_H
#define __UFSD_MMNGR_H


//
// Use inline memset/memcpy/memcmp/memmove, if possible
//

#ifndef UFSD_NO_CRT_USE

#if defined _MSC_VER && _MSC_VER < 1400
  //
  // VC 8 ignores intrinsic functions (BUG?)
  //
  extern "C"{
    void *  __cdecl memset(void *, int, size_t);
    void *  __cdecl memcpy(void *, const void *, size_t);
    int     __cdecl memcmp( const void *buf1, const void *buf2, size_t count );
  }

  #pragma intrinsic(memcpy, memset, memcmp)
  #define Memset2  memset
  #define Memcpy2  memcpy
  #define Memcmp2  memcmp
  #define UFSD_USE_BUILTIN_MEMCPY

#elif defined __GNUC__

  #define Memset2   __builtin_memset
  #define Memcpy2   __builtin_memcpy
  #define Memcmp2   __builtin_memcmp
  #define UFSD_USE_BUILTIN_MEMCPY
  #if __GNUC__ >= 4 && !defined __i386__
    #define Memmove2  __builtin_memmove
    #define UFSD_USE_BUILTIN_MEMMOVE
  #endif

  #if defined UFSD_DRIVER_LINUX && defined __LINUX_ARM_ARCH__
    extern "C" void   __memzero(void*, size_t);
    #define Memzero2  __memzero
  #elif defined UFSD_DRIVER_LINUX && defined  __mips__
    // unfortunately mips kernel does not export bzero/__bzero
  #else
    #define Memzero2  __builtin_bzero
  #endif

#endif

#endif // #ifndef UFSD_NO_CRT_USE

#ifndef Memset2
  #define Memset2  m_Mm->Memset
  #define Memcpy2  m_Mm->Memcpy
  #define Memcmp2  m_Mm->Memcmp
#endif

#ifndef Memzero2
  #define Memzero2(p,s) Memset2( (p), 0, (s) )
#endif

#ifndef Memmove2
  #define Memmove2  m_Mm->Memmove
#endif

#if defined(__APPLE__) && !defined(KERNEL)
  #include <stdlib.h>
#endif

namespace UFSD {

class UFSD_ABSTRACT_CLASS CMemoryManager{
protected:
    // Make compiler happy:
    CMemoryManager(){}
    CMemoryManager(const CMemoryManager&){}
    CMemoryManager& operator=(const CMemoryManager&){return *this;}

public:

#ifdef UFSD_DRIVER_LINUX

#ifndef UFSD_USE_BUILTIN_MEMCPY
    // Sets the first count chars of dest to the character c.
    virtual void Memset(
        IN void*          dest,
        IN unsigned char  c,
        IN size_t         count
        ) = 0;

    // Copies count bytes of 'from' to 'to'. If the source and destination overlap,
    // the behavior of Memcpy is undefined.
    // Use Memmove to handle overlapping regions.
    virtual void Memcpy(
        IN void*        to,
        IN const void*  from,
        IN size_t       count
        ) = 0;

    // Compares the first count bytes of buf1 and buf2 and returns a value indicating
    // their relationship.
    // < 0  b1 less than b2
    //   0  b1 identical to b2
    // > 0  b1 greater than b2
    virtual int Memcmp(
        IN const void* b1,
        IN const void* b2,
        IN size_t      c
        ) = 0;
#endif

#ifndef UFSD_USE_BUILTIN_MEMMOVE
    // Copies count bytes of characters from 'from' to 'to'. If some regions of the source
    // area and the destination overlap, Memmove ensures that the original source bytes
    // in the overlapping region are copied before being overwritten
    virtual void Memmove(
        IN void*        to,
        IN const void*  from,
        IN size_t       count
        ) = 0;
#endif

    // Returns total ram in bytes or 0 if not implemented
    virtual UINT64 GetTotalRam() const = 0;

    // Returns page size or 0 if not implemented
    virtual unsigned int GetPageSize() const = 0;

#else

    // Sets the first count chars of dest to the character c.
    virtual void Memset(
        IN void*          dest,
        IN unsigned char  c,
        IN size_t         count
        )
    {
      unsigned char* beg = (unsigned char*)dest;
      unsigned char* end = beg + count;
      switch (count & 7) {
      case 0:
        while (beg != end) {
              *beg++ = c;
      case 7: *beg++ = c;
      case 6: *beg++ = c;
      case 5: *beg++ = c;
      case 4: *beg++ = c;
      case 3: *beg++ = c;
      case 2: *beg++ = c;
      case 1: *beg++ = c;
        }
      }
    }

    // Copies count bytes of 'from' to 'to'. If the source and destination overlap,
    // the behavior of Memcpy is undefined.
    // Use Memmove to handle overlapping regions.
    virtual void Memcpy(
        IN void*        to,
        IN const void*  from,
        IN size_t       count
        )
    {
      unsigned char* beg = (unsigned char*)to;
      unsigned char* end = beg + count;
      const unsigned char* src = (const unsigned char*)from;
      switch (count & 7) {
      case 0:
        while (beg != end) {
              *beg++ = *src++;
      case 7: *beg++ = *src++;
      case 6: *beg++ = *src++;
      case 5: *beg++ = *src++;
      case 4: *beg++ = *src++;
      case 3: *beg++ = *src++;
      case 2: *beg++ = *src++;
      case 1: *beg++ = *src++;
        }
      }
    }

    // Compares the first count bytes of buf1 and buf2 and returns a value indicating
    // their relationship.
    // < 0  b1 less than b2
    //   0  b1 identical to b2
    // > 0  b1 greater than b2
    virtual int Memcmp(
        IN const void* b1,
        IN const void* b2,
        IN size_t      c
        )
    {
      const unsigned char* p1 = (const unsigned char*)b1;
      const unsigned char* p2 = (const unsigned char*)b2;
      while( 0 != c-- ){
        int d = *p1++ - *p2++;
        if ( d )
          return d;
      }
      return 0;
    }

    // Copies count bytes of characters from 'from' to 'to'. If some regions of the source
    // area and the destination overlap, Memmove ensures that the original source bytes
    // in the overlapping region are copied before being overwritten
    virtual void Memmove(
        IN void*        to,
        IN const void*  from,
        IN size_t       count
        )
    {
      if ( to <= from || (char *)to >= ((char *)from + count)) {
        //
        // Non-Overlapping Buffers
        //
        unsigned char* beg = (unsigned char*)to;
        unsigned char* end = beg + count;
        unsigned char* src = (unsigned char*)from;
        switch (count & 7) {
        case 0:
          while (beg != end) {
                *beg++ = *src++;
        case 7: *beg++ = *src++;
        case 6: *beg++ = *src++;
        case 5: *beg++ = *src++;
        case 4: *beg++ = *src++;
        case 3: *beg++ = *src++;
        case 2: *beg++ = *src++;
        case 1: *beg++ = *src++;
          }
        }
      } else {
        //
        // Overlapping Buffers
        //
        unsigned char* end = (unsigned char*)to;
        unsigned char* beg = end + count;
        const unsigned char* src = (const unsigned char*)from + count;
        switch (count & 7) {
        case 0:
          while (beg != end) {
                *--beg = *--src;
        case 7: *--beg = *--src;
        case 6: *--beg = *--src;
        case 5: *--beg = *--src;
        case 4: *--beg = *--src;
        case 3: *--beg = *--src;
        case 2: *--beg = *--src;
        case 1: *--beg = *--src;
          }
        }
      }
    }

    // Returns total ram in bytes or 0 if not implemented
    virtual UINT64 GetTotalRam() const { return 0; }

    // Returns page size or 0 if not implemented
    virtual unsigned int GetPageSize() const { return 0; }

#endif // #ifdef UFSD_DRIVER_LINUX

#ifndef UFSD_HeapAlloc
    // Malloc returns a void pointer to the allocated space, or NULL if there is
    // insufficient memory available
    virtual void* Malloc(
        IN size_t size
        )
#if defined __GNUC__ && (__GNUC__ >= 4 && __GNUC_MINOR__ >= 6)
        __attribute__((alloc_size(1)))
#endif
        = 0 ;

    // The Free function deallocates a memory block 'p' that was previously
    // allocated by a call to Malloc
    virtual void Free(
        IN void*  p
        ) = 0;
#endif

};


#if defined UFSD_DRIVER_LINUX && !defined UFSD_USE_GLOBAL_MM
  // Activate this define if you want to use the only global Memory manager for your project
  #define UFSD_USE_GLOBAL_MM
#endif


#ifdef UFSD_USE_GLOBAL_MM
//
// We save a lot of memory removing m_Mm from each ufsd object
//
extern CMemoryManager* m_Mm;
  #define GetMm( o )  m_Mm
#else
  #define GetMm( o ) (o)->m_Mm
#endif


#ifdef UFSD_HeapAlloc
  #define malloc( s )         UFSD_HeapAlloc( s, 0 )
  #define calloc( n, s )      UFSD_HeapAlloc( (s) * (n), 1 )
  #define free( p )           UFSD_HeapFree( p )
  #define Malloc2( s )        UFSD_HeapAlloc( s, 0 )
  #define Calloc2( m, n, s )  UFSD_HeapAlloc( (s) * (n), 1 )
  #define Free2( p )          UFSD_HeapFree( p )
#else
  #define Malloc2( s )  m_Mm->Malloc( s )
  #define Free2( p )    m_Mm->Free( p )
  static inline void* Calloc2( CMemoryManager* m_Mm, size_t num, size_t size ) {
    size *= num;
    void* p = m_Mm->Malloc( size );
    if ( NULL != p )
      Memzero2( p, size );
    return p;
  }
#endif


//  Class with overloaded "new/delete" operators.
//  Class is used to allocate and free memory
//  using our own memory manager.
template<class T>
class CMemBased{
private:
  // Make compiler happy (just declare but not implement):
  CMemBased();
  CMemBased(const CMemBased&);
  CMemBased& operator=(const CMemBased&);
  // By default new[] and delete[] are empty
  // redefine in your class if necessary
  static void* __cdecl operator new[] (size_t) throw() { return NULL; }
  static void __cdecl operator delete[](void*){}

protected:
  ~CMemBased(){} // non virtual!

#ifdef UFSD_USE_GLOBAL_MM
  CMemBased( CMemoryManager* ) {}
#else
  CMemBased( CMemoryManager * mm ) : m_Mm(mm) {}
public:
  CMemoryManager * m_Mm;
#endif

public:

  CMemoryManager* GetMemoryManager() const{
    return m_Mm;
  }

  //
  // Inplace operator new
  //
  static void* __cdecl operator new( size_t, void* p ) throw(){ return p; }

  //
  // operator new for CMemBased derived classes with virtual functions
  //
  static void* __cdecl operator new( size_t cb, CMemoryManager * m_Mm ) throw(){
#ifdef UFSD_HeapAlloc
    return UFSD_HeapAlloc( cb, 1 );
#else
    void* p = m_Mm->Malloc( cb );
#ifndef UFSD_MALLOC_CANT_FAIL
    if ( NULL == p )
      return NULL;
#endif
    Memzero2( p, cb );
    return p;
#endif
  }

  //
  // operator delete for CMemBased derived classes
  //
  static void __cdecl operator delete(void* p){
#ifdef UFSD_HeapAlloc
    UFSD_HeapFree( p );
#else
    GetMm( (T*)p )->Free( p );
#endif
  }

#ifndef UFSD_HeapAlloc
  //=======================================================
  // a-la "C" runtime memory related functions
  //=======================================================
  void* malloc( size_t size ) const
#if defined __GNUC__ && (__GNUC__ >= 4 && __GNUC_MINOR__ >= 6)
        __attribute__((alloc_size(1)))
#endif
  {
    return m_Mm->Malloc( size );
  }

  void* calloc( size_t num, size_t size ) const{
    size *= num;
    void* p = m_Mm->Malloc( size );
#ifndef UFSD_MALLOC_CANT_FAIL
    if ( NULL == p )
      return NULL;
#endif
    Memzero2( p, size );
    return p;
  }

  void free( void* p ) const{
    m_Mm->Free( p );
  }

  //=======================================================
  // End of "C" runtime memory related functions
  //=======================================================
#endif
};

} // namespace UFSD {

#endif // __UFSD_MMNGR_H
