/****************************************************
 *  File rwblck.h
 *  Copyright (C) 2000 Paragon Software,
 *
 *  Here is the definition of CRWBlock class
 *  for low level block read/write operations.
 ****************************************************/

#ifndef __UFSD_RWBLOCK_H
#define __UFSD_RWBLOCK_H


// Possible options for SetWait
#define UFSD_RW_WAIT_SYNC        0x01
#define UFSD_RW_WAIT_BARRIER     0x02

// Possible options for Map
#define UFSD_RW_MAP_NO_READ       0x01
//#define UFSD_RW_MAP_FIXUP         0x02

#ifndef UFSDAPI_CALL
  #define UFSDAPI_CALL
#endif

extern "C"{
struct buffer_head;
struct super_block;
}

namespace UFSD {

#ifndef UFSD_DRIVER_LINUX
// default source id {144F461A-0377-4403-8D78-F15AB1F255AF}
// no additional parameters
static const GUID UFSD_ERROR_SOURCE_ID = { 0x144f461a, 0x377, 0x4403, { 0x8d, 0x78, 0xf1, 0x5a, 0xb1, 0xf2, 0x55, 0xaf } };

// Log class
// http://portal.paragon-software.com/SiteDirectory/cto/Development Documents/Active Projects/PRM 3.0/Architecture/Error Handling.docx
class CLog
{
public:

   // add error to log
   virtual void AddError(
       const GUID&  sourceId,                  // source identifier
       unsigned int errCode,                   // error code, specific to the source
       const char*  file,                      // source file
       unsigned int line,                      // line number
       const char*  description,               // text description for the error (can be null)
       const void*  param,                     // buffer containing additional parameters
       unsigned int paramLen                   // buffer length
       )                                       = 0;

   // add message
   virtual void AddMessage(
       const char* format,                     // format string
       ...
       )                                       = 0;

   // add message
   virtual void AddVMessage(
       const char* format,                     // format string
       va_list     arg                         // argument list
       )                                       = 0;
};
#endif // UFSD_DRIVER_LINUX

struct BcbBuf{
  struct buffer_head* Bcb;
  void* Buf;
};


class UFSD_ABSTRACT_CLASS CRWBlock
{
protected:
    // Make compiler happy:
    CRWBlock(){}
    CRWBlock(const CRWBlock&){}
    CRWBlock& operator=(const CRWBlock&);

public:

    // =================================
    //   PURE VIRTUAL FUNCTIONS
    // =================================

    //
    // Read blocks [Blck, Blck+BlckCount) in 'Buffer'.
    // Returns:
    //  0 - if success
    //  error code otherwise
    //
    virtual unsigned ReadBlocks(
        IN  const UINT64& Blck,
        OUT void *        Buffer,
        IN  size_t        BlckCount
        ) = 0;

    //
    // Write 'Buffer' to blocks [Blck, Blck+BlckCount).
    // Returns:
    //  0 - if success
    //  error code otherwise
    virtual unsigned WriteBlocks(
        IN const UINT64& Blck,
        IN const void *  Buffer,
        IN size_t        BlckCount
        ) = 0;

    // Retrieve block size in bytes
    virtual unsigned int GetBlockSize(
        ) = 0;

    // Sets the logical block size in bytes
    virtual void SetBlockSize(
        IN unsigned int NewBlockSize
        ) = 0;

    //
    // Retrieve number of blocks available with RWBlock
    // Returns:
    //  0 - if success;desc non zero;
    //  error code otherwise.
    virtual UINT64 GetNumberOfBlocks() = 0;

#ifdef UFSD_DRIVER_LINUX

    // flush cache buffers
    // returns  zero if OK, else UFSD error code.
    virtual unsigned Flush() = 0;

    // returns non zero for read-only mode
    virtual int IsReadOnly() const = 0;

    // Returns size of sector in bytes
    virtual unsigned int GetSectorSize() const = 0;

    // Discard range of bytes (known as trim for SSD)
    virtual unsigned DiscardRange(
        IN  const UINT64& Offset,
        IN  const UINT64& Bytes
        ) = 0;

    //
    // Set mode of next writes
    // Returns previous mode
    //
    virtual size_t SetWait(
        IN size_t Flags // - one of the UFSD_RW_WAIT_XXX
        ) = 0;

    //
    // Map/UnMap/SetDirty functions
    //
    virtual unsigned Map(
        IN const UINT64&  Offset,
        IN size_t         Bytes,
        IN size_t         Flags,  // - one of the UFSD_RW_MAP_XXX
        OUT struct buffer_head**  Bcb,
        OUT void**        Buffer
        ) = 0;

    virtual void UnMap(
        IN struct buffer_head*  Bcb,
        IN bool   Forget = false
        ) = 0;

    virtual unsigned SetDirty(
        IN struct buffer_head*  Bcb,
        IN size_t   Wait  = 0
        ) = 0;

    virtual void LockBuffer(
        IN struct buffer_head*  Bcb
        ) = 0;

    virtual void UnLockBuffer(
        IN struct buffer_head*  Bcb
        ) = 0;

    //
    // Zero range of bytes [Offset, Offset + Bytes)
    //
    virtual unsigned Zero(
        IN  const UINT64& Offset,
        IN  size_t        Bytes
        ) = 0;

    // =================================
    //   END OF PURE VIRTUAL FUNCTIONS
    // =================================

#else

    // =================================
    //   END OF PURE VIRTUAL FUNCTIONS
    // =================================


    // =================================
    //   NOT PURE VIRTUAL FUNCTIONS
    // =================================

    // flush cache buffers
    // returns  zero if OK, else UFSD error code.
    virtual unsigned Flush() { return 0; }

    // returns non zero for read-only mode
    virtual int IsReadOnly() const { return 0; }

    // Returns size of sector in bytes
    virtual unsigned int GetSectorSize() const { return 512; }

    // Discard range of bytes (known as trim for SSD)
    virtual unsigned DiscardRange(
        IN  const UINT64& ,//Offset,
        IN  const UINT64& //Bytes
        )
    {
      return ERR_NOTIMPLEMENTED;
    }

    // Verifies disk. ( A-la surface test )
    // Returns 0 if block is OK
    virtual unsigned Verify(
        IN const UINT64&, // Blck
        IN size_t         // BlckCount
        )
    {
      // Default implementation does nothing
      return 0;
    }

    //
    // Set mode of next writes
    // Returns previous mode
    //
    virtual size_t SetWait(
        IN size_t // Flags - one of the UFSD_RW_WAIT_XXX
        )
    {
      return 0;
    }

    // Function a-la DeviceIoControl
    virtual unsigned IoControl(
        IN  size_t          ,//FsIoControlCode,
        IN  const void*     = NULL, // OPTIONAL
        IN  size_t          = 0,    // OPTIONAL
        OUT void*           = NULL, // OPTIONAL
        IN  size_t          = 0,    // OPTIONAL
        OUT size_t*         = NULL  // OPTIONAL
        )
    {
      return ERR_NOTIMPLEMENTED;
    }

    // =================================
    //   END OF NOT PURE VIRTUAL FUNCTIONS
    // =================================


    ///////////////////////////////////////////////////////////
    // Helper class that allows to:
    // Switch temporary the size of block to NewBytesPerBlock
    // and then switch block size back to OldBytesPerBlock
    ///////////////////////////////////////////////////////////
    class CBlockSize : public ufsd_noncopyable{
      CRWBlock*     m_Rw;
      unsigned int  m_OldBytesPerBlock;

    public:
      CBlockSize( IN CRWBlock*      Rw,
                  IN unsigned int   NewBytesPerBlock,
                  IN unsigned int   OldBytesPerBlock )
        : m_Rw( Rw )
        , m_OldBytesPerBlock( OldBytesPerBlock )
      {
        Rw->SetBlockSize( NewBytesPerBlock );
      }
      CBlockSize( IN CRWBlock*      Rw,
                  IN unsigned int   NewBytesPerBlock )
        : m_Rw( Rw )
        , m_OldBytesPerBlock( Rw->GetBlockSize() )
      {
        Rw->SetBlockSize( NewBytesPerBlock );
      }

      ~CBlockSize()
      {
        m_Rw->SetBlockSize( m_OldBytesPerBlock );
      }
    };
#endif // #ifdef UFSD_DRIVER_LINUX

};

#if !defined UFSD_DRIVER_LINUX || defined UFSD_NTFS_512_4K || defined NEW_ANDROID_FUSE

enum UFSD_RWIOCONTROL{
  RWIOCTL_GET_RWBLOCK_TYPE              = 1,
  RWIOCTL_SET_BADBLOCKS                 = 2,
  RWIOCTL_GET_FILERWBLOCK_FREESPACE     = 4,
  RWIOCTL_SET_RW_ACCESS_TYPE            = 6,

  // Some compilers can use BYTE or WORD for enumerators
  // depending on enumerator values
  RWIOCTL_PLACEHOLDER           = 0xFFFFFFFFu
};


//===================================================================
//
// RWIOCTL_GET_RWBLOCK_TYPE
//
// input  - none
//
// output - size_t RWBLOCK_XXX
#define RWBLOCK_PROTECTED 0xff000001


//===================================================================
//
// RWIOCTL_SET_BADBLOCKS
//
// input  - RWBADBLOCKS_BUFFER
//
// output - none
//
// NOTE: Update bad blocks information.
//

struct RWBADBLOCKS_BUFFER {
  unsigned int Cmd;             // 0-add 1-clear 2-clear all BB list
  unsigned int ExtentCount;     // Count of Extents structure elements in the present RWBADBLOCKS_BUFFER structure
  struct {
      unsigned int  BadLLsn;    // The LSN at which the current bad blocks begins on the volume.
      unsigned int  BadHLsn;    // low and high parts
      unsigned int  Len;        // The number of continuous bad sectors
  } Extents[1];
};


//===================================================================
//
// RWIOCTL_SET_BLOCK_SIZE
//
// input  - 'unsigned int' casted to 'void*'
//
// output - none
//

//===================================================================
//
// RWIOCTL_GET_FILERWBLOCK_FREESPACE
//
// input  - none
//
// output - UINT64 size of free space in bytes on host volume
//          (need for increasing size of file based RWBlock)


//===================================================================
//
// RWIOCTL_GET_SUPER_BLOCK
//
// input  - none
//
// output - pointer to pointer
//

//===================================================================
//
// RWIOCTL_SET_RW_ACCESS_TYPE
// specify data and access type for read/write operation
// (RW_DATA_TYPE_XXX | RW_ACCESS_XXX)
// not forgot to set default type RW_DATA_TYPE_ANY before exiting
//
// input  - 'unsigned int' data type and access
//
// output - none
//
//data type
#define RW_DATA_TYPE_ANY        0       /*default read/write any(mixed) data type*/
#define RW_DATA_TYPE_LAYOUT     1       /*read/write layout data (fdisk) only*/
#define RW_DATA_TYPE_METADATA   2       /*read/write FS metadata only*/
#define RW_DATA_TYPE_FILEDATA   3       /*read/write file data only*/

//access type
#define RW_ACCESS_SEQUENTIAL    0x10    /*any data sectors will be writed once*/

//
// Helper class to overwrite only some of functions (e.g. ReadBlocks/WriteBlocks)
//
struct CRWBlockA : public CRWBlock NONCOPYABLE{

  CRWBlock* m_Rw;

  virtual unsigned ReadBlocks(
      IN const UINT64&  Blck,
      IN void*          Buffer,
      IN size_t         BlckCount
      )
  {
    return m_Rw->ReadBlocks( Blck, Buffer, BlckCount );
  }

  virtual unsigned WriteBlocks(
      IN const UINT64&  Blck,
      IN const void*    Buffer,
      IN size_t         BlckCount
      )
  {
    return m_Rw->WriteBlocks( Blck, Buffer, BlckCount );
  }

  virtual unsigned int GetBlockSize() { return m_Rw->GetBlockSize(); }

  virtual void SetBlockSize( IN unsigned int NewBlockSize ) { m_Rw->SetBlockSize( NewBlockSize ); }

  virtual UINT64 GetNumberOfBlocks() { return m_Rw->GetNumberOfBlocks(); }

  virtual unsigned Flush(){ return m_Rw->Flush(); }

  virtual int IsReadOnly() const { return m_Rw->IsReadOnly(); }

  virtual unsigned int GetSectorSize() const { return m_Rw->GetSectorSize(); }

  virtual unsigned DiscardRange( const UINT64& Offset, const UINT64& Bytes ) { return m_Rw->DiscardRange( Offset, Bytes ); }

  virtual size_t SetWait( size_t Flags ) { return m_Rw->SetWait( Flags ); }

#ifdef UFSD_DRIVER_LINUX
  virtual unsigned Map(
      IN const UINT64& Offset, IN size_t Bytes, IN size_t Flags,
      OUT struct buffer_head** Bcb, OUT void** Buffer )
  {
    return m_Rw->Map( Offset, Bytes, Flags, Bcb, Buffer );
  }

  virtual void UnMap( struct buffer_head* bh, bool Forget ) { m_Rw->UnMap( bh, Forget ); }
  virtual unsigned SetDirty( struct buffer_head* bh, size_t w ) { return m_Rw->SetDirty( bh, w ); }
  virtual void LockBuffer( struct buffer_head* bh ) { m_Rw->LockBuffer( bh ); }
  virtual void UnLockBuffer( struct buffer_head* bh ) { m_Rw->UnLockBuffer( bh ); }
  virtual unsigned Zero( const UINT64& Offset, size_t Bytes ) { return m_Rw->Zero( Offset, Bytes ); }
#else
  virtual unsigned IoControl(
        IN  size_t          FsIoControlCode,
        IN  const void*     InBuffer       = NULL,
        IN  size_t          InBufferSize   = 0,
        OUT void*           OutBuffer      = NULL,
        IN  size_t          OutBufferSize  = 0,
        OUT size_t*         BytesReturned  = NULL
        )
  {
    return m_Rw->IoControl( FsIoControlCode, InBuffer, InBufferSize, OutBuffer, OutBufferSize, BytesReturned );
  }
  virtual unsigned Verify( const UINT64& Blck, size_t BlckCount ) { return m_Rw->Verify( Blck, BlckCount ); }
#endif

};

#endif // #if !defined UFSD_DRIVER_LINUX || defined UFSD_NTFS_512_4K


///////////////////////////////////////////////////////////
// GetLog2OfBlockSize
//
// Simple function that returns log2 for most common BytesPerBlock
///////////////////////////////////////////////////////////
static inline
unsigned char
GetLog2OfBlockSize(
    IN unsigned int BytesPerBlock
    )
{
  switch( BytesPerBlock )
  {
    case 1<<9:  return 9;   // 512
    case 1<<10: return 10;  // 1K
    case 1<<11: return 11;  // 2K
    case 1<<12: return 12;  // 4K
    case 1<<13: return 13;  // 8K
    case 1<<14: return 14;  // 16K
    case 1<<15: return 15;  // 32K
    case 1<<16: return 16;  // 64K
    case 1<<17: return 17;  // 128K
    case 1<<18: return 18;  // 256K
    case 1<<19: return 19;  // 512K
    case 1<<20: return 20;  // 1M
    case 1<<21: return 21;  // 2M
    case 1<<22: return 22;  // 4M
    case 1<<23: return 23;  // 8M
    case 1<<24: return 24;  // 16M
    case 1<<25: return 25;  // 32M
  }
  return 9; // default
}

} // namespace UFSD

#endif //__UFSD_RWBLOCK_H

