/****************************************************
 *  File UFSD/include/strings.h
 *  Copyright (C) 2000 Paragon Software,
 *
 *  Here is the definition for Base class CStrings
 *  of strings object used in the UFSD
 *
 ****************************************************/

#ifndef __UFSD_STRINGS_H
#define __UFSD_STRINGS_H

namespace UFSD {

//
//  StrASCII means:
// - OEM character set for names in FAT
// - UTF-8 for names in SysV/Htfs/Ext2/Reiser
//
//  StrUNICODE means:
// - UTF-16 for names in FAT/NTFS/Reg/HFS+
//

typedef char STRType;

const char StrUnknown = 0;
const char StrASCII   = 1;
const char StrUNICODE = 2;



///////////////////////////////////////////////////////////
// GetCharSize
//
//Returns character size
//  for unicode it is sizeof(unsigned short)
//  for ascii it is sizeof(unsigned char)
//if bad type is passed then 0 is returned
///////////////////////////////////////////////////////////
static inline unsigned
GetCharSize(
    IN STRType type
    )
{
//  C_ASSERT( StrASCII == sizeof(char) );
//  C_ASSERT( StrUNICODE == sizeof(unsigned short) );
//  return type;
  switch (type)
  {
    case StrASCII   : return sizeof( unsigned char );
    case StrUNICODE : return sizeof( unsigned short );
    default         : return 0;
  }
}

///////////////////////////////////////////////////////////
// IsEmpty
//
// Returns true if string is empty
///////////////////////////////////////////////////////////
static inline bool
IsEmpty(
    IN STRType      type,
    IN const void*  name
    )
{
  if ( StrASCII == type )
    return 0 == *(const char*)name;
  if ( StrUNICODE == type )
    return 0 == *(const short*)name;
  return true;
}


#if defined __GNUC__ & defined toupper
 #undef toupper
#endif

class UFSD_ABSTRACT_CLASS CStrings
{
protected:
    // Make compiler happy:
    CStrings(){}
    CStrings(const CStrings&);
    CStrings& operator=(const CStrings&);

public:

    //Convert string strSrc from UNICODE to ASCII and vice versa
    //Result is in strDest.
    //  SrcLen - the length of source string in characters
    //            if (long)SrcLen < 0 then zero terminated string is assumed
    //  MaxOut - number of characters in output buffer
    //  DstLen - the length of destination string in characters
    //Returns:
    //  0 - if success
    virtual unsigned Convert(
        IN  STRType       SrcType,
        IN  const void   *SrcStr,
        IN  size_t        SrcLen,
        OUT void         *DstStr,
        IN  size_t        MaxOut,
        OUT size_t*       DstLen// OPTIONAL
        ) = 0;

    //Retrieve string length - the number of characters until zero
    virtual size_t strlen(
        IN STRType     type,
        IN const void *str
        ) = 0;

    // Compare not zero terminated filename and mask
    // type       - string type ASCII or UNICODE
    // name       - string of length "len"
    // len        - length of name
    // mask       - zero terminated mask
    // MetaMatch  - "inverted" sign. ( If MetaMatch is set then MatchMask is equivalent strn(i)cmp )
    // Returns 0 if name matches the mask
    virtual unsigned MatchnMask(
        IN STRType       type,
        IN const void   *name,
        IN const void   *mask,
        IN size_t        len,
        IN unsigned      MetaMatch = 0
        ) = 0;

    // Exclude this bullshit from driver
#ifndef UFSD_DRIVER_LINUX

    // The strcpy function copies strSource, including the terminating null character,
    // to the location specified by strDestination.
    // The behavior of strcpy is undefined if the source and destination strings overlap.
    // type - string type ASCII or UNICODE
    virtual void strcpy(
        IN STRType      type,
        IN OUT void  *  strDestination,
        IN const void*  strSource
        ) = 0;

    // The strncpy function copies the initial count characters of strSource
    // to strDest. If count is less than or equal to the length of strSource,
    // a null character is not appended automatically to the copied string.
    // If count is greater than the length of strSource, the destination string
    // is padded with null characters up to length count.
    // The behavior of strncpy is undefined if the source and destination strings overlap.
    // type - string type ASCII or UNICODE
    virtual void strncpy(
        IN STRType       type,
        IN OUT void   *  strDestination,
        IN const void *  strSource,
        IN size_t        Count
        ) = 0;

    // Converts a given lowercase letter to an uppercase letter if possible and appropriate.
    virtual unsigned toupper(
        IN unsigned c
        ) = 0;

    // The strcmp function compares string1 and string2 lexicographically
    // and returns a value indicating their relationship.
    // type - string type ASCII or UNICODE
    // Returns:
    //   < 0  - string1 less than string2
    //    0   - string1 is identical to string2
    //   > 0  - string1 greater than string2
    virtual int strcmp(
        IN STRType      type,
        IN const void * string1,
        IN const void * string2
        ) = 0;

    // The strcmp function lexicographically compares lowercase versions of string1 and string2
    // and returns a value indicating their relationship.
    // type - string type ASCII or UNICODE
    // Returns:
    //   < 0  - string1 less than string2
    //    0   - string1 is identical to string2
    //   > 0  - string1 greater than string2
    virtual int stricmp(
        IN STRType     type,
        IN const void *string1,
        IN const void *string2
        ) = 0;

    // The strnicmp function lexicographically compares, at most, the first count
    // characters of string1 and string2. The comparison is performed without regard to case;
    // strnicmp is a case-insensitive version of strncmp.
    // The comparison ends if a terminating null character is reached in either string
    // before count characters are compared. If the strings are equal when a terminating
    // null character is reached in either string before count characters are compared,
    // the shorter string is lesser.
    // Two strings containing characters located between 'Z' and 'a' in the ASCII table
    // ('[', '\', ']', '^', '_', and '`') compare differently, depending on their case.
    // For example, the two strings "ABCDE" and "ABCD^" compare one way if the comparison
    // is lowercase ("abcde" > "abcd^") and the other way ("ABCDE" < "ABCD^") if it is uppercase.
    //  type - string type ASCII or UNICODE
    // Returns:
    //   < 0  - string1 substring less than string2 substring
    //    0   - string1 substring identical to string2 substring
    //   > 0  - string1 substring greater than string2 substring
    virtual int strnicmp(
        IN STRType       type,
        IN const void   *string1,
        IN const void   *string2,
        IN size_t        count
        ) = 0;

    // The strncmp function lexicographically compares, at most, the first count
    // characters in string1 and string2 and returns a value indicating the relationship
    // between the substrings
    //type - string type ASCII or UNICODE
    // Returns:
    //   < 0  - string1 substring less than string2 substring
    //    0   - string1 substring identical to string2 substring
    //   > 0  - string1 substring greater than string2 substring
    virtual int strncmp(
        IN STRType       type,
        IN const void   *string1,
        IN const void   *string2,
        IN size_t        count
        ) = 0;

    // The strcat function appends strSource to strDestination and terminates the resulting
    // string with a null character. The initial character of strSource overwrites
    // the terminating null character of strDestination.
    // The behavior of strcat is undefined if the source and destination strings overlap.
    // type - string type ASCII or UNICODE
    virtual void strcat(
        IN STRType     type,
        IN OUT void   *strDestination,
        IN const void *strSource
        ) = 0;

    // The strstr function returns a pointer to the first occurrence of strSearch in string.
    // The search does not include terminating null characters.
    // type - string type ASCII or UNICODE
    // Returns:
    //  NULL - if no string found
    //  pointer to the substring if it is found
    virtual const void* strstr(
        IN STRType     type,
        IN const void *string,
        IN const void *strSearch
        ) = 0;

    // The strchr function finds the first occurrence of c in string, or it returns NULL
    // if c is not found.
    // The null-terminating character is included in the search.
    // type - string type ASCII or UNICODE
    virtual const void* strchr(
        IN STRType     type,
        IN const void *string,
        IN unsigned    c
        ) = 0;

    // The strrchr function finds the last occurrence of c in string.
    // The search includes the terminating null character.
    // type - string type ASCII or UNICODE
    // Returns a pointer to the last occurrence of c in string, or NULL if c is not found.
    virtual const void* strrchr(
        IN STRType     type,
        IN const void *string,
        IN unsigned    c
        ) = 0;

    //Convert OEM string strSrc to STRType string (UNICODE or ASCII)
    //Result is in strDest.
    //  SrcLen - the length of source string in characters
    //            if (long)SrcLen < 0 then zero terminated string is assumed
    //  MaxOut - number of characters in output buffer
    //  DstLen - the length of destination string in characters
    // Parameter type defines type of result string StrASCII or StrUNICODE.
    //Returns:
    //  0 - if success
    // -1 - otherwise
    // The default implementation just does call Convert for unicode strings or copy for ASCII strings

    virtual unsigned ConvertOEM(
        IN  STRType       type,
        IN  const char   *strSrc,
        IN  size_t        SrcLen,
        OUT void         *strDest,
        IN  size_t        MaxOut,
        OUT size_t*       DstLen// OPTIONAL
        )
    {
        switch(type)
        {
        case StrASCII:
            strcpy(StrASCII, strDest, strSrc);
            return 0;
        case StrUNICODE:
            return Convert(StrASCII, strSrc, SrcLen, strDest, MaxOut, DstLen);
        default:
            return unsigned(-1);
        }
    }

    //Convert STRType string (UNICODE or ASCII) to OEM string strSrc
    //Result is OEM string in strDest.
    //  SrcLen - the length of source string in characters
    //            if (long)SrcLen < 0 then zero terminated string is assumed
    //  MaxOut - number of characters in output buffer
    //  DstLen - the length of destination string in characters
    // Parameter type defines type of source string StrASCII or StrUNICODE.
    //Returns:
    //  0 - if success
    // -1 - otherwise
    // The default implementation just call Convert for unicode strings or copy for ASCII strings

    virtual unsigned ConvertToOEM(
        IN  STRType       type,
        IN  const void   *strSrc,
        IN  size_t        SrcLen,
        OUT char         *strDest,
        IN  size_t        MaxOut,
        OUT size_t*       DstLen,// OPTIONAL
        IN  bool          // bUpcase
        )
    {
        switch(type)
        {
        case StrASCII:
            strcpy(StrASCII, strDest, strSrc);
            return 0;
        case StrUNICODE:
            return Convert(StrUNICODE, strSrc, SrcLen, strDest, MaxOut, DstLen);
        default:
            return unsigned(-1);
        }
    }

    // Compare filename and mask
    // type       - string type ASCII or UNICODE
    // name       - zero terminated name
    // mask       - zero terminated mask
    // MetaMatch  - "inverted" sign. ( If MetaMatch is set then MatchMask is equivalent str(i)cmp )
    // Returns 0 if name matches the mask
    virtual unsigned MatchMask(
        IN STRType     type,
        IN const void *name,
        IN const void *mask,
        IN unsigned    MetaMatch = 0
        ) = 0;

#endif // #ifndef UFSD_DRIVER_LINUX
};

} // namespace UFSD


#endif //__UFSD_STRINGS_H

