/****************************************************
 *  File UFSD/include/types.h
 *  Copyright (C) 2000 Paragon Software,
 *
 *  Here must be placed types redefinitions for UFSD.
 ****************************************************/

//*******************************************************
//
// Alexander Mamaev:
// NOTE: According to "Preparing Code for the IA-64 Architecture":
//       http://www.intel.com/cd/ids/developer/emea/rus/dc/64bit/17969.htm
//
// The UNIX/64 and Windows 2000 (64-bit) data models have slightly diverged, differing on
// some data types. They still share the same meaning for the types char, short, int,
// float, and double. However, in Windows 2000 (64-bit), long is still 32 bits, while in
// UNIX/64 long is 64 bits. In addition, there is no ANSI C/C++ standardization for a 64-bit
// model, as of yet.
//
//*******************************************************

#ifndef __UFSD_TYPES_H
#define __UFSD_TYPES_H

//
// We need to know what is 'size_t'
//
#ifndef KERNEL
  #include <stddef.h>
  #include <stdarg.h>
#endif

#ifdef __APPLE__
  #include <sys/types.h>
  #include <stdarg.h>
#endif


#ifndef NULL
  #define NULL   0L
#endif


// there is a lot of cases when size does matter, i.e.
// for kernel-mode drivers. __declspec(novtable) saves
// from vtables for pure virtual classes to be created.
#if !defined(__declspec) && (!defined(_MSC_VER) || (_MSC_VER < 1200))
  #define __declspec(a)
#endif

#if !defined(UFSD_ABSTRACT_CLASS)
  #if defined(_MSC_VER) && (_MSC_VER >= 1200)
    #define UFSD_ABSTRACT_CLASS __declspec(novtable)
  #else
    #define UFSD_ABSTRACT_CLASS
  #endif
#endif


// Some functions require explicit call declaration
// especially when UFSD is used as external library.
// Gnu requires different form of stdcall, cdecl
// so do not use it now
#ifdef __GNUC__
  #ifndef __stdcall
    #define __stdcall
    #define _stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
    #define _cdecl
  #endif
#endif

#ifndef UFSD_API
  #ifdef UFSD_DLL
    // This define is used to export classes.
    // When UFSD is compiled as a part of DLL
    // this define must be set to __declspec(dllexport)
    // When the DLL will be used this define
    // must be set to  __declspec(dllimport)
    #ifdef UFSD_EXPORTS
      #ifdef __GNUC__
        #define UFSD_API __attribute__ ((visibility ("default")))
      #else
        #define UFSD_API __declspec(dllexport)
      #endif
    #else
      #define UFSD_API __declspec(dllimport)
    #endif

  #else
    // Use this way when UFSD's sources are used
    // to compile a project
    #define UFSD_API
  #endif
#endif

#if defined _MSC_VER
  // These #pragma warning allows to compile UFSD at level 4 warning
  #pragma warning( disable : 4291 ) //'declaration' : no matching operator delete found;
                                    // memory will not be freed if initialization throws an exception
  #pragma warning( disable : 4514 ) // "..." : unreferenced inline function has been removed
//  #pragma warning( disable : 4711 ) // function "..." selected for automatic inline expansion
//  #pragma warning( disable : 4710 ) // function "..." not inlined
  #pragma warning( disable : 4511 ) // copy constructor could not be generated
  #pragma warning( disable : 4512 ) // assignment operator could not be generated
//  #pragma warning( disable : 4328 ) // indirection alignment of formal parameter

  #ifdef UFSD_DLL
    #pragma warning( disable : 4275 ) // non dll-interface class 'ufsd_noncopyable' used as base for dll-interface class
  #endif
#endif

#ifndef FlagOn
  #define SetFlag(flags, flag) (flags |= (flag))
  #define ClearFlag(flags, flag) (flags &= ~(flag))
  #define FlagOn(flags, flag) ( 0 != ((flags) & (flag)))
  #define FlagOnXact(flags, flag1, flag2) ( (flag2) == ((flags) & (flag1)) )
#endif

#ifndef IN
  #define IN
  #define OUT
#endif

#ifndef ARRSIZE
  #define ARRSIZE(x) (sizeof((x))/sizeof((x)[0]))
#endif

#if !defined _SIZE_T_DEFINED && !defined _SIZE_T
  #ifdef _WIN64
    typedef unsigned __int64 size_t;
    typedef signed __int64 ptrdiff_t;
    #define __WORDSIZE  64
  #elif defined _WIN32
    typedef unsigned int size_t;
    typedef signed int ptrdiff_t;
    #define __WORDSIZE  32
  #elif defined __unix
    #if defined (__LP64__) || ( __WORDSIZE == 64)
      typedef unsigned long long size_t;
      typedef signed long long ptrdiff_t;
    #elif (__GNUC__ <= 2 && __GNUC_MINOR__ <= 95) || (__GNUC__ >= 4 && __GNUC_MINOR__ >= 0)
      typedef unsigned long size_t;
      typedef signed long ptrdiff_t;
    #else
      typedef unsigned int size_t;
      typedef signed int ptrdiff_t;
    #endif
  #endif
  #define _SIZE_T_DEFINED
  #define _SIZE_T
#endif


//
// Helper macroses
//
#ifndef QuadAlign
  #define QuadAlign( n )      (((n) + 7) & (~7) )
  #define IsQuadAligned( n )  (0 == ((size_t)(n) & 7))
  #define Quad2Align( n )     (((n) + 15) & (~15) )
  #define IsQuad2Aligned( n ) (0 == ((size_t)(n) & 15))
  #define Quad4Align( n )     (((n) + 31) & (~31) )
  #define IsSizeTAligned( n ) (0 == ((size_t)(n) & (sizeof(size_t)-1)))
  #define DwordAlign( n )     (((n) + 3) & (~3) )
  #define IsDwordAligned( n ) (0 == ((size_t)(n) & 3))
  #define WordAlign( n )      (((n) + 1) & (~1) )
  #define IsWordAligned( n )  (0 == ((size_t)(n) & 1))
#endif

#ifndef Add2Ptr
  #define Add2Ptr(P,I)   ((unsigned char*)(P) + (I))
  #define PtrOffset(B,O) ((size_t)((size_t)(O) - (size_t)(B)))
#endif

//
// This macro defines a simple unsigned value 0xFFFFFFFF (~0U) or 0xFFFFFFFFFFFFFFFF(~0ULL)
//
#define MINUS_ONE_T    ((size_t)(-1))

//
// UFSD_BIGENDIAN - external define
// Should the UFSD be compiled in bigendian memory model?
//

//#if (('1234' >> 24) != '1')  // == if little endian
//  #define UFSD_BIGENDIAN
//#endif

#ifdef UFSD_BIGENDIAN
  #define BE_ONLY(x)  x
  #define LE_ONLY(x)
#else
  #define BE_ONLY(x)
  #define LE_ONLY(x)  x
#endif

// Define empty __attribute__ for compilers that does not support it
#if !defined(__attribute__)
  #if !defined(__GNUC__)
    #define __attribute__(a)
  #endif
#endif


// Not all compilers understand #pragma pack( push, 1 )/#pragma pack( pop )
// To suppress warning use this define
#if defined __arm__ || defined __mips__
  #define NO_PRAGMA_PACK
#endif

//
//   compiler generates default constructor, copy constructor and operator= .
//   Nobody should call these functions anytime.
//   To reduce code size we just declare but not implement these functions
//   Really this macros is used to reduce the size of DLL of exported classes
//   And to suppress the warnings of gcc (see -Weffc++)
//
class ufsd_noncopyable
{
 protected:
    ufsd_noncopyable() {}
    ~ufsd_noncopyable() {}
 private:  // emphasize the following members are private
    ufsd_noncopyable( const ufsd_noncopyable& );
    const ufsd_noncopyable& operator=( const ufsd_noncopyable& );
};

//
// GCC used in MAC does not allow to use multiple base classes
//
#ifndef __APPLE__
  #define NONCOPYABLE ,public ufsd_noncopyable
#else
  #define NONCOPYABLE
#endif

// Helper template from Loki library
template <int v>
struct Int2Type
{
  enum { value = v };
};

#ifndef __PARAGON_UINT64_DEFINED
#define __PARAGON_UINT64_DEFINED

#if defined _MSC_VER
  typedef unsigned __int64    UINT64;
  typedef signed   __int64    INT64;
#elif defined __GNUC__
  typedef unsigned long long  UINT64;
  typedef long long           INT64;
#else
  #error "Does your compiler support build-in 64bit values"
#endif

#ifndef PU64
  //
  // Used to declare 64bit constants
  // F.e. #define DOS_DATE_BASE PU64(0x01A8E79FE1D58000)
  // is   #define DOS_DATE_BASE 0x01A8E79FE1D58000ULL  in GCC
  // and  #define DOS_DATE_BASE 0x01A8E79FE1D58000UI64 in VC
  //
  #define __PMERGE(a,b) a ## b
  #if defined _MSC_VER && _MSC_VER <= 1200
    #define PU64(x) __PMERGE(x, UI64)
    #define PI64(x) __PMERGE(x, I64 )
  #else
    #define PU64(x) __PMERGE(x, ULL)
    #define PI64(x) __PMERGE(x, LL)
  #endif

#endif // #ifndef PU64

#ifndef PLL
  //
  // Used to printf 64bit values
  // F.e. UINT64 a;
  // printf( "%"PLL"x\n", a );
  // is   printf( "%llx\n", a ); in GCC
  // and  printf( "%I64x\n", a ); in VC
  //
  #if defined _MSC_VER
    #define PLL "I64"
  #else
    #define PLL "ll"
  #endif
#endif // #ifndef PLL

#ifndef PZZ
  //
  // Used to printf size_t values
  // F.e. size_t a;
  // printf( "%"PZZ"x\n", a );
  // is   printf( "%Zx\n", a ); in GCC
  // and  printf( "%Ix\n", a ); in VC
  //
  #if defined _MSC_VER
    // MSDN: printf supports 'I', but really it does not
    #if _MSC_VER > 1200
      #define PZZ "I"
    #else
      #define PZZ ""
    #endif
  #elif defined __DJGPP__
    // DJGPP doesnt support 'Z'
    #define PZZ "l"
  #elif defined __QNX__
    // QNX doesnt support 'Z'
    #define PZZ ""
  #elif defined __APPLE__
    //C99 standard
    #define PZZ "z"
  #else
    // GCC supports 'Z'
    #define PZZ "Z"
  #endif
#endif

#ifndef MINUS_ONE_64
  #define MINUS_ONE_64  PU64(0xffffffffffffffff)
#endif


///////////////////////////////////////////////////////////
// UINT64 MERGE_UINT64
//
// Useful function to merge two 32 bit values into 64 bit value
///////////////////////////////////////////////////////////
static inline
UINT64 MERGE_UINT64( IN unsigned int low, IN unsigned int high ){
  return (((UINT64)high) << 32) | low;
}

#undef LODWORD
#undef HIDWORD
///////////////////////////////////////////////////////////
// LODWORD
//
// Simple function to access to low 32 bit of 64 bit value
// NOTE: this function returns r-value
///////////////////////////////////////////////////////////
static inline
unsigned int LODWORD( IN const UINT64& x ){
  return (unsigned int)(x);
}


///////////////////////////////////////////////////////////
// HIDWORD
//
// Simple function to access to high 32 bit of 64 bit value
// NOTE: this function returns r-value
///////////////////////////////////////////////////////////
static inline
unsigned int HIDWORD( IN const UINT64& x ){
  return (unsigned int)(x>>32);
}

#endif // __PARAGON_UINT64_DEFINED

#ifndef GUID_DEFINED
#define GUID_DEFINED
typedef struct _GUID {    // size is 16
    unsigned int   Data1;
    unsigned short Data2;
    unsigned short Data3;
    unsigned char  Data4[8];
} GUID;
#endif

#ifndef likely

  #if __GNUC__ >= 3 || (__GNUC__ == 2 && __GNUC_MINOR__ >= 96)
    // The __builtin_expect is a method that gcc (versions >= 2.96) offer
    // for programmers to indicate branch prediction information to the compiler.
    // The return value of __builtin_expect is the first argument
    // (which could only be an integer) passed to it.
    #define likely(x)       __builtin_expect((x),1)
    #define unlikely(x)     __builtin_expect((x),0)

  #else

    #define likely(x) x
    #define unlikely(x) x

  #endif // #if __GNUC__ >= 3 || (__GNUC__ == 2 && __GNUC_MINOR__ >= 96)

#endif // #ifndef likely

namespace UFSD{
  // Returns non zero if v is a power of 2 and not zero
  static inline bool IsPowerOf2( size_t v ){
    return 0 != v && 0 == (v & (v-1) );
  }
}

#ifndef PAGE_SIZE
  #define PAGE_SIZE   4096
  #define PAGE_SHIFT  12
#endif

#if defined UFSD_DRIVER_LINUX || defined UFSD_DRIVER_OSX
  #define UFSD_DRIVER_PURE_VIRTUAL = 0
#else
  #define UFSD_DRIVER_PURE_VIRTUAL
#endif


#define U_QUOTE1(name) #name
#define U_QUOTE2(name) U_QUOTE1(name)


#ifndef ARGUMENT_PRESENT
  #define ARGUMENT_PRESENT(p) (NULL != (p))
#endif

#endif //__UFSD_TYPES_H
