/****************************************************
 *  File UFSD/include/ufsd.h
 *  Copyright (C) 2000 Paragon Software,
 *
 *  UFSD Core is defined here.
 *  Here is CUFSD class definition for UFSD.
 *  Also here is the other stuff for UFSD.

  Since 2001 this file is supported by Alexander Mamaev

 *
 ****************************************************/
#ifndef __UFSD_UFSD_H
#define __UFSD_UFSD_H

namespace UFSD {

extern "C" unsigned
UFSD_API
__stdcall
InitUfsdTrace(
    IN void (__cdecl *Trace )( UINT64 context, const char*, va_list ),
    IN void (__cdecl *Error )( UINT64 context, int Err, const char* FileName, int Line )
    );

class CDir;
class CFile;
class CFSObject;
class CEntryNumerator;
struct IMessage;

#ifndef SLASH_CHAR
  #define SLASH_CHAR  '/'
  #define SLASH_ASCII "/"
#endif


////////////////////////////////
// File open flags
////////////////////////////////
#define OPEN_CREATE 0x00000001  // Create file if it doesn't exist
#define OPEN_TRUNC  0x00000002  // If file exists, truncate it to zero
#define OPEN_NEW    0x00000004  // File must not exist


//
// Additional options for CUFSD::FormatFs
//
union OptFormat{

  struct{
      unsigned int    HiddenSectors;      // Number of hidden sectors for bootable volume ( default 0 )
      unsigned short  wSectorsPerTrack;   // Sectors per track ( default 63 )
      unsigned short  wHeadNumber;        // Number of heads ( default 255 )
      unsigned char   MediaByte;          // Default 0xF8 - harddisk
      unsigned char   bEnableCompression; // Mark root directory as compressed ( default false )
      unsigned char   BcFs;               // BCFS signature
      unsigned char   Res;
      int             Version;            // See below
      unsigned int    BytesPerRecord;     // Desired size of MFT record ( 0 means default )

      //
      // Version
      //  610         - 3.1 Windows 7/8
      //   61         - 3.1 Windows 7/8 without /$Extend/$RmMetadata
      //  600         - 3.1 Vista
      //   60         - 3.1 Vista without /$Extend/$RmMetadata
      //   other      - 3.1 2K/XP
      //
  } Ntfs;

  struct{
      unsigned int    HiddenSectors;      // Number of hidden sectors for bootable volume ( default 0 )
      unsigned short  wSectorsPerTrack;   // Sectors per track ( default 63 )
      unsigned short  wHeadNumber;        // Number of heads ( default 255 )
      unsigned char   MediaByte;          // Default 0xF8 - harddisk
      int             Version;            // See below

      //
      // Version
      //   61         - Windows 7/8
      //   other      - 2K
      //
  } Fat;

  struct{
      UINT64          HiddenSectors;      // Number of hidden sectors for bootable volume ( default 0 )
      unsigned int    AllocationUnit;     // In bytes (default 0)
      unsigned char   TexFat;             // Format TexFAT
      unsigned char   BcFs;               // BCFAT signature
  } exFat;

  struct{
      unsigned char Guid[16];
  } Refs;

  struct{
      unsigned int    cNodeSize;          // Catalog node size
      unsigned int    eNodeSize;          // Extents node size
      //
      // j, c
      // 0, 0 - Mac OS Extended
      // 0, 1 - Mac OS Extended (Case-sensitive)
      // 1, 0 - Mac OS Extended (Journaled)
      // 1, 1 - Mac OS Extended (Journaled,Case-sensitive)
      //
      bool Journal;                 // default false
      bool CaseSensitive;           // default false
  } Hfs;

  struct{
      unsigned int InodeSize;       //0-def(128), min value 128 bytes, max block size, power of 2
      unsigned int InodeRatio;      //one Inodes per slice of volume, min 1024(block size) max 65536 * 1024
      unsigned int ReservedRatio;   //percent of blocks reserved for superuser, def 5%, max 50%
      bool         Journal;         //0-ext2 without journal
      bool         LazyFormat;      //0-full format, 1-not format some structures
      bool         RootWriteAccess; //0-umask 022(default) else umask 0
      //ext4
      bool         Ext4;            //1-with extents and other features
      unsigned int FlexBGsize;      //0-standart, 2...pow(2)-number of groups that will be packet together, 16-default, 1024-maximum
  } Ext;

  struct{
      unsigned int Version;         //0-new version(default), 0xffffffff-old version
  } LSwap;
};

//C_ASSERT( FIELD_OFFSET(OptFormat,Ntfs.HiddenSectors)    == FIELD_OFFSET(OptFormat,Fat.HiddenSectors) );
//C_ASSERT( FIELD_OFFSET(OptFormat,Ntfs.wSectorsPerTrack) == FIELD_OFFSET(OptFormat,Fat.wSectorsPerTrack) );
//C_ASSERT( FIELD_OFFSET(OptFormat,Ntfs.wHeadNumber)      == FIELD_OFFSET(OptFormat,Fat.wHeadNumber) );
//C_ASSERT( FIELD_OFFSET(OptFormat,Ntfs.MediaByte)        == FIELD_OFFSET(OptFormat,Fat.MediaByte) );


struct OptCheck{
  size_t  Flags;          // See CHK_OPTIONS_XXX
  size_t  MaxMem;         // Amount of memory to use guard. 0 - no limits
  bool    bFix;           // true to fix inconsistencies
  bool    bCheckForBad;   // true to check for bad blocks
};


#ifndef UFSD_DRIVER_LINUX

////////////////////////////////////////////////////////////////
// CUFSD Class
// Kernel (Main ) class for Universal File System Driver
////////////////////////////////////////////////////////////////
class UFSD_API CUFSD : public CMemBased<CUFSD>
{
protected:
    list_head         m_OpenFiles;
    list_head         m_OpenSearches;
    CFileSystem*      m_FileSystem;     // Pointer to File System
    CStrings*         m_Strings;        // Pointer to Strings Manager

    //Parse Path function checks if the path is valid
    //and opens or closes all subdirectories of the path
    //  type        - path string type UNICODE or ASCII
    //  path        - path
    //  close       - defines if all directories should be closed
    //  tail_skip   - interpret last Item in path
    //  dir         - Pointer to the result directory
    //Returns:
    //  0 if OK
    //  error code otherwise
    unsigned ParsePath(
        IN  STRType     type,
        IN  const void* path,
        OUT CDir *&     dir,
        IN  bool        get_last_item,
        OUT void *      buffer,
        IN  size_t      BytesPerBuffer,
        OUT size_t*     BufferLen
        );

    //Runs to the root directory and increases
    //Reference count on the way
    //  dir  -  pointer to start directory
    //  root - pointer to the root directory
    //Returns: error code
    unsigned OpenDirs(
        IN CDir * dir,
        IN CDir * root
        );

    //Flush certain dir
    //  dir - pointer to dir
    //Returns:
    //  Error
    static unsigned __stdcall FlushDir(
        IN CDir * dir
        );

    // Helper function for GetCwd and GetAltCwd
    // See comments for these functions
    unsigned GetCwd2(
        IN STRType  type,
        IN void *   dirname,
        IN size_t   size,
        OUT size_t* name_len,
        IN bool     bAltName
        );

    // close all searches before remove or rename directory
    unsigned CloseSearches(
        IN CDir * dir
        );

    // Helper function to remove directories resursivle
    unsigned  RmDirHlp(
        IN CDir*        Parent,
        IN STRType      Type,
        IN const void*  DirName,
        IN size_t       Name_len
        );

public:

    // Mm - Pointer to memory management object
    CUFSD( CMemoryManager * Mm )
      : CMemBased<CUFSD>( Mm )
      , m_FileSystem(NULL)
      , m_Strings(NULL)
    {
      m_OpenFiles.init();
      m_OpenSearches.init();
    }

    ~CUFSD() { Dtor(); } // non virtual!

    void Dtor(); // Used to cheat gcc

    //Initializes Universal File System Driver for specified File System
    //  Ss      - pointer to Strings management object
    //  Rw      - pointer to Read Write Block  object
    //  fs      - Mask of possible file systems
    //  Tt      - Pointer to time management object
    //  real_fs - Really initiated file system
    //            (This parameter is filled only for successful
    //             initiated file system.
    //             Useful for masked fs )
    //  fs_option - parameter passed to filesystem
    //              It is FS dependent (!!)
    //Returns:
    //  0 - if success;desc non zero;
    //  error code otherwise.
    unsigned Init(
        IN  CStrings* Ss,
        IN  CRWBlock* Rw,
        IN  size_t    fs,
        IN  CTime *   Tt,
        OUT size_t*   real_fs   = NULL, // OPTIONAL
        IN  size_t    fs_option = 0,
        OUT size_t*   fs_flags  = NULL  // OPTIONAL, See UFSD_FLAGS_XXX
        );

    //Initializes Universal File System Driver for specified File System
    //  Ss      - pointer to Strings management object
    //  Fs      - pointer to CFileSystem object
    //Returns:
    //  0 - if success;desc non zero;
    //  error code otherwise.
    unsigned Init(
        IN  CStrings*    Ss,
        IN  CFileSystem* Fs
        )
    {
      m_Strings    = Ss;
      m_FileSystem = Fs;
      return ERR_NOERROR;
    }

    // Helper function to check file descriptor
    bool DescValid(
        IN void * desc
        );

    // Helper function to check search handle
    bool HandleValid(
        IN void * handle
        );

    // Returns string manager used by this UFSD
    CStrings* GetStringManager() const{ return m_Strings; }

    // Returns current file system
    CFileSystem* GetFileSystem() const{ return m_FileSystem; }

    // Sets current directory
    bool SetCurrentDir(CDir * dir);

    //Open file with name specified
    //  desc - structure describing current state of opened file;
    //  type  - name type UNICODE/ASCII;
    //  fname - file name;
    //  mode  - File open mode
    //Returns:
    //  0 - if success;desc non zero;
    //  error code otherwise.
    unsigned Open(
        OUT void*&      desc,
        IN STRType      type,
        IN const void*  fname,
        IN size_t       flags   // 0,OPEN_CREATE,OPEN_TRUNC,OPEN_NEW
        );

    //Move to specified position
    //desc - structure describing current state of opened file;
    //  offset - move offset
    //  origin - code of start position
    //Returns:
    //  0             - if success
    //  error code    - otherwise
    unsigned Seek(
        IN void*          desc,
        IN const UINT64&  offset,
        IN SEEKOrig       orig
        );

    //Retriev current file position
    //Returns:
    //  0             - if success
    //  error code    - otherwise
    unsigned Tell(
        IN void *     desc,
        OUT UINT64&   offset
        );

    //Set end of file into current position
    //Returns:
    //  0             - if success
    //  error code    - otherwise
    unsigned SetEOF(
        IN void * desc
        );

    //Tests for end-of-file.
    //Returns:
    //  1             - end of file
    //  0             - if not
    unsigned FEOF(
        IN void * desc
        );

    //Close file
    //Returns:
    //  0             - if success
    //  error code    - otherwise
    unsigned Close(
        IN void * desc
        );

    //Read file
    //  buffer  - buffer for data read
    //  size    - block size in bytes
    //  bytes_read - number of bytes read
    //Returns:
    //  0             - if success
    //  error code    - otherwise
    unsigned Read(
        OUT size_t& bytes_read,
        IN void *   buffer,
        IN size_t   size,
        IN void *   desc
        );

    //Write file
    //  buffer  - buffer of data to be written
    //  size    - block size in bytes
    //  bytes_read - number of bytes written
    //Returns:
    //  0             - if success
    //  error code    - otherwise
    unsigned Write(
        OUT size_t&     written,
        IN const void*  buffer,
        IN size_t       size,
        IN void *       desc
        );

    //Delete file with specified name
    // type - name type - UNICODE or ASCII
    //Returns:
    //  0             - if success
    //  error code    - otherwise
    unsigned Unlink(
        IN STRType      type,
        IN const void*  name
        );

    //Move/rename file/or directory named oldname to newname
    //  type - name type - UNICODE or ASCII
    //Returns:
    //  0             - if success
    //  error code    - otherwise
    unsigned Rename(
        IN STRType      type,
        IN const void*  oldname,
        IN const void*  newname
        );

    //Create directory dirname
    //  type - name type - UNICODE or ASCII
    //Returns:
    //  0             - if success
    //  error code    - otherwise
    unsigned MkDir(
        IN STRType      type,
        IN const void*  dirname,
        OUT CDir**      dir = NULL
        );

    //Remove directory dirname
    //  type - name type - UNICODE or ASCII
    //Returns:
    //  0             - if success
    //  error code    - otherwise
    unsigned RmDir(
        IN STRType      type,
        IN const void*  dirname
        );

    //Change current directory
    //  type - name type - UNICODE or ASCII
    //Returns:
    //  0             - if success
    //  error code    - otherwise
    unsigned ChDir(
        IN STRType      type,
        IN const void*  dirname
        );

    //Get current working directory
    //  type    - type of directory name (ASCII or UNICODE)
    //  dirname - directory name
    //  size    - maximum number of bytes to be retrieved
    //  name_len- name length in symbols
    //Returns:
    //  0 - success
    //  error code otherwise
    unsigned GetCwd(
        IN  STRType type,
        OUT void *  dirname,
        IN  size_t  size,
        OUT size_t* name_len = NULL
        );

    //Get current working directory in alternative names
    //  type    - type of directory name (ASCII or UNICODE)
    //  dirname - directory name
    //  size    - maximum number of bytes to be retrieved
    //  name_len- name length in symbols
    //Returns:
    //  0 - success
    //  error code otherwise
    unsigned GetAltCwd(
        IN STRType  type,
        OUT void *  dirname,
        IN size_t   size,
        OUT size_t* name_len = NULL
        );

    //Flush file
    //  desc - opened file to flush
    //Returns:
    //  0 - success
    //  error code otherwise
    unsigned Flush(
        IN void * desc
        );

    //Flushes all opened files and directories
    //Returns:
    //  0 - success
    //  error code otherwise
    unsigned FlushAll();

    //Retrieves file attributes for file desc,
    //  attr        - common attributes,
    //                if NULL then it is not retrieved
    //  fs_attrs    - file system specific attributes,
    //                if NULL then it is not retrieved
    //Returns:
    //  0 - success
    //  error code otherwise
    unsigned GetAttr(
        IN STRType        type,
        IN const void *   name,
        OUT unsigned int* attr,
        OUT unsigned int* fs_attrs = NULL
        );

    //Retrieves file creation, reference and modification times
    //  CrTime,       // local time of last write to file
    //  ReffTime,     // local time of last write to file
    //  ModiffTime,   // local time of last file modification
    // if NULL is passed as an argument then this time will not be retrieved.
    //Returns:
    //  0 - success
    //  error code otherwise
    unsigned GetTimes(
        IN STRType      type,
        IN const void*  name,
        OUT UINT64*     CrTime,
        OUT UINT64*     ReffTime,
        OUT UINT64*     ModiffTime
        );

    //Set file creation, reference and modification times
    //  CrTime,       // local time of last write to file
    //  ReffTime,     // local time of last write to file
    //  ModiffTime,   // local time of last file modification
    // if NULL is passed as an argument then this time will not be retrieved.
    //Returns:
    //  0 - success
    //  error code otherwise
    unsigned SetTimes(
        IN STRType        type,
        IN const void*    name,
        IN const UINT64*  CrTime,
        IN const UINT64*  ReffTime,
        IN const UINT64*  ModiffTime
        );

    // Sets file attributes for file desc,
    //  attr    -   attributes, if highest bit is set to one then
    //              file system specific attributes are set
    //Returns:
    //  0 - success
    //  error code otherwise
    unsigned SetAttr(
        IN STRType      type,
        IN const void*  name,
        IN unsigned int attr
        );

    //Find first file matching file mask
    //  type      - name type - UNICODE or ASCII
    //  filemask  - file search mask
    //  fi        - result structure with file information
    //  handle    - pointer to structure describing file
    //Returns:
    //  0             - if success
    //  error code    - otherwise
    unsigned FindFirst(
        IN STRType      type,
        IN const void*  filemask,
        OUT FileInfo&   fi,
        OUT void*&      handle,
        IN  size_t      options = 0 // See UFSD_ENUM_XXX (u_fsbase.h)
        );

    //Find next file after findfirst
    //  handle  - file found
    //  fi      - result structure with file information
    //Returns:
    //  0             - if found
    //  error code    - otherwise
    unsigned FindNext(
        IN void*      handle,
        OUT FileInfo& fi
        );

    //Close search
    //Returns:
    //  0             - if success
    //  error code    - otherwise
    unsigned FindClose(
        IN OUT void*&   handle
        );

    //Opens directory wilt name specified
    //  pointer to the opened dir
    //Returns:
    //  0             - if success
    //  error code    - otherwise
    unsigned OpenDir(
        IN STRType      type,
        IN const void*  dir_name,
        OUT void*&      dir
        );

    //Closes directory opened with opendir
    unsigned CloseDir(
        IN void* dir
        );

    // Get volume information.
    unsigned GetVolumeInfo(
        OUT UINT64&       FreeClusters,
        OUT UINT64&       TotalClusters,
        OUT size_t&       BytesPerCluster,
        OUT unsigned char VolumeSerial[16]        = NULL,       // OPTIONAL
        IN  STRType       VolumeNameType          = StrUnknown, // OPTIONAL
        OUT void*         VolumeNameBuffer        = NULL,       // OPTIONAL
        IN  size_t        VolumeNameBufferLength  = 0,          // OPTIONAL, in (wide)chars
        OUT VolumeState*  VolState                = NULL,       // OPTIONAL
        OUT size_t*       BytesPerSector          = NULL        // OPTIONAL
        );

    // A short form of 'GetVolumeInfo'
    unsigned  GetVolumeState(
        OUT VolumeState* VolState
        );

    // Set volume information.
    unsigned SetVolumeInfo(
        IN const unsigned char VolumeSerial[16],  // OPTIONAL
        IN STRType      VolumeNameType,           // OPTIONAL
        IN const void*  VolumeNameBuffer,         // OPTIONAL
        IN VolumeState  NewState = NoChange       // OPTIONAL
        );

    // This function gets file information by its' path
    unsigned GetFSObjInfo(
        IN STRType      type,
        IN const void*  Path,
        OUT FileInfo&   fi
        );

    // This function gets FileInfo by handle to CFSObject
    unsigned GetFSObjInfo(
        IN void*      FSObj,
        OUT FileInfo& fi
        );

    // Function a-la DeviceIoControl
    unsigned FsIoControl(
        IN  size_t      FsIoControlCode,
        IN  const void* InBuffer       = NULL,  // OPTIONAL
        IN  size_t      nInBufferSize  = 0,     // OPTIONAL
        OUT void*       OutBuffer      = NULL,  // OPTIONAL
        IN  size_t      nOutBufferSize = 0,     // OPTIONAL
        OUT size_t*     BytesReturned  = NULL   // OPTIONAL
        );
};


///////////////////////////////////////////////////////
//                     GetFsFormatParameters
//
//
// Used to get format parameters
//
// Fs                   - FileSystem, supported by UFSD
// FsSizeInSectors      - Volume size in sectors
// Params               - Format possible parameters
//
// Return value:
// true if parameters can be used to format Fs using UFSD
///////////////////////////////////////////////////////
bool UFSD_API __stdcall
GetFsFormatParameters(
    IN size_t           Fs,
    IN const UINT64&    VolumeSizeInSectors,
    IN unsigned short   BytesPerSector,
    IN const OptFormat* Opt,
    IN IMessage*        IMsg,
    OUT t_FormatParams* Params
    );

///////////////////////////////////////////////////////
//                           FormatFs
//
// Used to format fs
//
// Fs                   - FileSystem, supported by UFSD
// Mm,Ss,Rw,Tt          - services used to format volume
// Msg                  - service to output messages (can be NULL)
// bQuickFormat         - If not zero UFSD will verify volume
// BytesPerSector       - Desired size of sector in bytes. If 0 then RWBlock is used to get BytesPerSector
// BytesPerCluster      - Cluster size in bytes
// VolumeName           - Volume name (label)
// OptFormat            - Additional options
//
// Note: Size of volume is retrieved from CRWBlock
// Return UFSD error code
///////////////////////////////////////////////////////
unsigned  UFSD_API __stdcall
FormatFs(
    IN size_t           Fs,
    IN CMemoryManager*  Mm,
    IN CStrings *       Ss,
    IN CRWBlock *       Rw,
    IN CTime *          Tt,
    IN IMessage*        Msg,
    IN bool             bQuickFormat,
    IN unsigned short   BytesPerSector,
    IN unsigned int     BytesPerCluster,
    IN STRType          VolumeNameType,
    IN const void*      szVolumeName,
    IN const OptFormat* Opt = NULL,
    IN const char*      Hint = NULL
    );

///////////////////////////////////////////////////////
//            CheckFs  (static function)
//
// Used to check fs
//
// Fs                   - FileSystem, supported by UFSD
// Mm,Ss,Rw,Tt          - services used to check volume
// Msg                  - service to output messages (can not be NULL)
// Opt                  - additional options
// Hint                 - Name of volume. Used to show report
//
// Note: Size of volume is retrieved from CRWBlock
// Return UFSD error code
///////////////////////////////////////////////////////
unsigned UFSD_API __stdcall
CheckFs(
    IN size_t          Fs,
    IN CMemoryManager* Mm,
    IN CStrings *      Ss,
    IN CRWBlock *      Rw,
    IN CTime *         Tt,
    IN IMessage*       Msg,
    IN const OptCheck* Opt,
    IN const char*     Hint,
    OUT CHKDSK_INFO&   ChkInfo
    );


///////////////////////////////////////////////////////
//            DefragFs  (static function)
//
// Used to defragment fs
//
// Fs                   - FileSystem, supported by UFSD
// Mm,Ss,Rw,Tt          - services used to defrag volume
// Msg                  - service to output messages (can not be NULL)
// FsDefrag             - Input parameters of defragment task
// Report               - Output struct for report
// BytesPerReport       - The size of report structure
// nBytesReturned       - Filled bytes in report,
// Stat                 - FileSystem statistic during operation
// Return UFSD error code
///////////////////////////////////////////////////////
unsigned UFSD_API __stdcall
DefragFs(
    IN  size_t                 Fs,
    IN  CMemoryManager*        Mm,
    IN  CRWBlock*              Rw,
    IN  CStrings*              Ss,
    IN  CTime*                 Tt,
    IN  IMessage*              Msg,
    IN  const UFSD_FS_DEFRAG*  FsDefrag,
    OUT UFSD_DEFRAG_REPORT*    Report,
    IN  size_t                 BytesPerReport,
    OUT size_t*                nBytesReturned,
    OUT UFSD_FILESYSTEM_STATISTICS* Stat = NULL
    );

///////////////////////////////////////////////////////
//            GetVolumeMetaBitmap (static function)
//
// Used to dump filesystem
//
// Fs                   - mask of possible filesystems, supported by UFSD
// Mm,Rw                - services used to try volume
// Msg                  - service to output messages (can be NULL)
// Bitmap               - bitmap of size '(ClustersPerVolume+7)/8' bytes
// ClustersPerVolume    - Total clusters in bitmap
// Log2OfCluster        - Log( Bytes per cluster )
// Errors               - Count of errors while operation
//
// Return UFSD error code
///////////////////////////////////////////////////////
unsigned UFSD_API __stdcall
GetVolumeMetaBitmap(
    IN  size_t            Fs,
    IN  CMemoryManager*   Mm,
    IN  CRWBlock*         Rw,
    IN  IMessage*         Msg, // OPTIONAL
    OUT unsigned char**   Bitmap,
    OUT unsigned int*     ClustersPerVolume,
    OUT unsigned int*     Log2OfCluster,
    OUT unsigned int*     Errors,
    OUT size_t*           real_fs = NULL
    );


///////////////////////////////////////////////////////////
// DoFsDump
//
// Do fsdump image from rw block
///////////////////////////////////////////////////////////
unsigned UFSD_API __stdcall
DoFsDump(
    IN CMemoryManager*  Mm,
    IN CRWBlock*        Rw,
    IN CRWBlock*        FileRw,
    IN IMessage*        Msg,
    IN const char*      szDevice,
    IN int (__cdecl * Sprintf)(char *, const char *, ...)
    );


///////////////////////////////////////////////////////////
// DoFsRestore
//
// Restores fsdump image into rw block
///////////////////////////////////////////////////////////
unsigned UFSD_API __stdcall
DoFsRestore(
    IN CMemoryManager*  Mm,
    IN CRWBlock*        Rw,
    IN CRWBlock*        FileRw,
    IN IMessage*        Msg,
    IN const char*      szDevice,
    IN int (__cdecl * Sscanf)(const char *, const char *, ...)
    );


///////////////////////////////////////////////////////////
// GetFullName
//
// Returns the length of filled buffer
// If FullName is NULL then the returned value is the length
// (not including last zero!) required for full name
// chSize         - Maximum number of symbols in buffer (including last zero)
///////////////////////////////////////////////////////////
size_t UFSD_API __stdcall
GetFullName(
    IN  CFSObject*  Fso,
    IN  CStrings*   Str,
    OUT char*       FullName,
    IN  size_t      chSize
    );

///////////////////////////////////////////////////////
//                     GetFsMinSize
//
//
// Estimate minimal volume size
//
// Fs                     - FileSystem, supported by UFSD
// BytesPerSector
// BytesPerCluster
// DataCluster            - Size of used data in clusters on volume
// NumFiles               - Number of files on volume
// NumSubdirs             - Number of subdirs on volume
// MinVolumeSizeInSectors - returns minimal volume size for these params
//
// Return value:
///////////////////////////////////////////////////////
unsigned UFSD_API __stdcall
GetFsMinSize(
    IN size_t           Fs,
    IN unsigned short   BytesPerSector,
    IN unsigned int     BytesPerCluster,
    IN unsigned int     DataClusters,
    IN size_t           NumFiles,
    IN size_t           NumSubdirs,
    OUT UINT64*         MinVolumeSizeInSectors
    );

// CFile*
// Handle2File(
//     IN void* h  // handle returned by CUFSD::Open
//     );
#define Handle2File(h) ((UFSD::CFile*)*(void**)(Add2Ptr(h,sizeof(UINT64))))

// CDir*
// Handle2Dir(
//     IN void* h  // handle returned by CUFSD::OpenDir
//     );
#define Handle2Dir(h) ((UFSD::CDir*)(h))


// CEntryNumerator*
// Handle2Enum(
//     IN void* h // handle returned by CUFSD::FindFirst
//     );
#define Handle2Enum(h) ((UFSD::CEntryNumerator*)*(void**)(h))

#endif // #ifndef UFSD_DRIVER_LINUX

} // namespace UFSD {

#endif //__UFSD_UFSD_H

