////////////////////////////////////////////////////////////////
// June of 2005. (C) Paragon Software
// If this code works, it was written by Alexander Mamaev
// If not, I don't know who wrote it.
//
// This file contains the code that:
// - parses command line
// - calls UFSD::CheckFs( FS_NTFS, ... )
//
////////////////////////////////////////////////////////////////
//#define FSDUMP_FILE_NAME "Z:\\bugs\\60327_ntfs_journal\\dump"

#include <stdio.h>
#include <stdlib.h> // exit
#include <string.h> // memset
#include <assert.h>

//
// Include main ufsd header
//
#include "ufsd.h"

//
// All UFSD classes and functions are placed in UFSD namespace
//
using namespace UFSD;

#ifndef NLS_CAT_NAME
# define NLS_CAT_NAME "chkntfs"
#endif


//
// This file contains declarations of useful UFSD-client functions
//
#include "funcs.h"
#include "gettext.h"

#ifdef _WIN32
//---------------------------------------------------------
// To compile windows utility
//---------------------------------------------------------

#else
//---------------------------------------------------------
// To compile linux utility
//---------------------------------------------------------

#include <unistd.h>
#include <fcntl.h>
#include <locale.h>

//
// When this file is compiled as a Linux utility
// then some useful functions (GetAbsPath,ParsePath) that are shared for
// many projects are placed in file misc.cpp
//
# include "misc.h"

//---------------------------------------------------------
// End of Linux specific
//---------------------------------------------------------
#endif // _WIN32


///////////////////////////////////////////////////////////
// OnAbout
//
//
///////////////////////////////////////////////////////////
static void
OnAbout(void)
{
  fprintf( stdout, _("Check NTFS volume for error\n") );
}


///////////////////////////////////////////////////////////
// OnUsage
//
// print a list of the parameters to the program
///////////////////////////////////////////////////////////
static void
OnUsage()
{
  OnAbout();
  fprintf( stdout, _(
"Usage " NLS_CAT_NAME " device [-f]\n"
"  -f              fix errors on the disk.\n"
"  -a              auto check. Full check if volume is dirty\n"
"  -m:size         memory limit\n"
"  -h              display this help\n"
"  --short         check the minimum\n"
"  --safe          do not try to fix errors;\n"
"                  only clear 'dirty' flag if there are no errors\n"
"  --help          display this help\n"
"  --showminors    show minor errors\n"
"  --no-orphans    do not restore real orphan files\n"
"  --trace         turn on UFSD trace\n"
"  --verbose       explain what is being done\n"
"  --nopercents    do not print percents\n"
"  --version       show version and exit\n") );
#ifdef _WIN32
  fprintf( stdout,
_("E.g. " NLS_CAT_NAME " g:\n"
  "E.g. " NLS_CAT_NAME " \\\\?\\Volume{d7c5ce70-3220-11db-b0fb-005056c00008}\n") );
#else
  fprintf( stdout, _("E.g. " NLS_CAT_NAME " /dev/hdb1\n") );
#endif
}

//
// Chkntfs options structure.
// It will be filled from user' parameter string.
//
struct chk_options : public OptCheck {

  unsigned
     bAuto:1,
     nopercents:1,
     verbose:1;
};


///////////////////////////////////////////////////////////
// parse_options
//
// Full trusted function
// Returns false if error
///////////////////////////////////////////////////////////
static bool
parse_options(
    IN  int           argc,
    IN  const char*   argv[],
    OUT chk_options*  opts,
    OUT const char**  szDevice
    )
{
  int i;
  const char* cmd = NULL;

  *szDevice  = NULL;

  if ( argc <= 1 )
    return false; // Force to call OnUsage

  for ( i = 1; i < argc; i++ )
  {
    cmd = argv[i];
    //
    // Make sure arg is in proper format.
    //
    if ( '-' != cmd[0] )
    {
      // The first "non options argument" is a device name
      if ( NULL != *szDevice )
        goto Exit;
      *szDevice = cmd;
    }
    //
    // Which arg is it?
    //
    else switch ( cmd[1] )
    {
    case 'f':
      opts->bFix = true;
      break;

    case 'a':
      opts->bAuto = true;
      break;

    case 'h':
      return false; // Force to call OnUsage

    case 'm':
      if ( ':' != cmd[2] || 0 == (opts->MaxMem = ParseSize( cmd + 3 )) )
        goto Exit;
      break;

    case '-':
      if ( 0 == strcmp( "showminors", cmd+2 ) )
      {
        SetFlag( opts->Flags, CHK_OPTIONS_SHOWMINOR );
        break;
      }
      else if ( 0 == strcmp( "no-orphans", cmd+2 ) )
      {
        SetFlag( opts->Flags, CHK_OPTIONS_NOORPHANS );
        break;
      }
      else if ( 0 == strcmp( "version", cmd+2 ) )
      {
        fprintf( stdout, VER_STRING );
        exit(1);
      }
      else if ( 0 == strcmp( "trace", cmd+2 ) )
      {
        EnableTrace();
        break;
      }
      else if ( 0 == strcmp( "verbose", cmd+2 ) )
      {
        opts->verbose = 1;
        break;
      }
      else if ( 0 == strcmp( "nopercents", cmd + 2 ) )
      {
        opts->nopercents = 1;
        break;
      }
      else if ( 0 == strcmp( "help", cmd + 2 ) )
      {
        // Force to call OnUsage
        return false;
      }
      else if ( 0 == strcmp( "safe", cmd + 2 ) )
      {
        SetFlag( opts->Flags, CHK_OPTIONS_SAFE );
        break;
      }
      else if ( 0 == strcmp( "short", cmd + 2 ) )
      {
        SetFlag( opts->Flags, CHK_OPTIONS_SHORT );
        break;
      }

      //
      // !!!No break here!!!
      //

    default:
      goto Exit;    // double break
    }
  }

  return true;

Exit:
  fprintf( stderr, _("Error in argument: \"%s\"\n"), cmd );
  return false; // Force to call OnUsage
}


///////////////////////////////////////////////////////////
// main
//
//
///////////////////////////////////////////////////////////
int __cdecl
main(
    IN int argc,
    IN const char* argv[]
    )
{
  chk_options opts;
  unsigned Status       = ERR_BADPARAMS;
  const char* szDevice  = NULL;
  CRWBlock* Rw          = NULL;
  int Ret = 0;

#ifndef _WIN32
  setlocale(LC_MESSAGES, "");
  setlocale(LC_CTYPE, "");
  bindtextdomain(NLS_CAT_NAME, LOCALEDIR);
  textdomain(NLS_CAT_NAME);
#endif

  if ( argc < 1 )
    return -1;  // Impossible

  const char* szErr = CheckUFSDEndian();
  if ( NULL != szErr )
  {
    fprintf( stderr, "%s\n", szErr );
    return -1;
  }

  // Set default options
  memset( &opts, 0, sizeof(opts) );

  //
  // Parse command line
  // And exit if there are errors
  //
  if ( !parse_options( argc, argv, &opts, &szDevice ) )
  {
    OnUsage();
    return -2;
  }

#ifdef FSDUMP_FILE_NAME
  Rw      = GetRwBlock( FSDUMP_FILE_NAME );
  Status  = ERR_NOERROR;
#else
  if ( NULL == szDevice )
  {
    fprintf( stderr, _("Device is not supplied\n") );
    return -3;
  }

  if ( FlagOn( opts.Flags, CHK_OPTIONS_SAFE ) && !opts.bFix )
  {
    fprintf( stderr, _("\"--safe\" option requires \"-f\" to work properly\n") );
    return -4;
  }

#ifdef _WIN32

  char szVolume[64];
  if ( ':' == szDevice[1] )
    _snprintf( szVolume, ARRSIZE(szVolume), "\\\\.\\%s", szDevice );
  else
    _snprintf( szVolume, ARRSIZE(szVolume), szDevice );
  szVolume[ARRSIZE(szVolume)-1] = 0;
  szDevice = szVolume;

#else

  //
  // Check for mounted devices
  //
  const char* szMount = GetMount( szDevice, opts.verbose );
  if ( NULL != szMount )
  {
    fprintf( stderr, _("\nThe device \"%s\" is mounted as \"%s\"\nImpossible to check\n\n"), szDevice, szMount );
    return -4;
  }

#endif

  //
  // Init libufsd.so.
  // Without such initialization library is not able to show messages and errors
  //
  InitUfsdLibrary();

  //
  // Allocate BlockDevice
  //
  Status = UFSD_IOHandlerCreate( szDevice, !opts.bFix, &Rw, NULL, false, opts.verbose );
#endif

  if ( ERR_NOERROR != Status )
  {
    // Here Status is an 'errno'
    // Return it in negative form
    Ret = -(int)Status;
  }
  else
  {
    CMemoryManager* Mm  = UFSD_GetMemoryManager();
    CStrings* Ss        = UFSD_GetStringsService();
    CTime* Tt           = UFSD_GetTimeService();
    IMessage* IMsg      = UFSD_GetMessageService( stdout, NULL, opts.nopercents );
    bool SkipCheck      = false;
    size_t FsFlags      = 0;

    assert( NULL != Rw );

    if ( !opts.bFix )
      fprintf( stdout, _("WARNING!  f parameter not specified.\nRunning " NLS_CAT_NAME " in read-only mode.\n\n") );

    if ( opts.bAuto )
    {
      //
      // Assume NTFS is consistent. Check native journal and dirty flag
      //
      VolumeState VolState;
      CUFSD Ufsd( Mm );

      //
      // If '/f' and native journal is not empty
      // then Ufsd.Init() returns ERR_NEED_REPLAY
      //
      // If '/f' is not set and native journal is not empty
      // then FsFlags will contain bit UFSD_FLAGS_NEED_REPLAY_NATIVE
      //
      if ( UFSD_SUCCESS( Ufsd.Init( Ss, Rw, FS_NTFS, Tt, NULL, UFSD_OPTIONS_RECOGNIZE_FS, &FsFlags ) ) )
      {
        if ( FsFlags & UFSD_FLAGS_NEED_REPLAY_NATIVE )
        {
          SkipCheck = true;
          fprintf( stdout, _("NTFS quick check: native NTFS's journal needs to be replayed!\n") );
          Ret = 16;
        }
        else if ( UFSD_SUCCESS( Ufsd.GetVolumeState( &VolState ) )
          && Clean == VolState )
        {
          SkipCheck = true;
          fprintf( stdout, _("NTFS quick check: seems to be consistent\n") );
          Ret = 0;
        }
      }
    }

    if ( !SkipCheck )
    {
      //
      // Call UFSD code
      //
      CHKDSK_INFO ChkInfo = {0};

      Status = UFSD::CheckFs( FS_NTFS, Mm, Ss, Rw, Tt, IMsg, &opts, szDevice, ChkInfo );

      if ( ERR_NOERROR == Status )
      {
        // Report message was printed via message service'
        switch( ChkInfo.ExitStatus )
        {
        case UFSD_CHK_SUCCESS:
          Ret = 1; break;
        case UFSD_CHK_FIXED:
          Ret = 2; break;
        case UFSD_CHK_MINOR:
          Ret = 3; break;
        case UFSD_CHK_NOT_FIXED:
          Ret = opts.bFix? 4 : 5; break;
        case UFSD_CHK_FAILED:
          Ret = 6; break;
        case UFSD_CHK_BAD_BLOCKS:
          Ret = 15; break;
        default:
          Ret = 7;  // internal error
        }
      }
      else switch( Status )
      {
        case ERR_BADPARAMS:
          Ret = 8; break;
        case ERR_FSUNKNOWN:
          Ret = 9; break;
        case ERR_NOMEMORY:
          Ret = 10; break;
        case ERR_READFILE:
          Ret = 11; break;
        case ERR_WRITEFILE:
          Ret = 12; break;
        case ERR_CANCELLED:
          Ret = 13; break;
        case ERR_NOMEDIA:
          Ret = 17; break;
        default:
          Ret = 14; break;
      }
    }
  }

  //
  // Free all resources
  //
#ifndef FSDUMP_FILE_NAME
  UFSD_IOHandlerFree( Rw );
#endif

  //
  // Translate error code to string
  //
  if ( IS_UFSD_ERROR( Status ) )
    PrintUfsdStatusIfError( Status );

  if ( opts.verbose )
    fprintf( stdout, _(NLS_CAT_NAME " returns %d\n"), Ret );

  // Exit to host OS
  return Ret;
}
