////////////////////////////////////////////////////////////////
// October of 2004. (C) Paragon Software
// If this code works, it was written by Alexander Mamaev
// If not, I don't know who wrote it.
//
// This file contains helper functions that is used to debug UFSD library
////////////////////////////////////////////////////////////////
#include <stdio.h>
#include <stdarg.h>
#include <assert.h>
#include <locale.h>
#include <stdlib.h> // exit()
#include <unistd.h> // fsync

#include "gettext.h"
#include "u_errors.h"

#define IN
#define OUT

static FILE* flog = NULL;

///////////////////////////////////////////////////////////
// EnableTrace
//
// Enables trace from UFSD
///////////////////////////////////////////////////////////
void
EnableTrace( void )
{
  if ( NULL == flog )
    flog = fopen( "ufsd.log", "wt" );
}


///////////////////////////////////////////////////////////
// CloseTrace
//
//
///////////////////////////////////////////////////////////
void
CloseTrace( void )
{
  if ( NULL != flog )
  {
    fclose( flog );
    flog = NULL;
  }
}


///////////////////////////////////////////////////////////
// vUFSDTrace
//
// Traces the formatted messages into log file
///////////////////////////////////////////////////////////
void
vUFSDTrace(
    IN const char*format,
    IN va_list arglist
    )
{
  if ( NULL != flog )
  {
    vfprintf( flog, format, arglist );
    fflush( flog );
    fsync( fileno( flog ) );
  }
}


#ifdef UFSD_LIB_STATIC

///////////////////////////////////////////////////////////
// __cxa_pure_virtual
//
//
///////////////////////////////////////////////////////////
void
__cxa_pure_virtual(void)
{
//    CYG_FAIL("attempt to use a virtual function before object has been constructed");
    for ( ; ; )
    {
      ;
    }
}


///////////////////////////////////////////////////////////
// _UFSDTrace
//
//
///////////////////////////////////////////////////////////
void
_UFSDTrace(
    IN const char* format,
    IN ...
    )
{
  va_list arglist;
  va_start( arglist, format );
  vUFSDTrace( format, arglist );
  va_end( arglist );
}


///////////////////////////////////////////////////////////
// UFSDError
//
//
///////////////////////////////////////////////////////////
void
UFSDError(
    IN int iErr,
    IN const char* szFileName,
    IN int iLine
    )
{
  _UFSDTrace( "Ufsd error %x: %s, line %d\n", iErr, szFileName, iLine );
}

#else

//
// Extern function from libufsd.so
//
const char*
InitSo(
    IN void (*pTrace )( const char*, va_list)
    );

#endif

///////////////////////////////////////////////////////////
// InitUfsdLibrary
//
// Inits libufsd.so
///////////////////////////////////////////////////////////
void
InitUfsdLibrary()
{
#ifndef UFSD_LIB_STATIC
  // Pass pointer to vUFSDTrace function
  // UFSD library will use this function to show message and errors
  const char* szErr = InitSo( &vUFSDTrace );
  if ( NULL != szErr )
  {
    fprintf( stderr, "%s\n", szErr );
    exit(1);
  }
#endif

#ifndef NDEBUG
  //
  // Activate trace in debug version
  //
  fprintf( stdout, "WARNING: you use debug version. Keep in mind\n" );
  EnableTrace();
#endif
}


///////////////////////////////////////////////////////////
// GetUfsdErrorStr
//
// Translates UFSD error code to string.
///////////////////////////////////////////////////////////
const char*
GetUfsdErrorStr(
    IN unsigned nError
    )
{
  switch( nError )
  {
  case ERR_BADPARAMS:       return _("Bad parameters");
  case ERR_NOMEMMNGR:       return _("Memory manager is not valid");
  case ERR_BADRWBLOCK:      return _("Read/Write manager is not valid");
  case ERR_FSUNKNOWN:       return _("Unknown FileSystem");
  case ERR_NOMEMORY:        return _("Not enough memory");
  case ERR_BADUFSDFILE:     return _("Invalid file descriptor");
  case ERR_OPENROOT:        return _("Can't open root");
  case ERR_OPENFILE:        return _("Can't open file");
  case ERR_CREATEFILE:      return _("Can't create file");
  case ERR_READFILE:        return _("Can't read file");
  case ERR_WRITEFILE:       return _("Can't write file");
  case ERR_NOFILEEXISTS:    return _("File does not exist");
  case ERR_FILEEXISTS:      return _("Such file already exists");
  case ERR_NOROOTDIR:       return _("Invalid root directory");
  case ERR_NOUFSDINIT:      return _("UFSD is not initialized");
  case ERR_NOCURRENTDIR:    return _("Current directory is not set");
  case ERR_BADNAME:         return _("Ivalid file name");
  case ERR_BADNAME_LEN:     return _("File name length is invalid");
  case ERR_BADSTRTYPE:      return _("Invalid string type");
  case ERR_CLOSEUNABLE:     return _("Unable to close searches");
  case ERR_NODIRENTRY:      return _("The specified directory entry does not exist");
  case ERR_BADOBJECTTYPE:   return _("Invalid type of object");
  case ERR_CREATEDIR:       return _("Can't create directory");
  case ERR_NOFSINTEGRITY:   return _("File system is not consistent");
  case ERR_BADATTRIBUTES:   return _("Invalid file attributes");
  case ERR_WPROTECT:        return _("Volume is write protected");
  case ERR_DIRNOTEMPTY:     return _("Can't delete directory 'cause it is not empty");
  case ERR_FILENODATA:      return _("No such file");
  case ERR_NOSPC:           return _("No space left on device");
  case ERR_NOTIMPLEMENTED:  return _("This function is not implemented");
  case ERR_INSUFFICIENT_BUFFER: return _("The output buffer is too small to return any data");
  case ERR_MORE_DATA:       return _("the output buffer is too small to hold all of the data but can hold some entries");
  case ERR_BADBLOCK:        return _("Attempt to read/write bad block");
  case ERR_CANCELLED:       return _("Operation has been canceled");
  case ERR_NOMEDIA:         return _("Media is ejected");
  }

  return NULL;
}


///////////////////////////////////////////////////////////
// PrintUfsdStatusIfError
//
//
///////////////////////////////////////////////////////////
void
PrintUfsdStatusIfError(
    IN unsigned Status
    )
{
  if ( ERR_NOERROR != Status )
  {
    const char* szErr = GetUfsdErrorStr( Status );
    if ( NULL == szErr )
      fprintf( stderr, _("Error 0x%x\n"), Status );
    else
      fprintf( stderr, "%s\n", szErr );
  }
}
