////////////////////////////////////////////////////////////////
// August of 2006. (C) Paragon Software
// If this code works, it was written by Alexander Mamaev
// If not, I don't know who wrote it.
//
// This file contains the code that:
// - parses command line
// - Enumerates all directories and files
// - Count hard links
// - Prints final statistic
//
////////////////////////////////////////////////////////////////

#include <stdio.h>
#include <stdlib.h> // exit
#include <string.h> // memset
#include <assert.h>
#include <fcntl.h>
#include <ctype.h>  // isprint

//
// Include main ufsd header
//
#include "ufsd.h"

//
// All UFSD classes and functions are placed in UFSD namespace
//
using namespace UFSD;

#ifndef NLS_CAT_NAME
# define NLS_CAT_NAME "extents"
#endif

//
// This file contains declarations of useful UFSD-client functions
//
#include "funcs.h"
#include "gettext.h"

//---------------------------------------------------------
// To compile windows utility
#if defined _WIN32
//---------------------------------------------------------

#else
//---------------------------------------------------------
// To compile linux utility
//---------------------------------------------------------

#include <unistd.h>
#include <locale.h>

//
// When this file is compiled as a Linux utility
// then some useful functions (GetAbsPath,ParsePath) that are shared for
// many projects are placed in file misc.cpp
//
# include "misc.h"

//---------------------------------------------------------
// End of Linux specific
//---------------------------------------------------------
#endif // _WIN32

// "true" when [s,s+c) intersects with [l,l+w)
#define IS_IN_RANGE(s,c,l,w)                            \
  ( ( (c) > 0 && (w) > 0 )                              \
   && ( ( (l) <= (s) && (s) < ((l)+(w)) )                \
     || ( (s) <= (l) && ((s)+(c)) >= ((l)+(w)) )         \
     || ( (l) <  ((s)+(c)) && ((s)+(c)) < ((l)+(w)) ) ) )

#include "uenum.h"

///////////////////////////////////////////////////////////
// OnAbout
//
//
///////////////////////////////////////////////////////////
static void
OnAbout(void)
{
  fprintf( stdout, _("Traces files/folders extents\n") );
}


///////////////////////////////////////////////////////////
// OnUsage
//
// print a list of the parameters to the program
///////////////////////////////////////////////////////////
static void
OnUsage()
{
  OnAbout();
  fprintf( stdout, _(
"Usage " NLS_CAT_NAME " device\n"
"  -o              report file. Default is stdout\n"
"  -v              explain what is being done\n"
"  -h              display this help\n"
"  --trace         turn on UFSD trace\n"
"  --version       show version and exit\n") );

#ifdef _WIN32
  fprintf( stdout,
_("E.g. " NLS_CAT_NAME " G: -o report.txt\n"
  "E.g. " NLS_CAT_NAME " \\\\?\\Volume{d7c5ce70-3220-11db-b0fb-005056c00008}\n") );
#else
  fprintf( stdout, _("E.g. " NLS_CAT_NAME " /dev/hdb1 -o report.txt\n"));
#endif
}

//
// Extents options structure.
// It will be filled from user' parameter string.
//

struct ext_options {
  UINT64  Lcn;
  UINT64  Len;
  FILE*   rep;
  unsigned
     nopercents:1,
     verbose : 1;

};


///////////////////////////////////////////////////////////
// parse_options
//
// Full trusted function
// Returns false if error
///////////////////////////////////////////////////////////
static bool
parse_options(
    IN int argc,
    IN const char *argv[],
    OUT struct ext_options* opts,
    OUT const char** szDevice
    )
{
  const char* a = NULL;
  int LcnLen = 0; // 1 - Lcn parsed, 2 - Len parsed

  *szDevice  = NULL;

  if ( argc <= 1 )
    return false;

  for ( int i = 1; i < argc; i++ )
  {
    //
    // Make sure arg is in proper format.
    //
    a = argv[i];

    if ( '-' != a[0] )
    {
      // The first "non options argument" is a device name
      if ( NULL != *szDevice )
      {
        // The second "non options arguments" is an range Lcn
        // The third "non options arguments" is an range Len
        if ( (0 == LcnLen && 1 == sscanf( a, "%" PLL "x", &opts->Lcn ))
          || (1 == LcnLen && 1 == sscanf( a, "%" PLL "x", &opts->Len )) )
        {
          LcnLen += 1;
          continue;
        }
        goto Exit;
      }
      *szDevice = a;
    }

    //
    // Which arg is it?
    //
    else switch ( a[1] )
    {
    case 'o':
      if ( 0 != a[2] || i + 1 >= argc )
        goto Exit;
      opts->rep = fopen( argv[++i], "wt" );
      if ( NULL == opts->rep )
        goto Exit;
      break;

    case 'v':
      opts->verbose = 1;
      break;

    case 'h':
      // Force to call OnUsage
      return 0;

    case '-':
      if ( 0 == strcmp( "version", a+2 ) )
      {
        fprintf( stdout, VER_STRING );
        exit(1);
      }
      else if ( 0 == strcmp( "trace", a+2 ) )
      {
        EnableTrace();
        break;
      }

      //
      // !!!No break here!!!
      //

    default:
      goto Exit;    // double break
    }
  }

  // Set default length if only 'Lcn' is given
  if ( 1 == LcnLen )
    opts->Len = 1;

  return true;

Exit:
  fprintf( stderr, _("Error in argument: \"%s\"\n"), a );
  return false; // Force to call OnUsage
}



///////////////////////////////////////////////////////////
// PrintUnicode
//
// Helper function to print unicode strings
///////////////////////////////////////////////////////////
static void
PrintUnicode(
    IN FILE* fid,
    IN const unsigned short* u,
    IN size_t len
    )
{
  for ( size_t i = 0; i < len; i += 1 )
  {
    if ( 0 == (u[i] & 0xFF00)
     && isprint( (char)u[i] ) )
    {
      fprintf( fid, "%c", (char)u[i] );
    }
    else
    {
      fprintf( fid, "\\x%02x", u[i] );
    }
  }
}


///////////////////////////////////////////////////////////
// Helper class to enumerate files and folders via UFSD
//
///////////////////////////////////////////////////////////
class CExtents : public CUfsdEnum
{
  IMessage*     m_IMsg;
  ext_options*  m_Opts;
  char          m_Name[512];      // Buffer for full path
  FILE*         m_Rep;

  //
  // Helper buffers
  //
  RETRIEVAL_POINTERS_BUFFER*    m_Ext;
  size_t                        m_BytesPerExt;

  UFSD_STREAM_INFORMATION*      m_Streams;
  size_t                        m_BytesPerStreams;

  UFSD_GET_RETRIEVAL_POINTERS*  m_GetRp;
  size_t                        m_BytesPerRp;

public:

  CExtents( IN CMemoryManager*  Mm,
            IN IMessage* Msg,
            IN ext_options* opts
           )
    : CUfsdEnum( Mm, false, true )
    , m_IMsg( Msg )
    , m_Opts( opts )
    , m_Ext(NULL)
    , m_BytesPerExt(0)
    , m_Streams(NULL)
    , m_BytesPerStreams(0)
    , m_GetRp(NULL)
    , m_BytesPerRp(0)
  {
    m_Rep = m_Opts->rep;
    if ( NULL == m_Rep )
      m_Rep = stdout;
  }

  ~CExtents()
  {
    free( m_Ext );
    free( m_Streams );
    free( m_GetRp );
  }

  //
  // This virtual function is called from base class for each file/folder
  //
  unsigned ProcessObject(
      IN CFSObject*         Fso,
      IN const t_FileInfo&  FileInfo
      )
  {
    if ( NULL != m_IMsg && m_IMsg->IsBreak() )
      return ERR_CANCELLED;

    if ( NULL == m_Streams )
    {
      m_BytesPerStreams = sizeof(UFSD_STREAM_INFORMATION) + 1024;
      m_Streams         = (UFSD_STREAM_INFORMATION*)malloc( m_BytesPerStreams );
      if ( NULL == m_Streams )
      {
NoMemory:
        fprintf( m_Rep, "No memory\n" );
        return ERR_NOMEMORY;
      }
    }

    if ( NULL == m_Ext )
    {
      m_BytesPerExt = MAXEXTENTS_TO_BYTES( 500 );
      m_Ext         = (RETRIEVAL_POINTERS_BUFFER*)malloc( m_BytesPerExt );
      if ( NULL == m_Ext )
        goto NoMemory;
    }

    if ( NULL == m_GetRp )
    {
      m_BytesPerRp  = sizeof(UFSD_GET_RETRIEVAL_POINTERS) + 0;//512;
      m_GetRp       = (UFSD_GET_RETRIEVAL_POINTERS*)malloc( m_BytesPerRp );
      if ( NULL == m_GetRp )
        goto NoMemory;
    }

    //
    // Print file name
    //
    int NameIsPrinted = 0;

    size_t Tmp;
    unsigned Status;

    m_GetRp->Handle.FsObject = Fso;

    while( ERR_MORE_DATA == (Status =
                FsIoControl( IOCTL_GET_STREAMS2, &m_GetRp->Handle, sizeof(m_GetRp->Handle),
                             m_Streams, m_BytesPerStreams, &Tmp ) ) )
    {
      void* ptr = malloc( m_BytesPerStreams * 2 );
      if ( NULL == ptr )
      {
        fprintf( m_Rep, "No memory\n" );
        return ERR_NOMEMORY;
      }

      free( m_Streams );
      m_BytesPerStreams *= 2;
      m_Streams          = (UFSD_STREAM_INFORMATION*)ptr;
    }

    size_t BytesPerStreams = Tmp;

    //
    // Get extents of all streams
    //
    for ( size_t offset = 0; offset + sizeof(UFSD_STREAM_INFORMATION) <= BytesPerStreams; )
    {
      const UFSD_STREAM_INFORMATION* info = (const UFSD_STREAM_INFORMATION*)((char*)m_Streams+ offset);

      m_GetRp->NameLen      = (unsigned short)info->NameLength;
      m_GetRp->DataOffset   = 0;
      m_GetRp->Type         = info->Type;
      m_GetRp->Instance     = info->Instance;
      m_GetRp->UseInstance  = 1;
      m_GetRp->StartingVcn  = 0;
      if ( sizeof(UFSD_GET_RETRIEVAL_POINTERS) + info->NameLength > m_BytesPerRp )
      {
        void* ptr = malloc( sizeof(UFSD_GET_RETRIEVAL_POINTERS) + info->NameLength );
        if ( NULL == ptr )
          goto NoMemory;
        memcpy( ptr, m_GetRp, sizeof(UFSD_GET_RETRIEVAL_POINTERS) );
        free( m_GetRp );
        m_BytesPerRp  = sizeof(UFSD_GET_RETRIEVAL_POINTERS) + info->NameLength;
        m_GetRp       = (UFSD_GET_RETRIEVAL_POINTERS*)ptr;
      }
      // Copy stream name
      memcpy( m_GetRp + 1, info + 1, info->NameLength );

      size_t StartExtent = 0;
      do {

        //
        // Query retrieval points
        //
        Status = FsIoControl( IOCTL_GET_RETRIEVAL_POINTERS2,
                              m_GetRp, sizeof(UFSD_GET_RETRIEVAL_POINTERS) + m_GetRp->NameLen,
                              m_Ext, m_BytesPerExt, &Tmp );

        //
        // Check errors
        //
        if ( !UFSD_SUCCESS(Status) && ERR_MORE_DATA != Status )
          return Status;

        if ( Tmp < sizeof(RETRIEVAL_POINTERS_BUFFER) )
          return ERR_BADPARAMS;

        if ( 0 == m_Ext->ExtentCount )
          break; // No extents

        const RETRIEVAL_POINTERS_BUFFER::EXTENTS* Ext = m_Ext->Extents;
        unsigned int PrevVcn = m_Ext->StartingVcn;

        if ( 0 != m_Opts->Len )
        {
          for ( size_t i = 0; i < m_Ext->ExtentCount; i += 1, Ext += 1 )
          {
            unsigned int Len = Ext->NextVcn - PrevVcn;
            if ( IS_IN_RANGE( Ext->Lcn, Len, m_Opts->Lcn, m_Opts->Len ) )
              goto TraceExt;
            PrevVcn = Ext->NextVcn;
          }
        }
        else
        {
TraceExt:
          if ( 0 == NameIsPrinted )
          {
            NameIsPrinted = 1;
            UFSD::GetFullName( Fso, GetStringManager(), m_Name, ARRSIZE(m_Name) );
            fprintf( m_Rep, "%s(r=%" PLL "x) :  %s\n", FlagOn(FileInfo.Attrib, UFSD_SUBDIR)? "Folder" : "File", FileInfo.Id, m_Name );
          }

          //
          // Trace extents
          //
          if ( 0 == m_GetRp->StartingVcn )
          {
            fprintf( m_Rep, " Stream 0x%x:", info->Type );
            PrintUnicode( m_Rep, (const unsigned short*)(info+1), info->NameLength/sizeof(short) );
            fprintf( m_Rep, "\n   #      VCN      LCN    COUNT\n" );
          }

          PrevVcn = m_Ext->StartingVcn;
          Ext     = m_Ext->Extents;
          for ( size_t i = 0; i < m_Ext->ExtentCount; i += 1, Ext += 1 )
          {
            fprintf( m_Rep, "%4u %8x %8x %8x\n", (unsigned)(i + StartExtent), PrevVcn, Ext->Lcn, Ext->NextVcn - PrevVcn );
            PrevVcn = Ext->NextVcn;
          }
        }

        StartExtent          += m_Ext->ExtentCount;
        m_GetRp->StartingVcn  = PrevVcn;

      } while( ERR_MORE_DATA == Status );

      if ( 0 == info->NextEntryOffset )
        break;

      offset += info->NextEntryOffset;
    }

    if ( 0 == NameIsPrinted && 0 == m_Opts->Len )
    {
//      NameIsPrinted = 1;
      UFSD::GetFullName( Fso, GetStringManager(), m_Name, ARRSIZE(m_Name) );
      fprintf( m_Rep, "%s(r=%" PLL "x) :  %s\n", FlagOn(FileInfo.Attrib, UFSD_SUBDIR)? "Folder" : "File", FileInfo.Id, m_Name );
    }

    return ERR_NOERROR;
  }
};



///////////////////////////////////////////////////////////
// main
//
//
///////////////////////////////////////////////////////////
int
__cdecl
main(
    IN int argc,
    IN const char* argv[]
    )
{
  ext_options opts;
  unsigned Status       = ERR_BADPARAMS;
  const char* szDevice  = NULL;
  CRWBlock* Rw          = NULL;

#ifndef _WIN32
  setlocale(LC_MESSAGES, "");
  setlocale(LC_CTYPE, "");
  bindtextdomain(NLS_CAT_NAME, LOCALEDIR);
  textdomain(NLS_CAT_NAME);
#endif

  if ( argc < 1 )
    return -1;  // Impossible

  const char* szErr = CheckUFSDEndian();
  if ( NULL != szErr )
  {
    fprintf( stderr, "%s\n", szErr );
    return -1;
  }

  // Set default options
  memset( &opts, 0, sizeof(opts) );

  // Parse command line
  // And exit if there are errors

  if ( !parse_options( argc, argv, &opts, &szDevice ) )
  {
    OnUsage();
    return -1;
  }

  if ( NULL == szDevice )
  {
    fprintf( stderr, _("Device is not supplied\n") );
    return -2;
  }

#ifdef _WIN32

  char szVolume[64];
  if ( ':' == szDevice[1] )
    _snprintf( szVolume, ARRSIZE(szVolume), "\\\\.\\%s", szDevice );
  else
    _snprintf( szVolume, ARRSIZE(szVolume), szDevice );
  szVolume[ARRSIZE(szVolume)-1] = 0;
  szDevice = szVolume;

#else

  //
  // Check for mounted devices
  //
  const char* szMount = GetMount( szDevice, opts.verbose );
  if ( NULL != szMount )
  {
    fprintf( stderr, _("\nThe device \"%s\" is mounted as \"%s\"\n"
                     "Enumerate is not implemented yet\n\n"),
                     szDevice, szMount );
    return -3;
  }

#endif

  //
  // Init libufsd.so.
  // Without such initialization library is not able to show messages and errors
  //
  InitUfsdLibrary();

  //
  // Allocate BlockDevice
  //
  Status = UFSD_IOHandlerCreate( szDevice, true, &Rw, NULL, false, opts.verbose );

  if ( ERR_NOERROR == Status )
  {
    assert(NULL != Rw);

    //
    // Init UFSD and enumerate all files/folders
    //
    CMemoryManager* mm  = UFSD_GetMemoryManager();
    CStrings* str       = UFSD_GetStringsService();
    CTime*    tm        = UFSD_GetTimeService();
    IMessage* msg       = UFSD_GetMessageService();
    size_t real_fs;

    CExtents  ext( mm, msg, &opts );
    Status = ext.Init( str, Rw, MINUS_ONE_T, tm, &real_fs, NTFS_OPTIONS_INCLUDE_BCFS );
    if ( ERR_NOERROR == Status )
    {
      //
      // Call recursive function
      //
      Status = ext.EnumDir( NULL );
    }
  }

  //
  // Free all resources
  //
  UFSD_IOHandlerFree( Rw );

  if ( NULL != opts.rep )
    fclose( opts.rep );

  //
  // Translate error code to string
  //
  PrintUfsdStatusIfError( Status );

  // Exit to host OS
  return Status;
}
