////////////////////////////////////////////////////////////////
// March of 2008. (C) Paragon Software
// If this code works, it was written by Alexander Mamaev
// If not, I don't know who wrote it.
//
// This file contains the code that:
// - parses command line
// - calls UFSD library
//
////////////////////////////////////////////////////////////////

#include <stdio.h>
#include <stdlib.h> // exit
#include <string.h> // memset
#include <assert.h>

//
// Include main ufsd header
//
#include "ufsd.h"

//
// All UFSD classes and functions are placed in UFSD namespace
//
using namespace UFSD;

#ifndef NLS_CAT_NAME
# define NLS_CAT_NAME "mkhfs"
#endif

//
// This file contains declarations of useful UFSD-client functions
//
#include "funcs.h"
#include "gettext.h"

#ifdef _WIN32
//---------------------------------------------------------
// To compile windows utility
//---------------------------------------------------------
#include <windows.h>  // DeviceIoControl
#pragma warning( disable: 4201 ) // nonstandard extension used : nameless struct/union
#include <winioctl.h>
#include <io.h>       // _get_osfhandle

#else
//---------------------------------------------------------
// To compile linux utility
//---------------------------------------------------------

#ifndef __QNX__
#include <linux/hdreg.h>
#endif
#include <sys/ioctl.h>
#include <locale.h>

//
// When this file is compiled as a Linux utility
// then some useful functions (GetAbsPath,ParsePath) that are shared for
// many projects are placed in file misc.cpp
//
# include "misc.h"

//---------------------------------------------------------
// End of Linux specific
//---------------------------------------------------------
#endif // _WIN32



///////////////////////////////////////////////////////////
// OnAbout
//
//
///////////////////////////////////////////////////////////
static void
OnAbout()
{
  fprintf( stdout, _("Create a Mac OS Extended (HFS+) volume on a user specified (block) device.\n") );
}


///////////////////////////////////////////////////////////
// OnUsage
//
// print a list of the parameters to the program
///////////////////////////////////////////////////////////
static void
OnUsage()
{
  OnAbout();
  fprintf( stdout, _(
"Usage: \"%s\" [options] device\n"
"  -v:label          Specify the volume label.\n"
"  -a:size           Override the default allocation unit size. Default settings\n"
"                    are strongly recommended for general use.\n"
"                    HFS supports 512, 1024, 2048, 4096, 8192, 16K, 32K, 64K.\n"
"  -ne:size          Extents b-tree node size 512-32K\n"
"  -nc:size          Catalog b-tree node size 4K-32K\n"
"  -f                Force to format without question\n"
"  -j                Journaled\n"
"  -c                Case-sensitive\n"
"  -h                display this help\n"
"  --help            Display this help\n"
"  --trace           Turn on UFSD trace\n"
"  --verbose         Explain what is being done\n"
"  --nopercents      Do not print percents\n"
"  --version         Show version and exit\n"), NLS_CAT_NAME );
#ifdef _WIN32
fprintf( stdout,
"  --type:type       MBR partition type in hex ( 7 - ntfs, af - hfs )\n" );
#endif
#ifdef _WIN32
  fprintf( stdout,
_("E.g. \"%s\" g:\n"
  "E.g. \"%s\" \\\\?\\Volume{d7c5ce70-3220-11db-b0fb-005056c00008}\n"), NLS_CAT_NAME, NLS_CAT_NAME );
#else
  fprintf( stdout, _("E.g. \"%s\" /dev/hdb1\n"), NLS_CAT_NAME );
#endif
}


//
// Format options structure.
// It will be filled from user' parameter string.
//

struct hfs_format_options {
  OptFormat     ufsd;
  const char*   szLabel;            // NULL for empty label
  unsigned int  BytesPerBlock;
#ifdef _WIN32
  unsigned int  Type;               // MBR partition type
#endif

  unsigned
     quiet:1,
     nopercents:1,
     verbose:1;
};


///////////////////////////////////////////////////////////
// parse_options
//
// Full trusted function
// Returns 0 if error
///////////////////////////////////////////////////////////
static bool
parse_options(
    IN int argc,
    IN const char *argv[],
    OUT hfs_format_options* opts,
    OUT const char** szDevice
    )
{
  const char* a = NULL;

  // Fill the default options
  opts->BytesPerBlock   = 4096;

  *szDevice = NULL;

  if ( argc <= 1 )
    return false;

  for ( int i = 1; i < argc; i++ )
  {
    //
    // Make sure arg is in proper format.
    //
    a = argv[i];

    if ( '-' != a[0] )
    {
      // The first "non options argument" is a device name
      if ( NULL != *szDevice )
        goto Exit;
      *szDevice = a;
      continue;
    }

    //
    // Which arg is it?
    //
    switch ( a[1] )
    {
    case 'v':
      if ( ':' != a[2] )
        goto Exit;
      opts->szLabel = a + 3;
      break;

    case 'q':
      break;

    case 'c':
      opts->ufsd.Hfs.CaseSensitive = true;
      break;

    case 'j':
      opts->ufsd.Hfs.Journal = true;
      break;

    case 'a':
      if ( ':' != a[2] )
        goto Exit;

      opts->BytesPerBlock = ParseSize( a + 3 );
      if ( !IsPowerOf2( opts->BytesPerBlock ) || opts->BytesPerBlock > 64*1024 )
        goto Exit;
      break;

    case 'n':
      if ( ':' == a[3] )
      {
        unsigned int Bytes = ParseSize( a + 4 );
        if ( !IsPowerOf2( Bytes ) || Bytes > 64*1024 )
          goto Exit;
        if ( 'e' == a[2] )
        {
          opts->ufsd.Hfs.eNodeSize = Bytes;
          break;
        }
        if ( 'c' == a[2] )
        {
          opts->ufsd.Hfs.cNodeSize = Bytes;
          break;
        }
      }
      goto Exit;

    case 'f':
      opts->quiet = 1;
      break;

    case 'h':
      return false; // Force to call OnUsage

    case '-':
      if ( 0 == strcmp( "version", a + 2 ) )
      {
        fprintf( stdout, VER_STRING );
        exit(1);
      }
      else if ( 0 == strcmp( "trace", a + 2 ) )
      {
        EnableTrace();
        break;
      }
      else if ( 0 == strcmp( "verbose", a + 2 ) )
      {
        opts->verbose = 1;
        break;
      }
      else if ( 0 == strcmp( "nopercents", a + 2 ) )
      {
        opts->nopercents = 1;
        break;
      }
      else if ( 0 == strcmp( "help", a + 2 ) )
      {
        // Force to call OnUsage
        return false;
      }
#ifdef _WIN32
      else if ( 0 == memcmp( "type:", a + 2, 5 ) && 1 == sscanf( a + 7, "%x", &opts->Type ) )
      {
        break;
      }
#endif

      //
      // !!!No break here!!!
      //

    default:
      goto Exit;  // double break
    }
  }

  return true;

Exit:
  fprintf( stderr, _("Error in argument: \"%s\"\n"), a );
  return false; // Force to call OnUsage
}


///////////////////////////////////////////////////////////
// main
//
//
///////////////////////////////////////////////////////////
int __cdecl
main(
    IN int argc,
    IN const char* argv[]
    )
{
  unsigned Status = ERR_BADPARAMS;
  const char* szDevice = NULL;
  hfs_format_options opts;
  CRWBlock* Rw = NULL;
  int Device = -1;

#ifndef _WIN32
  setlocale(LC_MESSAGES, "");
  setlocale(LC_CTYPE, "");
  bindtextdomain(NLS_CAT_NAME, LOCALEDIR);
  textdomain(NLS_CAT_NAME);
#endif

  memset( &opts, 0, sizeof(opts) );

  if ( argc < 1 )
    return -1;  // Impossible

  const char* szErr = CheckUFSDEndian();
  if ( NULL != szErr )
  {
    fprintf( stderr, "%s\n", szErr );
    return -1;
  }

  // Parse command line
  // And exit if there are errors
  if ( !parse_options( argc, argv, &opts, &szDevice ) )
  {
    OnUsage();
    return -1;
  }

  if ( NULL == szDevice )
  {
    fprintf( stderr, _("Device is not supplied\n") );
    return -2;
  }

#ifdef _WIN32

  char szVolume[64];
  if ( ':' == szDevice[1] )
    _snprintf( szVolume, ARRSIZE(szVolume), "\\\\.\\%s", szDevice );
  else
    _snprintf( szVolume, ARRSIZE(szVolume), szDevice );
  szVolume[ARRSIZE(szVolume)-1] = 0;
  szDevice = szVolume;

#elif !defined __QNX__

  //
  // Check for mounted devices
  //
  const char* szMount = GetMount( szDevice, opts.verbose );
  if ( NULL != szMount )
  {
    fprintf( stderr, _("\nThe device \"%s\" is mounted as \"%s\"\n"
                       "Impossible to format\n\n"),
                     szDevice, szMount );
    return -3;
  }

#endif


  //
  // Init libufsd.so.
  // Without such initialization library is not able to show messages and errors
  //
  InitUfsdLibrary();

  //
  // Allocate BlockDevice
  //
  Status = UFSD_IOHandlerCreate( szDevice, 0, &Rw, &Device, false, opts.verbose );

  if ( ERR_NOERROR == Status )
  {
    UFSD::CMemoryManager* Mm  = UFSD_GetMemoryManager();
    UFSD::CStrings*       Ss  = UFSD_GetStringsService();
    UFSD::CTime*          Tt  = UFSD_GetTimeService();
    UFSD::IMessage*       Msg = UFSD_GetMessageService( stdout, NULL, opts.nopercents );

    assert(NULL != Rw);

    //
    // Ask the user
    //
    unsigned short BytesPerSector = (unsigned short)Rw->GetSectorSize();
    unsigned short Log2OfSector   = (unsigned short)(9 + UFSD::Log2( BytesPerSector >> 9 ));
    UINT64 TotalSectors = Rw->GetNumberOfBlocks();
    unsigned int  Gb    = LODWORD(TotalSectors >> (30-Log2OfSector));
    unsigned int  Mb    = LODWORD(TotalSectors) & ((1U<<(30-Log2OfSector))-1);

    fprintf( stdout, _("WARNING, ALL DATA ON DEVICE \"%s\"(%u.%02u Gb) WILL BE LOST!\nProceed with Format (y/n)?"),
             szDevice,
             Gb, (Mb*100)>>21 );

    //
    // Ask the user
    //
    if ( 0 != opts.quiet )
      fprintf( stdout, "y\n" );
    else if ( 'y' != getc(stdin) )
    {
      Status = ERR_CANCELLED;
      goto Exit;
    }

    if ( opts.verbose )
    {
      fprintf ( stdout, _("Formatting %s:\n"),
                opts.ufsd.Hfs.CaseSensitive
                ? opts.ufsd.Hfs.Journal
                  ? "Mac OS Extended (Journaled,Case-sensitive)"
                  : "Mac OS Extended (Case-sensitive)"
                : opts.ufsd.Hfs.Journal
                  ? "Mac OS Extended (Journaled)"
                  : "Mac OS Extended" );
    }

    //
    // Call UFSD code
    //
    Status = UFSD::FormatFs( FS_HFSPLUS, Mm, Ss, Rw, Tt, Msg, true, // Quick
                             0, opts.BytesPerBlock,
                             StrASCII, opts.szLabel,
                             &opts.ufsd, szDevice );


    if ( ERR_NOERROR == Status )
    {
      //
      // We have to change fdisk type of partition (if Device is a basic partition)
      //
#ifdef _WIN32
      DWORD Tmp;
      HANDLE hDevice = (HANDLE)::_get_osfhandle(Device);
      BOOL bOK = FALSE;
      if ( 0 == opts.Type )
        opts.Type = 0xaf;

      PARTITION_INFORMATION PartInfo;
      if ( ::DeviceIoControl( hDevice, IOCTL_DISK_GET_PARTITION_INFO,
                              NULL, 0,
                              &PartInfo, sizeof(PartInfo),
                              &Tmp, NULL ) )
      {
        SET_PARTITION_INFORMATION SetType;
        SetType.PartitionType = (BYTE)opts.Type;
        bOK = SetType.PartitionType == PartInfo.PartitionType
             || ::DeviceIoControl( hDevice, IOCTL_DISK_SET_PARTITION_INFO,
                                   &SetType, sizeof(SetType),
                                   NULL, 0,
                                   &Tmp, NULL );
      }

      if ( !bOK )
      {
        fprintf( stderr, _("Partition is formatted but its fdisk type is not changed\n") );
      }
#else
      //
      // Unfortunately I don't know how to change partition type in Linux
      //
#endif
      fprintf( stdout, _("OK\n") );
    }
  }

Exit:

  //
  // Free all resources
  //
  UFSD_IOHandlerFree( Rw );

  //
  // Translate error code to string
  //
  PrintUfsdStatusIfError( Status );

  // Exit to Linux
  return Status;
}
