////////////////////////////////////////////////////////////////
// October of 2004. (C) Paragon Software
// If this code works, it was written by Alexander Mamaev
// If not, I don't know who wrote it.
//
// This file contains the code that:
// - parses command line
// - calls UFSD library
//
////////////////////////////////////////////////////////////////

#include <stdio.h>
#include <stdlib.h> // exit
#include <string.h> // memset
#include <assert.h>

//
// Include main ufsd header
//
#include "ufsd.h"

//
// All UFSD classes and functions are placed in UFSD namespace
//
using namespace UFSD;

#ifndef NLS_CAT_NAME
# define NLS_CAT_NAME "mkntfs"
#endif

//
// This file contains declarations of useful UFSD-client functions
//
#include "funcs.h"
#include "gettext.h"

#ifdef _WIN32
//---------------------------------------------------------
// To compile windows utility
//---------------------------------------------------------
#include <windows.h>  // DeviceIoControl
#pragma warning( disable: 4201 ) // nonstandard extension used : nameless struct/union
#include <winioctl.h>
#include <io.h>       // _get_osfhandle

#else
//---------------------------------------------------------
// To compile linux utility
//---------------------------------------------------------

#ifndef __QNX__
#include <linux/hdreg.h>
#endif
#include <sys/ioctl.h>
#include <locale.h>

//
// When this file is compiled as a Linux utility
// then some useful functions (GetAbsPath,ParsePath) that are shared for
// many projects are placed in file misc.cpp
//
# include "misc.h"

//---------------------------------------------------------
// End of Linux specific
//---------------------------------------------------------
#endif // _WIN32



///////////////////////////////////////////////////////////
// OnAbout
//
//
///////////////////////////////////////////////////////////
static void
OnAbout(void)
{
  fprintf( stdout, _("Create an NTFS volume on a user specified (block) device.\n") );
}


///////////////////////////////////////////////////////////
// OnUsage
//
// print a list of the parameters to the program
///////////////////////////////////////////////////////////
static void
OnUsage()
{
  OnAbout();
  fprintf( stdout, _(
"Usage: " NLS_CAT_NAME " [options] device\n"
"  -v:label          Specify the volume label.\n"
"  -c                Files created on the new volume will be compressed by default.\n"
"  -a:size           Override the default allocation unit size. Default settings\n"
"                    are strongly recommended for general use.\n"
"                    NTFS supports 512, 1024, 2048, 4096, 8192, 16K, 32K, 64K.\n"
"                    NTFS compression is not supported for allocation unit sizes above 4096.\n"
"  -b:size           Override the default block (sector) size. Default settings\n"
"                    are strongly recommended for general use.\n"
"                    One can use 512, 1024, 2048, 4096.\n"
"  -m:size           Override the default mft record size. Default settings\n"
"                    are strongly recommended for general use.\n"
"                    One can use 512, 1024, 2048, 4096.\n"
"  -f                Force to format without question\n"
"  -s:start          Specify \"hidden\" sectors  in boot area\n"
"  -g:tracks:sectors Specify the disk geometry that should be written in boot area\n"
"                    \"tracks\"  Specifies the number of tracks per disk side\n"
"                    \"sectors\" Specifies the number of sectors per track\n"
"                    The most known geometries are:\n"
"                    NORMAL: 63 sectors per track and 15(16) tracks per cylinder\n"
"                    LBA:    63 sectors per track and 255 tracks per cylinder\n"
"                    In general Windows uses LBA geometry (-g:255:63)\n"
"                    If -g is not specified this program gets geometry from Linux\n"
"  -winxp            Create NTFS for Windows XP (default)\n"
"  -winvista         Create NTFS for Windows Vista\n"
"  -win7             Create NTFS for Windows 7/8\n"
"  -h                Display this help\n"
"  --help            Display this help\n"
"  --trace           Turn on UFSD trace\n"
"  --verbose         Explain what is being done\n"
"  --nopercents      Do not print percents\n"
"  --version         Show version and exit\n") );
#ifdef _WIN32
  fprintf( stdout,
_("E.g. " NLS_CAT_NAME " g:\n"
  "E.g. " NLS_CAT_NAME " \\\\?\\Volume{d7c5ce70-3220-11db-b0fb-005056c00008}\n"));
#else
  fprintf( stdout, _("E.g. " NLS_CAT_NAME " /dev/hdb1\n") );
#endif
}


//
// Format options structure.
// It will be filled from user' parameter string.
//

struct ntfs_format_options {
  OptFormat     ufsd;
  const char*   szLabel;            // NULL for empty label
  unsigned int  BytesPerCluster;
  unsigned int  BytesPerSector;     // 0 means default

  unsigned
     quiet:1,
     use_hidden:1,  // Use HiddenSectors
     use_geom:1,    // Use SectorsPerTrack and Use HeadsNumber
     nopercents:1,
     verbose:1;
};



///////////////////////////////////////////////////////////
// parse_options
//
// Full trusted function
// Returns 0 if error
///////////////////////////////////////////////////////////
static bool
parse_options(
    IN int argc,
    IN const char *argv[],
    OUT struct ntfs_format_options* opts,
    OUT const char** szDevice
    )
{
  const char* a = NULL;
  unsigned int HeadsNumber, SectorsPerTrack, Bytes;

  // Fill the default options
  opts->BytesPerCluster   = 4096;

  *szDevice = NULL;

  if ( argc <= 1 )
    return false;

  for ( int i = 1; i < argc; i++ )
  {
    //
    // Make sure arg is in proper format.
    //
    a = argv[i];

    if ( '-' != a[0] )
    {
      // The first "non options argument" is a device name
      if ( NULL != *szDevice )
        goto Exit;
      *szDevice = a;
      continue;
    }

    //
    // Which arg is it?
    //

    else switch ( a[1] )
    {
    case 'v':
      if ( ':' != a[2] )
        goto Exit;
      opts->szLabel = a + 3;
      break;

    case 'q':
      break;

    case 'c':
      opts->ufsd.Ntfs.bEnableCompression = 1;
      break;

    case 'a':
      if ( ':' != a[2] )
        goto Exit;
      // Check for 512, 1024, 2048, 4096, 8192, 16K, 32K, 64K
      Bytes = ParseSize( a + 3 );
      if ( !IsPowerOf2( Bytes ) || Bytes > 64*1024 )
        goto Exit;

      // OK
      opts->BytesPerCluster = Bytes;
      break;

    case 'b':
      if ( ':' != a[2] )
        goto Exit;
      // Check for 512, 1024, 2048, 4096
      Bytes = ParseSize( a + 3 );
      if ( !IsPowerOf2( Bytes ) || Bytes > 4*1024 )
        goto Exit;

      // OK
      opts->BytesPerSector = Bytes;
      break;

    case 'm':
      if ( ':' != a[2] )
        goto Exit;
      // Check for 512, 1024, 2048, 4096
      Bytes = ParseSize( a + 3 );
      if ( !IsPowerOf2( Bytes ) || Bytes > 4*1024 )
        goto Exit;

      // OK
      opts->ufsd.Ntfs.BytesPerRecord = Bytes;
      break;

    case 'f':
      opts->quiet = 1;
      break;

    case 's':
      if ( ':' != a[2] )
        goto Exit;

      opts->use_hidden = 1;
      opts->ufsd.Ntfs.HiddenSectors = (unsigned)atol( a + 3 );
      break;

    case 'g':
      // Parse -g:tracks:sectors
      if ( ':' != a[2] || 2 != sscanf( a + 3, "%x:%x", &HeadsNumber, &SectorsPerTrack ) )
        goto Exit;

      opts->ufsd.Ntfs.wHeadNumber      = (unsigned short)HeadsNumber;
      opts->ufsd.Ntfs.wSectorsPerTrack = (unsigned short)SectorsPerTrack;
      opts->use_geom   = 1;
      break;

    case 'w':
      // default Win2K/XP, 6 - Vista, 61 - Win7
      if ( 0 == strcmp( "inxp", a + 2 ) )
        opts->ufsd.Ntfs.Version = 0; // Default
      else if ( 0 == strcmp( "invista", a + 2 ) )
        opts->ufsd.Ntfs.Version = 60;
      else if ( 0 == strcmp( "in7", a + 2 ) )
        opts->ufsd.Ntfs.Version = 610;
      else
        goto Exit;
      break;

    case 'h':
      return false; // Force to call OnUsage

    case '-':
      if ( 0 == strcmp( "version", a + 2 ) )
      {
        fprintf( stdout, VER_STRING );
        exit(1);
      }
      else if ( 0 == strcmp( "trace", a + 2 ) )
      {
        EnableTrace();
        break;
      }
      else if ( 0 == strcmp( "verbose", a + 2 ) )
      {
        opts->verbose = 1;
        break;
      }
      else if ( 0 == strcmp( "nopercents", a + 2 ) )
      {
        opts->nopercents = 1;
        break;
      }
      else if ( 0 == strcmp( "help", a + 2 ) )
      {
        // Force to call OnUsage
        return false;
      }
      else if ( 0 == strcmp( "bcfs", a + 2 ) )
      {
        opts->ufsd.Ntfs.BcFs = 1;
        return false;
      }

      //
      // !!!No break here!!!
      //

    default:
      goto Exit;  // double break
    }
  }

  //
  // Check cluster size and compress
  //
  if ( opts->BytesPerCluster > 4096
    && 0 != opts->ufsd.Ntfs.bEnableCompression )
  {
    fprintf( stderr, _("Impossible to create an compressed NTFS when cluster size is more than 4096!\n") );
    // Force to call OnUsage
    return false;
  }

  return true;

Exit:
  fprintf( stderr, _("Error in argument: \"%s\"\n"), a );
  return false; // Force to call OnUsage
}


///////////////////////////////////////////////////////////
// main
//
//
///////////////////////////////////////////////////////////
int __cdecl
main(
    IN int argc,
    IN const char* argv[]
    )
{
  unsigned Status = ERR_BADPARAMS;
  const char* szDevice = NULL;
  ntfs_format_options opts;
  CRWBlock* Rw = NULL;
  int Device = -1;

#ifndef _WIN32
  setlocale(LC_MESSAGES, "");
  setlocale(LC_CTYPE, "");
  bindtextdomain(NLS_CAT_NAME, LOCALEDIR);
  textdomain(NLS_CAT_NAME);
#endif

  memset( &opts, 0, sizeof(opts) );

  if ( argc < 1 )
    return -1;  // Impossible

  const char* szErr = CheckUFSDEndian();
  if ( NULL != szErr )
  {
    fprintf( stderr, "%s\n", szErr );
    return -1;
  }

  // Parse command line
  // And exit if there are errors
  if ( !parse_options( argc, argv, &opts, &szDevice ) )
  {
    OnUsage();
    return -1;
  }

  if ( NULL == szDevice )
  {
    fprintf( stderr, _("Device is not supplied\n") );
    return -2;
  }

#ifdef _WIN32

  char szVolume[64];
  if ( ':' == szDevice[1] )
    _snprintf( szVolume, ARRSIZE(szVolume), "\\\\.\\%s", szDevice );
  else
    _snprintf( szVolume, ARRSIZE(szVolume), szDevice );
  szVolume[ARRSIZE(szVolume)-1] = 0;
  szDevice = szVolume;

#elif !defined __QNX__

  //
  // Check for mounted devices
  //
  const char* szMount = GetMount( szDevice, opts.verbose );
  if ( NULL != szMount )
  {
    fprintf( stderr, _("\nThe device \"%s\" is mounted as \"%s\"\n"
                       "Impossible to format\n\n"),
                     szDevice, szMount );
    return -3;
  }

#endif


  //
  // Init libufsd.so.
  // Without such initialization library is not able to show messages and errors
  //
  InitUfsdLibrary();

  //
  // Allocate BlockDevice
  //
  Status = UFSD_IOHandlerCreate( szDevice, 0, &Rw, &Device, false, opts.verbose );

  if ( ERR_NOERROR == Status )
  {
    UFSD::CMemoryManager* Mm  = UFSD_GetMemoryManager();
    UFSD::CStrings*       Ss  = UFSD_GetStringsService();
    UFSD::CTime*          Tt  = UFSD_GetTimeService();
    UFSD::IMessage*       Msg = UFSD_GetMessageService( stdout, NULL, opts.nopercents );

    assert(NULL != Rw);

    //
    // Ask the user
    //
    unsigned short BytesPerSector = (unsigned short)Rw->GetSectorSize();
    unsigned short Log2OfSector   = (unsigned short)(9 + UFSD::Log2( BytesPerSector >> 9 ));
    UINT64 TotalSectors = Rw->GetNumberOfBlocks();
    unsigned int  Gb    = LODWORD(TotalSectors >> (30-Log2OfSector));
    unsigned int  Mb    = LODWORD(TotalSectors) & ((1U<<(30-Log2OfSector))-1);

    fprintf( stdout, _("WARNING, ALL DATA ON DEVICE \"%s\"(%u.%02u Gb) WILL BE LOST!\nProceed with Format (y/n)?"),
             szDevice,
             Gb, (Mb*100)>>21 );

    //
    // Ask the user
    //
    if ( 0 != opts.quiet )
      fprintf( stdout, "y\n" );
    else if ( 'y' != getc(stdin) )
    {
      Status = ERR_CANCELLED;
      goto Exit;
    }

    //
    // Hidden sectors - wSectorsPerTrack for logical partitions
    //                - start sector for primary partitions
    //                  the first primary partition allocates
    //                  at wSectorsPerTrack sector
    //
    // Really these values are used in boot sector only
    //                      offset    length
    // SectorsPerTrack    : 0x18        2
    // HeadNumber         : 0x1A        2
    // HiddenSectors      : 0x1C        4
    if ( !opts.use_hidden )
      opts.ufsd.Ntfs.HiddenSectors    = 63;
    if ( !opts.use_geom )
    {
      opts.ufsd.Ntfs.wSectorsPerTrack = 63;
      opts.ufsd.Ntfs.wHeadNumber      = 255;
    }

    //
    // Get drive geometry
    //
    if ( opts.use_geom && opts.use_hidden )
    {
      if ( opts.verbose )
        fprintf( stdout, "Use predefined disk geometry\n" );
    }
    else
    {
#ifdef _WIN32
      DWORD Tmp;
      HANDLE hDevice = (HANDLE)_get_osfhandle(Device);

      if ( !opts.use_geom )
      {
        DISK_GEOMETRY Geom;
        if ( !::DeviceIoControl( hDevice, IOCTL_DISK_GET_DRIVE_GEOMETRY, NULL, 0, &Geom, sizeof(Geom), &Tmp, NULL ) )
        {
          fprintf( stderr, "Can't get drive geometry for \"%s\". Use default\n", szDevice );
        }
        else
        {
          opts.ufsd.Ntfs.wSectorsPerTrack  = (unsigned short)Geom.SectorsPerTrack;
          opts.ufsd.Ntfs.wHeadNumber       = (unsigned short)Geom.TracksPerCylinder;
        }
      }

      if ( !opts.use_hidden )
      {
        PARTITION_INFORMATION PartInfo;
        if ( !::DeviceIoControl( hDevice, IOCTL_DISK_GET_PARTITION_INFO,
                                 NULL, 0,
                                &PartInfo, sizeof(PartInfo),
                                &Tmp, NULL ) )
        {
          fprintf( stderr, "Can't get partition position for \"%s\". Use default\n", szDevice );
        }
        else
        {
          opts.ufsd.Ntfs.HiddenSectors = PartInfo.HiddenSectors;
        }
      }
#elif !defined __QNX__
      struct hd_geometry geometry;

      if ( ioctl ( Device, HDIO_GETGEO, &geometry ) )
      {
        perror( "ioctl(HDIO_GETGEO) failed" );
        fprintf ( stderr, "Unable to get drive geometry, use default\n" );
      }
      else
      {
        // Try to detect primary/logical devices
        const char* last = szDevice + strlen(szDevice) - 1;
        while( last >= szDevice && *last >= '0' && *last <='9' )
          last--;

        long Num          = atol( last + 1 );
        bool bPrimary     = 1 <= Num && Num <= 4;

        if ( opts.verbose )
          fprintf( stdout, "Device \"%s\" is parsed as %s partition\n", szDevice, bPrimary?"primary":"logical" );

        if ( !opts.use_hidden )
          opts.ufsd.Ntfs.HiddenSectors     = bPrimary? geometry.start : geometry.sectors;
        if ( !opts.use_geom )
        {
          opts.ufsd.Ntfs.wSectorsPerTrack  = geometry.sectors;
          opts.ufsd.Ntfs.wHeadNumber       = geometry.heads;
        }
      }
#else
      opts.ufsd.Ntfs.HiddenSectors = 0;
#endif
    }

    if ( opts.verbose )
    {
      fprintf ( stdout, "Use following parameters in boot area:\n"
                        "Hidden sectors      = 0x%x\n"
                        "Sectors per track   = 0x%x (%u)\n"
                        "Tracks per cylinder = 0x%x (%u)\n",
                        opts.ufsd.Ntfs.HiddenSectors,
                        (unsigned)opts.ufsd.Ntfs.wSectorsPerTrack,(unsigned)opts.ufsd.Ntfs.wSectorsPerTrack,
                        (unsigned)opts.ufsd.Ntfs.wHeadNumber, (unsigned)opts.ufsd.Ntfs.wHeadNumber );
    }

    opts.ufsd.Ntfs.MediaByte = 0xF8;

    //
    // Call UFSD code
    //
    Status = UFSD::FormatFs( FS_NTFS, Mm, Ss, Rw, Tt, Msg, true, // QuickFormat
                             (unsigned short)opts.BytesPerSector,
                             opts.BytesPerCluster,
                             StrASCII, opts.szLabel,
                             &opts.ufsd, szDevice );


    if ( ERR_NOERROR == Status )
    {
      //
      // We have to change fdisk type of partition (if Device is a basic partition)
      //
#ifdef _WIN32
      DWORD Tmp;
      HANDLE hDevice = (HANDLE)_get_osfhandle(Device);
      BOOL bOK = FALSE;

      PARTITION_INFORMATION PartInfo;
      if ( ::DeviceIoControl( hDevice, IOCTL_DISK_GET_PARTITION_INFO,
                               NULL, 0,
                               &PartInfo, sizeof(PartInfo),
                               &Tmp, NULL ) )
      {
        SET_PARTITION_INFORMATION SetType;
        SetType.PartitionType = PARTITION_IFS;
        bOK = SetType.PartitionType == PartInfo.PartitionType
             || ::DeviceIoControl( hDevice, IOCTL_DISK_SET_PARTITION_INFO,
                                   &SetType, sizeof(SetType),
                                   NULL, 0,
                                   &Tmp, NULL );
      }

      if ( !bOK )
      {
        fprintf( stderr, _("Partition is formatted but its fdisk type is not changed\n") );
      }
#else
      //
      // Unfortunately I don't know how to change partition type in Linux
      //
#endif
      fprintf( stdout, _("OK\n") );
    }
  }

Exit:

  //
  // Free all resources
  //
  UFSD_IOHandlerFree( Rw );

  //
  // Translate error code to string
  //
  PrintUfsdStatusIfError( Status );

  // Exit to Linux
  return Status;
}
