#include <stdio.h>
#include <stdarg.h>
#include <errno.h>
#include <assert.h>
#include <sys/types.h>
#include <sys/stat.h>
#ifndef __USE_LARGEFILE64
#define __USE_LARGEFILE64
#endif
#ifndef __USE_FILEOFFSET64
#define __USE_FILEOFFSET64
#endif
#ifndef __USE_UNIX98
#define __USE_UNIX98
#endif
#include <fcntl.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h> // ispint

#include <unistd.h>
#include <dirent.h>
#include <poll.h>

#define _STDIN  0
#define _STDOUT 1

//
// Include main ufsd header
//
#include "ufsd.h"

//
// All UFSD classes and functions are placed in UFSD namespace
//
using namespace UFSD;

#ifndef NLS_CAT_NAME
# define NLS_CAT_NAME "sysdump"
#endif

//
// This file contains declarations of useful UFSD-client functions
//
#include "funcs.h"
#include "gettext.h"

//---------------------------------------------------------
// To compile linux utility
//---------------------------------------------------------

#include <locale.h>

//
// When this file is compiled as a Linux utility
// then some useful functions (GetAbsPath,ParsePath) that are shared for
// many projects are placed in file misc.cpp
//
# include "misc.h"

//---------------------------------------------------------
// End of Linux specific
//---------------------------------------------------------

#include "zlib.h"
#include <sys/utsname.h>
#include <sys/statfs.h>
#include <time.h>
#include <sys/ioctl.h>

// Add missing defines.
#ifndef O_BINARY
#define O_BINARY 0
#endif
#ifndef O_LARGEFILE
#define O_LARGEFILE 0
#endif

#ifndef S_IRUSR
#define S_IRUSR 0
#define S_IWUSR 0
#endif
#ifndef S_IRGRP
#define S_IRGRP 0
#define S_IWGRP 0
#endif

#ifndef S_ISDIR
#define S_ISDIR(m)  (((m) & S_IFMT) == S_IFDIR)
#endif

#ifndef __ANDROID__
# ifndef __off64_t_defined
#  define open64 open
#  define lseek64 lseek
#  define pread64 pread
#  define pwrite64 pwrite
# endif
#else
static inline unsigned int pread64(int h, void *b, unsigned int c, long long off) {
  lseek64(h, off, SEEK_SET);
  return read(h, b, c);
}
#endif

#ifndef verify
#ifdef NDEBUG
#define verify(f)     ((void)(f))
#else
#define verify(f)     assert(f)
#endif
#endif

// "true" when [s,s+c) intersects with [l,l+w)
#define IS_IN_RANGE(s,c,l,w)   (   ( (l) <= (s) && (s) < ((l)+(w)) ) || ( (s) <= (l) && ((s)+(c)) >= ((l)+(w)) ) || ( (l) <  ((s)+(c)) && ((s)+(c)) < ((l)+(w)) )    )

///////////////////////////////////////////////////////////
// GetBit
//
// Get bit value
///////////////////////////////////////////////////////////
inline bool
GetBit(
        IN const unsigned char* Map,
        IN size_t               Bit
        )
{
    return 0 != ( Map[Bit>>3] >> (Bit & 7) & 1 );
}

//
//  Lookup table that tells how many bits are set in a byte
//
static const unsigned char BitsSetTotal[] = {
    0,1,1,2,1,2,2,3,1,2,2,3,2,3,3,4,
    1,2,2,3,2,3,3,4,2,3,3,4,3,4,4,5,
    1,2,2,3,2,3,3,4,2,3,3,4,3,4,4,5,
    2,3,3,4,3,4,4,5,3,4,4,5,4,5,5,6,
    1,2,2,3,2,3,3,4,2,3,3,4,3,4,4,5,
    2,3,3,4,3,4,4,5,3,4,4,5,4,5,5,6,
    2,3,3,4,3,4,4,5,3,4,4,5,4,5,5,6,
    3,4,4,5,4,5,5,6,4,5,5,6,5,6,6,7,
    1,2,2,3,2,3,3,4,2,3,3,4,3,4,4,5,
    2,3,3,4,3,4,4,5,3,4,4,5,4,5,5,6,
    2,3,3,4,3,4,4,5,3,4,4,5,4,5,5,6,
    3,4,4,5,4,5,5,6,4,5,5,6,5,6,6,7,
    2,3,3,4,3,4,4,5,3,4,4,5,4,5,5,6,
    3,4,4,5,4,5,5,6,4,5,5,6,5,6,6,7,
    3,4,4,5,4,5,5,6,4,5,5,6,5,6,6,7,
    4,5,5,6,5,6,6,7,5,6,6,7,6,7,7,8,
};



///////////////////////////////////////////////////////////
// GetNumberOfSetBits
//
// This function returns the number of '1' bits in bitmap
// beginning from FirstBit bit until nBits-1 position relative FirstBit
// nBits  - the number of available bits after FirstBit
///////////////////////////////////////////////////////////
size_t
GetNumberOfSetBits(
        IN const unsigned char* Map,
        IN size_t               nBits
        )
{
    size_t Ones = 0;

    // Middle bits
    size_t BytePos = nBits%8;
    nBits  /= 8;
#if 1
    const unsigned char* Last = Map + nBits;
    switch( nBits & 7 )
    {
    case 0:
        while( Map < Last )
        {
            Ones += BitsSetTotal[*Map++];
    case 7: Ones += BitsSetTotal[*Map++];
            case 6: Ones += BitsSetTotal[*Map++];
            case 5: Ones += BitsSetTotal[*Map++];
            case 4: Ones += BitsSetTotal[*Map++];
            case 3: Ones += BitsSetTotal[*Map++];
            case 2: Ones += BitsSetTotal[*Map++];
            case 1: Ones += BitsSetTotal[*Map++];
        }
    }
#else
    while( nBits-- )
        Ones += BitsSetTotal[*Map++];
#endif

    // Tail bits
    if ( 0 != BytePos )
        Ones += BitsSetTotal[(unsigned char)((*Map)<<(8-BytePos))];

    return Ones;
}



///////////////////////////////////////////////////////////
// OnAbout
//
//
///////////////////////////////////////////////////////////
static void
OnAbout()
{
    printf("Dump NTFS/HFS/Ext volumes (metafiles only) and collect system information.\n");
}

///////////////////////////////////////////////////////////
// OnUsage
//
// print a list of the parameters to the program
///////////////////////////////////////////////////////////
static void
OnUsage()
{
    OnAbout();

    printf("Paragon's Sysdump utility doesn't collect any personal or user-sensitive information. " \
           "Please refer to the 'Privacy policy' described in the Sysdump utility guide for more information.\n" \
           "Usage: sysdump [-o <filename>] <-s|device>\n" \
           "  -s     collect only system information\n" \
           "  -o     new file name (if name without suffix \".tar\", then suffix \".tar\" will be added automatically)\n" \
           "E.g. sysdump /dev/hdb1\n");
}


///////////////////////////////////////////////////////////
// parse_options
//
// Full trusted function // used to be
// Returns 0 if error
///////////////////////////////////////////////////////////
static bool
parse_options(
        IN int argc,
        IN const char *argv[],
        OUT const char**  szDevice,
        OUT bool& onlyInfo,
        OUT const char** sFilename
        )
{
    *szDevice = NULL;

    if (argc > 4)
    {
        fprintf(stderr, "Too many arguments\n");
        return false; // Force to call OnUsage
    }

    for (int i = 1; i < argc; ++i)
    {
        if (0 == strcmp("-s", argv[i]))
        {
            onlyInfo = true;
        }

        else if (0 == strcmp("-o", argv[i]))
        {
            if (i >= argc - 1)
            {
                fprintf(stderr, "Specify a file name after -o option\n");
                return false; // Force to call OnUsage
            }

            *sFilename = argv[i+1];
            ++i;
        }

        else
        {
            // The first "non options argument" is a device name
            if (NULL != *szDevice)
            {
                fprintf(stderr, "Error in argument %s\n", argv[i]);
                return false; // Force to call OnUsage
            }

            *szDevice = argv[i];
        }
    }

    if ((onlyInfo && (NULL != *szDevice)) || ((!onlyInfo) && (NULL == *szDevice)))
    {
        fprintf(stderr, "You can use either device name or -s option\n");
        return false; // Force to call OnUsage
    }

    return true;
}

// 2Mb
#define BYTES_PER_BUFFER    (2*1024*1024)

#define TARHEADER static_cast<PosixTarHeader*>(header)

struct PosixTarHeader
{
    char name[100];
    char mode[8];
    char uid[8];
    char gid[8];
    char size[12];
    char mtime[12];
    char checksum[8];
    char typeflag[1];
    char linkname[100];
    char magic[6];
    char version[2];
    char uname[32];
    char gname[32];
    char devmajor[8];
    char devminor[8];
    char prefix[155];
    char pad[12];
};

/**
* A Tar Archive
*/

class Tar
{
private:
    bool m_bFinished;
    const char* m_sCurrentFilename;
    unsigned long long m_nStartPosOfCurrentFile;
    int m_iArchFd;
    const char* m_sArchiveName;

protected:
    void _init(void* const header)
    {
        memset(header, 0, sizeof(PosixTarHeader));
        snprintf(TARHEADER->magic, sizeof(TARHEADER->magic), "ustar");
        sprintf(TARHEADER->mtime, "%011lo", time(NULL));
        sprintf(TARHEADER->mode, "%07o", 0644);

        char * s = ::getlogin();
        if(s!=NULL)
            snprintf((char*)TARHEADER, 32, "%s", s);
        sprintf(TARHEADER->gname, "users");
    }

    void _checksum(void* const header)
    {
        unsigned int sum = 0;
        char *p = (char*) header;
        char *q = p + sizeof(PosixTarHeader);

        while (p < TARHEADER->checksum)
            sum += *p++ & 0xff;

        for (int i = 0; i < 8; ++i)
        {
            sum += ' ';
            ++p;
        }

        while (p < q)
            sum += *p++ & 0xff;

        sprintf(TARHEADER->checksum,"%06o", sum);
    }

    void _size(void* const header, const unsigned long& fileSize)
    {
        sprintf(TARHEADER->size,"%011llo", (long long unsigned int)fileSize);
    }

    void _filename(void* const header, const char* const filename)
    {
        snprintf(TARHEADER->name, 100, "%s", filename);
    }

    bool _endRecord(size_t len) const
    {
        bool bRet = true;
        while((len % sizeof(PosixTarHeader))!= 0)
        {
            if (write(m_iArchFd, "\0", 1) < 0)
                bRet = false;
            ++len;
        }
        return bRet;
    }

public:
    Tar(const char* sFilename) :
        m_bFinished(false),
        m_sArchiveName(sFilename)
    {
    }

    virtual ~Tar()
    {
        if(!m_bFinished)
        {
        }
    }

    bool openArchive(const int& nFlags, const mode_t& mode)
    {
        m_iArchFd = open(m_sArchiveName, nFlags, mode);

        return -1 == m_iArchFd ? false : true;
    }

    void closeArchive()
    {
        close(m_iArchFd);
        m_iArchFd = -1;
    }

    // writes 2 empty blocks. Should be always called before closing the Tar file
    bool finish()
    {
        m_bFinished = true;
        bool bRet = true;
        //The end of the archive is indicated by two blocks filled with binary zeros
        PosixTarHeader header;
        memset((void*)&header, 0, sizeof(PosixTarHeader));

        for (int i = 0; i < 2; ++i)
            if (write(m_iArchFd, (const char*)&header, sizeof(PosixTarHeader)) < 0)
                bRet = false;

        return bRet;
    }

    bool addFileWithZeroHeader(const char* sNameInArchive)
    {
        bool bRet = true;
        m_sCurrentFilename = sNameInArchive;
        PosixTarHeader header;
        memset((void*)&header, 0, sizeof(PosixTarHeader));

        if (write(m_iArchFd, (const char*)&header, sizeof(PosixTarHeader)) < 0)
            bRet = false;
        m_nStartPosOfCurrentFile = lseek(m_iArchFd, 0, SEEK_CUR);

        return bRet;
    }

    bool addContentInCurrentFile(const char* const buf, const unsigned int& size) const
    {
        return write(m_iArchFd, buf, size) >= 0;
    }

    bool closeCurrentFile()
    {
        unsigned long long nFilesize = lseek(m_iArchFd, 0, SEEK_END) - m_nStartPosOfCurrentFile;
        bool bRet = true;

        PosixTarHeader header;
        _init((void*)&header);
        _filename((void*)&header, m_sCurrentFilename);
        header.typeflag[0 ]= 0;
        _size((void*)&header, nFilesize);
        _checksum((void*)&header);

        lseek(m_iArchFd, m_nStartPosOfCurrentFile - sizeof(PosixTarHeader), SEEK_SET);
        if (write(m_iArchFd, (const char*)&header, sizeof(PosixTarHeader)) < 0)
            bRet = false;

        lseek(m_iArchFd, 0, SEEK_END);
        _endRecord(nFilesize);

        return bRet;
    }
};

class StdString
{
private:
    size_t m_nLength;
    size_t m_nCapacity;
    char* m_aStr;
    const Tar* m_archiveForWrite;

    StdString& appendStr(const StdString& str)
    {
        if ((str.m_nLength + 1) > (m_nCapacity - m_nLength))
        {
            if (NULL == m_archiveForWrite)
            {
                m_nCapacity += m_nCapacity > str.m_nLength ? m_nCapacity : str.m_nLength;
                m_aStr = (char*)realloc(m_aStr, m_nCapacity);
                if (NULL == m_aStr)
                {
                    perror("Realloc");
                    exit(1);
                }
            }
            else
            {
                m_archiveForWrite->addContentInCurrentFile(m_aStr, m_nLength);
                m_nLength = 0;
            }
        }
        strcpy(m_aStr + m_nLength, str.m_aStr);
        m_nLength += str.m_nLength;
        return *this;
    }

    StdString(const StdString& str);
    StdString & operator = (const StdString& str);


public:
    StdString(const size_t& nReserve, const Tar* arhive = NULL) :
        m_nLength(0),
        m_nCapacity(nReserve),
        m_archiveForWrite(arhive)
    {
        m_aStr = (char*)malloc(nReserve);
        if (NULL == m_aStr)
        {
            perror("Malloc");
            exit(1);
        }
    }

    StdString(const char* const str) :
        m_nLength(strlen(str)),
        m_nCapacity(strlen(str) + 1),
        m_archiveForWrite(NULL)
    {
        m_aStr = (char*)malloc(m_nLength + 1);
        if (NULL == m_aStr)
        {
            perror("Malloc");
            exit(1);
        }
        strcpy(m_aStr, str);
    }

    StdString& operator += (const StdString& str)
    {
        return appendStr(str);
    }

    StdString& operator + (const StdString& str)
    {
        return appendStr(str);
    }

    const size_t& length() const
    {
        return m_nLength;
    }

    const char* c_str() const
    {
        return m_aStr;
    }

    ~StdString()
    {
        free(m_aStr);
    }
};

class FileDumper;
class ListOfPartitions;

class Partition
{
    friend class FileDumper;
    friend class ListOfPartitions;

private:
    char       m_sFilesystem[128];
    char       m_sType[128];
    char       m_sMountedOn[128];
    char       m_sSize[128];
    char       m_sAvail[128];
    Partition* m_pNextPart;

    void getHumanReadableSize(const unsigned long long& nSize, char* sSize, size_t nSizeOfStr) const
    {
        double dblSize = nSize;
        int nCount = 0;

        while (dblSize >= 1024.0)
        {
            ++nCount;
            dblSize /= 1024.0;

            if (nCount == 4)
                break;
        }

        snprintf(sSize, nSizeOfStr, "%.1f", dblSize);

        const char* prfx[] = {"K", "M", "G", "T"};
        if ((nCount >=1) && (nCount <= 4) && (strlen(sSize) + 1 < nSizeOfStr))
            strcat(sSize, prfx[nCount -1]);
    }

public:

    Partition(const char*          sFilesystem,
              const char*          sType,
              const char*          sMountedOn,
              const struct statfs&   statBuf    ) :

        m_pNextPart    (NULL)
    {
        strncpy(m_sFilesystem, sFilesystem, sizeof(m_sFilesystem) - 1);
        m_sFilesystem[sizeof(m_sFilesystem) - 1] = '\0';
        strncpy(m_sType, sType, sizeof(m_sType) - 1);
        m_sType[sizeof(m_sType) - 1] = '\0';
        strncpy(m_sMountedOn, sMountedOn, sizeof(m_sMountedOn) - 1);
        m_sMountedOn[sizeof(m_sMountedOn) - 1] = '\0';
        getHumanReadableSize(statBuf.f_bsize * statBuf.f_blocks, m_sSize, sizeof(m_sSize));
        getHumanReadableSize(statBuf.f_bsize * statBuf.f_bavail, m_sAvail, sizeof(m_sAvail));
    }
};

class ListOfPartitions
{
    friend class FileDumper;

private:
    Partition* m_pFirstPart;
    Partition* m_pLastPart;

    bool isEmpty() const
    {
        return NULL == m_pFirstPart;
    }

public:
    ListOfPartitions() :
        m_pFirstPart(NULL),
        m_pLastPart(NULL)
    {
    }

    void pushBack(const Partition& part)
    {
        Partition* pPart = new Partition(part);

        if (isEmpty())
            m_pFirstPart = m_pLastPart = pPart;
        else
        {
            m_pLastPart->m_pNextPart = pPart;
            m_pLastPart = pPart;
        }
    }

    ~ListOfPartitions()
    {
        Partition* pPart = m_pFirstPart;
        while(NULL != pPart)
        {
            Partition* pNxtPart = pPart->m_pNextPart;
            delete pPart;
            pPart = pNxtPart;
        }
    }
};

class ListOfStrings
{
    friend class FileDumper;

private:
    class NodeStr
    {
        friend class FileDumper;
        friend class ListOfStrings;

    private:
        char* m_sStr;
        NodeStr* m_pNextNode;
        size_t m_nLength;

    public:
        NodeStr(const char* sStr) :
            m_pNextNode(NULL),
            m_nLength(strlen(sStr))
        {
            m_sStr = (char*)malloc(m_nLength + 1);
            strcpy(m_sStr, sStr);
        }

        ~NodeStr()
        {
            free(m_sStr);
        }
    };

    NodeStr* m_pFirstStr;
    NodeStr* m_pLastStr;

    bool isEmpty() const
    {
        return NULL == m_pFirstStr;
    }

public:
    ListOfStrings() :
        m_pFirstStr(NULL),
        m_pLastStr(NULL)
    {
    }

    void pushBack(const char* sStr)
    {
        NodeStr* pStr = new NodeStr(sStr);

        if (isEmpty())
            m_pFirstStr = m_pLastStr = pStr;
        else
        {
            m_pLastStr->m_pNextNode = pStr;
            m_pLastStr = pStr;
        }
    }

    ~ListOfStrings()
    {
        NodeStr* pStr = m_pFirstStr;
        while(NULL != pStr)
        {
            NodeStr* pNxtStr = pStr->m_pNextNode;
            delete pStr;
            pStr = pNxtStr;
        }
    }
};

struct hd_geometry {
    unsigned char heads;
    unsigned char sectors;
    unsigned short cylinders;	/* truncated */
    unsigned long start;
};

#ifndef HDIO_GETGEO
#  define HDIO_GETGEO 0x0301
#endif

#define FD_BUFSIZE 1024

class FileDumper
{
private:
    Tar m_tarball;
    ListOfStrings m_listOfMountPoints;
    bool m_bIsAndroid;

public:
    FileDumper(const char* sFileName) :
        m_tarball(sFileName),
        m_bIsAndroid(false)
    {
    }

    bool openArchive(const int& nFlags, const mode_t& mode = 0)
    {
        return m_tarball.openArchive(nFlags, mode);
    }

    void closeArchive()
    {
        m_tarball.closeArchive();
    }

    bool appendFileContent(const char* const sReadingFile) const
    {
        StdString sOut(FD_BUFSIZE);

        sOut += (StdString)"\n\n#################### Content of file " + sReadingFile + ": ####################\n\n";

        int iFd = open(sReadingFile, O_RDONLY);
        if (-1 == iFd)
        {
            sOut += (StdString)"File " + sReadingFile + " could not be opened: " + strerror(errno) + "\n";
            m_tarball.addContentInCurrentFile(sOut.c_str(), sOut.length());
            return false;
        }

        m_tarball.addContentInCurrentFile(sOut.c_str(), sOut.length());

        char aBuf[FD_BUFSIZE];
        size_t nNumOfRead;
        while ((nNumOfRead = read(iFd, aBuf, FD_BUFSIZE)) > 0)
            m_tarball.addContentInCurrentFile(aBuf, nNumOfRead);

        close(iFd);
        return true;
    }

    void appendKernelInfo() const
    {
        struct utsname unameData;
        uname(&unameData);

        StdString sOut(FD_BUFSIZE);

        sOut += (StdString)"\n\n#################### Kernel Info: ####################\n\n" \
                + "NIS or YP domain name:                              " + unameData.domainname + "\n" \
                + "Hardware identifier:                                " + unameData.machine + "\n" \
                + "Name within \"some implementation-defined network\":  " + unameData.nodename + "\n" \
                + "Operating system release:                           " + unameData.release + "\n" \
                + "Operating system name:                              " + unameData.sysname + "\n" \
                + "Operating system version:                           " + unameData.version + "\n";

        m_tarball.addContentInCurrentFile(sOut.c_str(), sOut.length());
    }

    bool appendDiskSpace()
    {
        char sFileSystem[128];
        char sType[128];
        char sMountedOn[128];

        unsigned nLongestFileSistem = strlen("Filesystem");
        unsigned nLongestType = strlen("Type");
        unsigned nLongestSize = strlen("Size");
        unsigned nLongestAvail = strlen("Available");

        FILE *fp;
        if (NULL == (fp = fopen("/proc/mounts", "r")))
            return false;

        ListOfPartitions listOfParts;

        while (fscanf(fp, "%s%s%s%*s%*s%*s", sFileSystem, sMountedOn, sType) != EOF)
        {
            struct statfs buf;
            statfs(sMountedOn, &buf);

            if (buf.f_blocks)
            {
                Partition part(sFileSystem, sType, sMountedOn, buf);
                listOfParts.pushBack(part);

                if (strlen(part.m_sFilesystem) > nLongestFileSistem)
                    nLongestFileSistem = strlen(part.m_sFilesystem);

                if (strlen(part.m_sType) > nLongestType)
                    nLongestType = strlen(part.m_sType);

                if (strlen(part.m_sSize) > nLongestSize)
                    nLongestSize = strlen(part.m_sSize);

                if (strlen(part.m_sAvail) > nLongestAvail)
                    nLongestAvail = strlen(part.m_sAvail);
            }

            if (m_bIsAndroid)
            {
                if ((0 != strcmp(sMountedOn, "/")) && (0 != strcmp(sMountedOn, "/system")))
                    m_listOfMountPoints.pushBack(sMountedOn);
            }
            else
                if (0 != strcmp(sMountedOn, "/"))
                    m_listOfMountPoints.pushBack(sMountedOn);
        }
        fclose(fp);

        StdString sOut(FD_BUFSIZE, &m_tarball);
        sOut += "\n\n#################### Disk space: ####################\n\n";

        char aBuf[FD_BUFSIZE];
        snprintf(aBuf, FD_BUFSIZE, "%-*s%-*s%*s%*s %-s\n\n", nLongestFileSistem + 1, "Filesystem", nLongestType + 1, "Type", nLongestSize, "Size", nLongestAvail + 1, "Available", "Mounted on");
        sOut += aBuf;

        Partition* pPart = listOfParts.m_pFirstPart;
        while(NULL != pPart)
        {
            snprintf(aBuf, FD_BUFSIZE, "%-*s%-*s%*s%*s %-s\n", nLongestFileSistem + 1, pPart->m_sFilesystem, nLongestType + 1, pPart->m_sType, nLongestSize, pPart->m_sSize, nLongestAvail + 1, pPart->m_sAvail, pPart->m_sMountedOn);
            sOut += aBuf;
            pPart = pPart->m_pNextPart;
        }

        m_tarball.addContentInCurrentFile(sOut.c_str(), sOut.length());
        return true;
    }

    void appendDirectoryContent(const char* const sDir) const
    {
        StdString sOut(FD_BUFSIZE, &m_tarball);
        sOut += (StdString)"\n\n#################### Content of directory " + sDir + ": ####################\n\n";

        ListOfStrings listOfFilenames;
        getDirectoryContent(sDir, listOfFilenames, &sOut);

        ListOfStrings::NodeStr* pStr = listOfFilenames.m_pFirstStr;
        while (NULL != pStr)
        {
            sOut += (StdString)pStr->m_sStr + "\n";
            pStr = pStr->m_pNextNode;
        }

        m_tarball.addContentInCurrentFile(sOut.c_str(), sOut.length());
    }

    void appendModulesList() const
    {
        const char* sTitle = "\n\n#################### Found modules: ####################\n\n";
        m_tarball.addContentInCurrentFile(sTitle, strlen(sTitle));

        ListOfStrings listOfModules;
        char aDir[4096];
        strcpy(aDir, "/");
        searchModulesRecursive(aDir, sizeof(aDir), listOfModules);

        char sModules[FD_BUFSIZE];
        size_t nOffset = 0;

        ListOfStrings::NodeStr* pStr = listOfModules.m_pFirstStr;
        while (NULL != pStr)
        {
            if ((nOffset + pStr->m_nLength + 1) <= FD_BUFSIZE)
            {
                memcpy(sModules + nOffset, pStr->m_sStr, pStr->m_nLength);
                nOffset += pStr->m_nLength + 1;
                sModules[nOffset - 1] = '\n';
                pStr = pStr->m_pNextNode;
            }
            else
            {
                m_tarball.addContentInCurrentFile(sModules, nOffset);
                nOffset = 0;
            }
        }

        m_tarball.addContentInCurrentFile(sModules, nOffset);
    }

    void appendCommandOutput(const char* const sCommand) const
    {
        StdString sOut(FD_BUFSIZE, &m_tarball);
        sOut += (StdString)"\n\n#################### Command " + sCommand + ": ####################\n\n";

        getCommandOutput(sCommand, sOut);
        m_tarball.addContentInCurrentFile(sOut.c_str(), sOut.length());
    }

    void appendLocationAndVersionOfBinary() const
    {
        StdString sOut(FD_BUFSIZE);
        sOut += "\n\n#################### Locations and versions of binary executable files: ####################\n\n";

        sOut += "busibox: ";
        getCommandOutput("which busybox", sOut);
        sOut += "         ";
        StdString sBusyboxOut(FD_BUFSIZE);
        sBusyboxOut += "\0";
        getCommandOutput("busybox", sBusyboxOut);
        const char* sBusyboxVersion = strtok((char*)sBusyboxOut.c_str(), "\n");
        if (sBusyboxVersion != NULL)
            sOut += (StdString)sBusyboxVersion + "\n";
        sOut += "\n";

        sOut += "smbd:    ";
        getCommandOutput("which smbd", sOut);
        sOut += "         ";
        getCommandOutput("smbd --version", sOut);
        sOut += "\n";

        sOut += "afpd:    ";
        getCommandOutput("which afpd", sOut);
        sOut += "         ";
        getCommandOutput("afpd -v", sOut);
        sOut += "\n";

        m_tarball.addContentInCurrentFile(sOut.c_str(), sOut.length());
    }

    void addTarfileWithZeroHeader(const char* const sNameInArchive)
    {
        m_tarball.addFileWithZeroHeader(sNameInArchive);
    }

    void closeCurrentTarfile()
    {
        m_tarball.closeCurrentFile();
    }

    void finishTarball()
    {
        m_tarball.finish();
    }

    void writeMd5(const char* const ar, const int& nSize) const
    {
        char aMd5[nSize + 1];
        memcpy(aMd5, ar, nSize);
        aMd5[nSize] = '\n';
        m_tarball.addContentInCurrentFile(aMd5, nSize + 1);
    }

    bool copyFile(const char* const sFilename) const
    {
        int iFd = open(sFilename, O_RDONLY);
        if (-1 == iFd)
            return false;

        char aBuf[FD_BUFSIZE];
        size_t nRead;
        while ((nRead = read(iFd, aBuf, FD_BUFSIZE)) > 0)
            m_tarball.addContentInCurrentFile(aBuf, nRead);

        close(iFd);
        return true;
    }

    bool findFileInDirectory(const char* const sFile, const char* const sDirectory) const
    {
        ListOfStrings listOfNames;
        getDirectoryContent(sDirectory, listOfNames, 0);

        ListOfStrings::NodeStr* pStr = listOfNames.m_pFirstStr;
        while (NULL != pStr)
        {
            if (0 == strcmp(pStr->m_sStr, sFile))
                return true;

            pStr = pStr->m_pNextNode;
        }

        return false;
    }

    void setAndroidOs(const bool& bIsAndroid)
    {
        m_bIsAndroid = bIsAndroid;
    }

    bool isAndroid()
    {
        return m_bIsAndroid;
    }

    void appendFdiskAndroid()
    {
        StdString sOut(FD_BUFSIZE, &m_tarball);
        sOut += "\n\n#################### Command fdisk -l DISK: ####################\n\n";

        FILE *procpt;
        char line[128 + 1], ptname[128 + 1], devname[256];
        int ma, mi;
        unsigned long long sz;
        char sFdiskCommand[256];

        procpt = fopen("/proc/partitions", "r");
        if (procpt == NULL)
        {
            sOut += (StdString)"File /proc/partitions could not be opened: " + strerror(errno) + "\n";
            m_tarball.addContentInCurrentFile(sOut.c_str(), sOut.length());
            return;
        }

        while (fgets(line, sizeof(line), procpt))
        {
            if (sscanf (line, " %d %d %llu %128[^\n ]", &ma, &mi, &sz, ptname) != 4)
                continue;

            snprintf(devname, sizeof(devname), "/dev/block/%s", ptname);

            int fd = open(devname, O_RDONLY);
            if (fd == -1)
                continue;

            struct hd_geometry geometry;
            int i = ioctl(fd, HDIO_GETGEO, &geometry);
            close(fd);
            if (i != 0)
                continue;

            if (geometry.start == 0)
            {
                snprintf(sFdiskCommand, sizeof(sFdiskCommand), "fdisk -l %s", devname);
                getCommandOutput(sFdiskCommand, sOut);
                sOut += "\n";
            }
        }

        fclose(procpt);
        m_tarball.addContentInCurrentFile(sOut.c_str(), sOut.length());
    }

private:
    bool getDirectoryContent(const char* const sDirectory, ListOfStrings& listOfFilenames, StdString* sOut) const
    {
        DIR * dir;
        struct dirent * de;

        if ((dir = opendir(sDirectory)) == NULL)
        {
            if (NULL != sOut)
                (*sOut) += (StdString)"Directory " + sDirectory + " could not be opened: " + strerror(errno) + "\n";

            return false;
        }

        while ((de = readdir(dir)) != NULL)
        {
            const char* sName = de->d_name;
            if ((0 != strcmp(".", sName)) && (0 != strcmp("..", sName)))
                listOfFilenames.pushBack(sName);
        }

        closedir(dir);
        return true;
    }

    bool getCommandOutput(const char* const sCommand, StdString& sOut) const
    {
        char sCmd[256];
        char buf[FD_BUFSIZE];
        FILE* f;
        char sTmpFile[16];

        if (m_bIsAndroid)
        {
            sprintf(sTmpFile, "tmpXXXXXX");
            int nTmpFd = mkstemp(sTmpFile);

            if (-1 == nTmpFd)
                return false;

            close(nTmpFd);

            snprintf(sCmd, 256, "%s > %s 2>&1", sCommand, sTmpFile);
            if ( -1 == system(sCmd))
            {
                sOut += (StdString)"Command " + sCommand + " could not be called\n";
                return false;
            }

            if ((f = fopen(sTmpFile,"r")) == NULL)
                return false;
        }
        else
        {
            snprintf(sCmd, 256, "%s 2>&1", sCommand);

            if ((f = popen(sCmd,"r")) == NULL)
            {
                sOut += (StdString)"Command " + sCommand + " could not be called\n";
                return false;
            }
        }

        while (fgets(buf, FD_BUFSIZE -1, f))
            sOut += buf;

        fclose(f);

        if (m_bIsAndroid)
            unlink(sTmpFile);

        return true;
    }

    void searchModulesRecursive(char* const sPath, const size_t& nSizeOfPath, ListOfStrings& listOfModules) const
    {
        DIR * dir;

        if ((dir = opendir(sPath)) == NULL)
            return;

        if (0 != strcmp("/", sPath))
            strncat(sPath, "/", nSizeOfPath - 2);
        size_t nLengthOfDir = strlen(sPath);

        struct dirent * de;
        while ((de = readdir(dir)) != NULL)
        {
            const char* sName = de->d_name;
            if ((0 == strcmp(".", sName)) || (0 == strcmp("..", sName)))
                continue;

            size_t nSizeOfName = strlen(sName);
            strncat(sPath, sName, nSizeOfPath - nSizeOfName - 1);

            struct stat bufInfo;
            lstat(sPath, &bufInfo);

            if (S_ISDIR(bufInfo.st_mode))
            {
                bool bIsMatch = false;
                ListOfStrings::NodeStr* pMountPoint = m_listOfMountPoints.m_pFirstStr;
                while (NULL != pMountPoint)
                {
                    if (0 == strcmp(pMountPoint->m_sStr, sPath))
                    {
                        bIsMatch = true;
                        break;
                    }
                    pMountPoint = pMountPoint->m_pNextNode;
                }

                if (!bIsMatch)
                    searchModulesRecursive(sPath, nSizeOfPath, listOfModules);
            }
            else if (S_ISREG(bufInfo.st_mode))
            {
                if (nSizeOfName >= 3)
                    if ((sName[nSizeOfName - 3] == '.') && (sName[nSizeOfName - 2] == 'k') && (sName[nSizeOfName - 1] == 'o'))
                        listOfModules.pushBack(sPath);
            }
            sPath[nLengthOfDir] = '\0';
        }

        closedir(dir);
        return;
    }
};

typedef unsigned char md5_byte_t; /* 8-bit byte */
typedef unsigned int md5_word_t; /* 32-bit word */

/* Define the state of the MD5 Algorithm. */
typedef struct md5_state_s {
    md5_word_t count[2];	/* message length in bits, lsw first */
    md5_word_t abcd[4];		/* digest buffer */
    md5_byte_t buf[64];		/* accumulate block */
} md5_state_t;

#undef BYTE_ORDER	/* 1 = big-endian, -1 = little-endian, 0 = unknown */
#ifdef ARCH_IS_BIG_ENDIAN
#  define BYTE_ORDER (ARCH_IS_BIG_ENDIAN ? 1 : -1)
#else
#  define BYTE_ORDER 0
#endif

#define T_MASK ((md5_word_t)~0)
#define T1 /* 0xd76aa478 */ (T_MASK ^ 0x28955b87)
#define T2 /* 0xe8c7b756 */ (T_MASK ^ 0x173848a9)
#define T3    0x242070db
#define T4 /* 0xc1bdceee */ (T_MASK ^ 0x3e423111)
#define T5 /* 0xf57c0faf */ (T_MASK ^ 0x0a83f050)
#define T6    0x4787c62a
#define T7 /* 0xa8304613 */ (T_MASK ^ 0x57cfb9ec)
#define T8 /* 0xfd469501 */ (T_MASK ^ 0x02b96afe)
#define T9    0x698098d8
#define T10 /* 0x8b44f7af */ (T_MASK ^ 0x74bb0850)
#define T11 /* 0xffff5bb1 */ (T_MASK ^ 0x0000a44e)
#define T12 /* 0x895cd7be */ (T_MASK ^ 0x76a32841)
#define T13    0x6b901122
#define T14 /* 0xfd987193 */ (T_MASK ^ 0x02678e6c)
#define T15 /* 0xa679438e */ (T_MASK ^ 0x5986bc71)
#define T16    0x49b40821
#define T17 /* 0xf61e2562 */ (T_MASK ^ 0x09e1da9d)
#define T18 /* 0xc040b340 */ (T_MASK ^ 0x3fbf4cbf)
#define T19    0x265e5a51
#define T20 /* 0xe9b6c7aa */ (T_MASK ^ 0x16493855)
#define T21 /* 0xd62f105d */ (T_MASK ^ 0x29d0efa2)
#define T22    0x02441453
#define T23 /* 0xd8a1e681 */ (T_MASK ^ 0x275e197e)
#define T24 /* 0xe7d3fbc8 */ (T_MASK ^ 0x182c0437)
#define T25    0x21e1cde6
#define T26 /* 0xc33707d6 */ (T_MASK ^ 0x3cc8f829)
#define T27 /* 0xf4d50d87 */ (T_MASK ^ 0x0b2af278)
#define T28    0x455a14ed
#define T29 /* 0xa9e3e905 */ (T_MASK ^ 0x561c16fa)
#define T30 /* 0xfcefa3f8 */ (T_MASK ^ 0x03105c07)
#define T31    0x676f02d9
#define T32 /* 0x8d2a4c8a */ (T_MASK ^ 0x72d5b375)
#define T33 /* 0xfffa3942 */ (T_MASK ^ 0x0005c6bd)
#define T34 /* 0x8771f681 */ (T_MASK ^ 0x788e097e)
#define T35    0x6d9d6122
#define T36 /* 0xfde5380c */ (T_MASK ^ 0x021ac7f3)
#define T37 /* 0xa4beea44 */ (T_MASK ^ 0x5b4115bb)
#define T38    0x4bdecfa9
#define T39 /* 0xf6bb4b60 */ (T_MASK ^ 0x0944b49f)
#define T40 /* 0xbebfbc70 */ (T_MASK ^ 0x4140438f)
#define T41    0x289b7ec6
#define T42 /* 0xeaa127fa */ (T_MASK ^ 0x155ed805)
#define T43 /* 0xd4ef3085 */ (T_MASK ^ 0x2b10cf7a)
#define T44    0x04881d05
#define T45 /* 0xd9d4d039 */ (T_MASK ^ 0x262b2fc6)
#define T46 /* 0xe6db99e5 */ (T_MASK ^ 0x1924661a)
#define T47    0x1fa27cf8
#define T48 /* 0xc4ac5665 */ (T_MASK ^ 0x3b53a99a)
#define T49 /* 0xf4292244 */ (T_MASK ^ 0x0bd6ddbb)
#define T50    0x432aff97
#define T51 /* 0xab9423a7 */ (T_MASK ^ 0x546bdc58)
#define T52 /* 0xfc93a039 */ (T_MASK ^ 0x036c5fc6)
#define T53    0x655b59c3
#define T54 /* 0x8f0ccc92 */ (T_MASK ^ 0x70f3336d)
#define T55 /* 0xffeff47d */ (T_MASK ^ 0x00100b82)
#define T56 /* 0x85845dd1 */ (T_MASK ^ 0x7a7ba22e)
#define T57    0x6fa87e4f
#define T58 /* 0xfe2ce6e0 */ (T_MASK ^ 0x01d3191f)
#define T59 /* 0xa3014314 */ (T_MASK ^ 0x5cfebceb)
#define T60    0x4e0811a1
#define T61 /* 0xf7537e82 */ (T_MASK ^ 0x08ac817d)
#define T62 /* 0xbd3af235 */ (T_MASK ^ 0x42c50dca)
#define T63    0x2ad7d2bb
#define T64 /* 0xeb86d391 */ (T_MASK ^ 0x14792c6e)

static void
md5_process(md5_state_t *pms, const md5_byte_t *data /*[64]*/)
{
    md5_word_t
            a = pms->abcd[0], b = pms->abcd[1],
            c = pms->abcd[2], d = pms->abcd[3];
    md5_word_t t;
#if BYTE_ORDER > 0
    /* Define storage only for big-endian CPUs. */
    md5_word_t X[16];
#else
    /* Define storage for little-endian or both types of CPUs. */
    md5_word_t xbuf[16];
    const md5_word_t *X;
#endif

    {
#if BYTE_ORDER == 0
        /*
     * Determine dynamically whether this is a big-endian or
     * little-endian machine, since we can use a more efficient
     * algorithm on the latter.
     */
        static const int w = 1;

        if (*((const md5_byte_t *)&w)) /* dynamic little-endian */
#endif
#if BYTE_ORDER <= 0		/* little-endian */
        {
            /*
         * On little-endian machines, we can process properly aligned
         * data without copying it.
         */
            if (!((data - (const md5_byte_t *)0) & 3)) {
                /* data are properly aligned */
                X = (const md5_word_t *)data;
            } else {
                /* not aligned */
                memcpy(xbuf, data, 64);
                X = xbuf;
            }
        }
#endif
#if BYTE_ORDER == 0
        else			/* dynamic big-endian */
#endif
#if BYTE_ORDER >= 0		/* big-endian */
        {
            /*
         * On big-endian machines, we must arrange the bytes in the
         * right order.
         */
            const md5_byte_t *xp = data;
            int i;

#  if BYTE_ORDER == 0
            X = xbuf;		/* (dynamic only) */
#  else
#    define xbuf X		/* (static only) */
#  endif
            for (i = 0; i < 16; ++i, xp += 4)
                xbuf[i] = xp[0] + (xp[1] << 8) + (xp[2] << 16) + (xp[3] << 24);
        }
#endif
    }

#define ROTATE_LEFT(x, n) (((x) << (n)) | ((x) >> (32 - (n))))

    /* Round 1. */
    /* Let [abcd k s i] denote the operation
       a = b + ((a + F(b,c,d) + X[k] + T[i]) <<< s). */
#define F(x, y, z) (((x) & (y)) | (~(x) & (z)))
#define SET(a, b, c, d, k, s, Ti)\
    t = a + F(b,c,d) + X[k] + Ti;\
    a = ROTATE_LEFT(t, s) + b
    /* Do the following 16 operations. */
    SET(a, b, c, d,  0,  7,  T1);
    SET(d, a, b, c,  1, 12,  T2);
    SET(c, d, a, b,  2, 17,  T3);
    SET(b, c, d, a,  3, 22,  T4);
    SET(a, b, c, d,  4,  7,  T5);
    SET(d, a, b, c,  5, 12,  T6);
    SET(c, d, a, b,  6, 17,  T7);
    SET(b, c, d, a,  7, 22,  T8);
    SET(a, b, c, d,  8,  7,  T9);
    SET(d, a, b, c,  9, 12, T10);
    SET(c, d, a, b, 10, 17, T11);
    SET(b, c, d, a, 11, 22, T12);
    SET(a, b, c, d, 12,  7, T13);
    SET(d, a, b, c, 13, 12, T14);
    SET(c, d, a, b, 14, 17, T15);
    SET(b, c, d, a, 15, 22, T16);
#undef SET

    /* Round 2. */
    /* Let [abcd k s i] denote the operation
          a = b + ((a + G(b,c,d) + X[k] + T[i]) <<< s). */
#define G(x, y, z) (((x) & (z)) | ((y) & ~(z)))
#define SET(a, b, c, d, k, s, Ti)\
    t = a + G(b,c,d) + X[k] + Ti;\
    a = ROTATE_LEFT(t, s) + b
    /* Do the following 16 operations. */
    SET(a, b, c, d,  1,  5, T17);
    SET(d, a, b, c,  6,  9, T18);
    SET(c, d, a, b, 11, 14, T19);
    SET(b, c, d, a,  0, 20, T20);
    SET(a, b, c, d,  5,  5, T21);
    SET(d, a, b, c, 10,  9, T22);
    SET(c, d, a, b, 15, 14, T23);
    SET(b, c, d, a,  4, 20, T24);
    SET(a, b, c, d,  9,  5, T25);
    SET(d, a, b, c, 14,  9, T26);
    SET(c, d, a, b,  3, 14, T27);
    SET(b, c, d, a,  8, 20, T28);
    SET(a, b, c, d, 13,  5, T29);
    SET(d, a, b, c,  2,  9, T30);
    SET(c, d, a, b,  7, 14, T31);
    SET(b, c, d, a, 12, 20, T32);
#undef SET

    /* Round 3. */
    /* Let [abcd k s t] denote the operation
          a = b + ((a + H(b,c,d) + X[k] + T[i]) <<< s). */
#define H(x, y, z) ((x) ^ (y) ^ (z))
#define SET(a, b, c, d, k, s, Ti)\
    t = a + H(b,c,d) + X[k] + Ti;\
    a = ROTATE_LEFT(t, s) + b
    /* Do the following 16 operations. */
    SET(a, b, c, d,  5,  4, T33);
    SET(d, a, b, c,  8, 11, T34);
    SET(c, d, a, b, 11, 16, T35);
    SET(b, c, d, a, 14, 23, T36);
    SET(a, b, c, d,  1,  4, T37);
    SET(d, a, b, c,  4, 11, T38);
    SET(c, d, a, b,  7, 16, T39);
    SET(b, c, d, a, 10, 23, T40);
    SET(a, b, c, d, 13,  4, T41);
    SET(d, a, b, c,  0, 11, T42);
    SET(c, d, a, b,  3, 16, T43);
    SET(b, c, d, a,  6, 23, T44);
    SET(a, b, c, d,  9,  4, T45);
    SET(d, a, b, c, 12, 11, T46);
    SET(c, d, a, b, 15, 16, T47);
    SET(b, c, d, a,  2, 23, T48);
#undef SET

    /* Round 4. */
    /* Let [abcd k s t] denote the operation
          a = b + ((a + I(b,c,d) + X[k] + T[i]) <<< s). */
#define I(x, y, z) ((y) ^ ((x) | ~(z)))
#define SET(a, b, c, d, k, s, Ti)\
    t = a + I(b,c,d) + X[k] + Ti;\
    a = ROTATE_LEFT(t, s) + b
    /* Do the following 16 operations. */
    SET(a, b, c, d,  0,  6, T49);
    SET(d, a, b, c,  7, 10, T50);
    SET(c, d, a, b, 14, 15, T51);
    SET(b, c, d, a,  5, 21, T52);
    SET(a, b, c, d, 12,  6, T53);
    SET(d, a, b, c,  3, 10, T54);
    SET(c, d, a, b, 10, 15, T55);
    SET(b, c, d, a,  1, 21, T56);
    SET(a, b, c, d,  8,  6, T57);
    SET(d, a, b, c, 15, 10, T58);
    SET(c, d, a, b,  6, 15, T59);
    SET(b, c, d, a, 13, 21, T60);
    SET(a, b, c, d,  4,  6, T61);
    SET(d, a, b, c, 11, 10, T62);
    SET(c, d, a, b,  2, 15, T63);
    SET(b, c, d, a,  9, 21, T64);
#undef SET

    /* Then perform the following additions. (That is increment each
        of the four registers by the value it had before this block
        was started.) */
    pms->abcd[0] += a;
    pms->abcd[1] += b;
    pms->abcd[2] += c;
    pms->abcd[3] += d;
}

void
md5_init(md5_state_t *pms)
{
    pms->count[0] = pms->count[1] = 0;
    pms->abcd[0] = 0x67452301;
    pms->abcd[1] = /*0xefcdab89*/ T_MASK ^ 0x10325476;
    pms->abcd[2] = /*0x98badcfe*/ T_MASK ^ 0x67452301;
    pms->abcd[3] = 0x10325476;
}

void
md5_append(md5_state_t *pms, const md5_byte_t *data, int nbytes)
{
    const md5_byte_t *p = data;
    int left = nbytes;
    int offset = (pms->count[0] >> 3) & 63;
    md5_word_t nbits = (md5_word_t)(nbytes << 3);

    if (nbytes <= 0)
        return;

    /* Update the message length. */
    pms->count[1] += nbytes >> 29;
    pms->count[0] += nbits;
    if (pms->count[0] < nbits)
        pms->count[1]++;

    /* Process an initial partial block. */
    if (offset) {
        int copy = (offset + nbytes > 64 ? 64 - offset : nbytes);

        memcpy(pms->buf + offset, p, copy);
        if (offset + copy < 64)
            return;
        p += copy;
        left -= copy;
        md5_process(pms, pms->buf);
    }

    /* Process full blocks. */
    for (; left >= 64; p += 64, left -= 64)
        md5_process(pms, p);

    /* Process a final partial block. */
    if (left)
        memcpy(pms->buf, p, left);
}

void
md5_finish(md5_state_t *pms, md5_byte_t digest[16])
{
    static const md5_byte_t pad[64] = {
        0x80, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
    };
    md5_byte_t data[8];
    int i;

    /* Save the length before padding. */
    for (i = 0; i < 8; ++i)
        data[i] = (md5_byte_t)(pms->count[i >> 2] >> ((i & 3) << 3));
    /* Pad to 56 bytes mod 64. */
    md5_append(pms, pad, ((55 - (pms->count[0] >> 3)) & 63) + 1);
    /* Append the length. */
    md5_append(pms, data, 8);
    for (i = 0; i < 16; ++i)
        digest[i] = (md5_byte_t)(pms->abcd[i >> 2] >> ((i & 3) << 3));
}

///////////////////////////////////////////////////////////
// main
//
//
///////////////////////////////////////////////////////////
int __cdecl
main(
        IN int argc,
        IN const char* argv[]
        )
{
    unsigned Status = ERR_BADPARAMS;
    const char* szDevice = NULL;
    CRWBlock* Rw = NULL;
    int Device  = -1;
    unsigned char* Bitmap = NULL;
    char* Buffer = NULL;
    unsigned int BytesPerBitmap = 0;
    unsigned int ClustersPerBuffer = 0;
    unsigned int ClustersPerVolume;
    unsigned int Log2OfCluster;
    unsigned int Gb, Mb, GbDev, MbDev;
    const char* hint;

    setlocale(LC_MESSAGES, "");
    setlocale(LC_CTYPE, "");
    bindtextdomain(NLS_CAT_NAME, LOCALEDIR);
    textdomain(NLS_CAT_NAME);

    if (getuid() != 0)
    {
        printf("root privileges needed\n");
        return -1;
    }

    if ( argc < 1 )
        return -1;  // Impossible

    if (1 == argc)
    {
        OnUsage();
        return 0;
    }

    // Parse command line
    // And exit if there are errors
    bool fOnlyInfo = false;
    const char* sEnteredArchiveName = NULL;
    if ( !parse_options( argc, argv, &szDevice, fOnlyInfo, &sEnteredArchiveName ) )
    {
        OnUsage();
        return -1;
    }

    char sDirectoryName[256];
    size_t nSizeOfDirectoryName = sizeof(sDirectoryName);
    char sArchiveName[256];
    size_t nSizeOfArchiveName = sizeof(sArchiveName);

    struct utsname unameData;
    uname(&unameData);
    time_t theTime = time(NULL);
    struct tm* aTime = localtime(&theTime);

    const char* sKern = unameData.release;
    unsigned uAkern, uBkern, uCkern;
    sscanf(sKern, "%u%*c%u%*c%u", &uAkern, &uBkern, &uCkern);

    snprintf(sDirectoryName, nSizeOfDirectoryName, "%s.%u.%u.%u.%i%2.2i%2.2i", unameData.machine, uAkern, uBkern, uCkern, aTime->tm_year + 1900, aTime->tm_mon + 1, aTime->tm_mday);
    snprintf(sArchiveName, nSizeOfArchiveName, "%s.tar", sDirectoryName);

    if (NULL != sEnteredArchiveName)
    {
        if (strlen(sEnteredArchiveName) >= 4)
        {
            strncpy(sArchiveName, sEnteredArchiveName, nSizeOfArchiveName - 1);
            sArchiveName[nSizeOfArchiveName - 1] = '\0';
            char sSuf[5] = {'\0', '\0', '\0', '\0', '\0'};
            strcpy(sSuf, sEnteredArchiveName + strlen(sEnteredArchiveName) - 4);
            if (0 != strcmp(".tar", sSuf))
                strncat(sArchiveName, ".tar", nSizeOfArchiveName - strlen(sArchiveName) - 1);
        }
        else
            snprintf(sArchiveName, nSizeOfArchiveName, "%s.tar", sEnteredArchiveName);

        char sEnteredDirectoryName[256];
        const char* sFoundSlash = strrchr(sArchiveName, '/');
        if (NULL == sFoundSlash)
            strcpy(sEnteredDirectoryName, sArchiveName);
        else
            strcpy(sEnteredDirectoryName, sFoundSlash + 1);

        sEnteredDirectoryName[strlen(sEnteredDirectoryName) - 4] = '\0';
        if ((0 != strlen(sEnteredDirectoryName)) && (0 != strcmp(sEnteredDirectoryName, ".")) && (0 != strcmp(sEnteredDirectoryName, "..")))
            strcpy(sDirectoryName, sEnteredDirectoryName);
    }

    FileDumper fileDmp(sArchiveName);

    if (!fOnlyInfo)
    {
        const char* szErr = CheckUFSDEndian();
        if ( NULL != szErr )
        {
            fprintf(stderr, "%s\n", szErr);
            return -1;
        }

        if ( NULL == szDevice )
        {
            fprintf(stderr, "Device is not supplied\n");
            return -2;
        }

        hint = "l"; // linux

        //
        // Allocate BlockDevice.
        // First try to open for 'rw' to sync it if dump
        //
        Status = UFSD_IOHandlerCreate( szDevice, false, &Rw, &Device, true );
        if (ERR_NOERROR != Status)
            return -1;

        //
        // For 'dump' case we can use 'ro'
        //
        if ( UFSD_SUCCESS( Status ) )
        {
            fsync( Device );
            UFSD_IOHandlerFree( Rw );
            Rw = NULL;
        }
        //
        // Open in 'ro'
        //
        Status = UFSD_IOHandlerCreate( szDevice, true, &Rw, &Device );

        if(!fileDmp.openArchive(O_CREAT | O_WRONLY | O_TRUNC, S_IRUSR | S_IWUSR | S_IRGRP| S_IROTH))
        {
            perror("Open archive");
            return -1;
        }

        md5_byte_t digest[16];
        char sFsdump[256];
        snprintf(sFsdump, sizeof(sFsdump), "%s/dumpbin", sDirectoryName);

        if ( ERR_NOERROR == Status )
        {
            assert( NULL != Rw );
            UFSD::CMemoryManager* Mm  = UFSD_GetMemoryManager();
            UFSD::IMessage*       Msg = UFSD_GetMessageService( stdout, NULL, 0 );
            unsigned int BytesPerSector = Rw->GetSectorSize();
            assert( BytesPerSector == Rw->GetBlockSize() );
            Rw->SetBlockSize( BytesPerSector );
            UINT64 SectorsPerRw       = Rw->GetNumberOfBlocks();
            UINT64 BytesPerVolume     = SectorsPerRw * BytesPerSector;
            const char *TailHint = "";

            //
            // Init libufsd.so.
            // Without such initialization library is not able to show messages and errors
            //
            InitUfsdLibrary();

            //
            // Used to print the size of volume
            //
            GbDev = UFSD::FormatSizeGb( SectorsPerRw * BytesPerSector, &MbDev );

            char sFsdumpGz[256];
            snprintf(sFsdumpGz, sizeof(sFsdumpGz), "%s.gz", sFsdump);
            fileDmp.addTarfileWithZeroHeader(sFsdumpGz);
            fileDmp.closeArchive();

            gzFile gzOut = gzopen(sArchiveName, "ab6");
            if (gzOut == NULL)
            {
                fprintf(stderr, "Cannot open output file\n");
                return -1;
            }

            md5_state_t state;
            md5_init(&state);

            Buffer = (char*)malloc( BYTES_PER_BUFFER );
            if ( NULL == Buffer )
            {
                Status = ERR_NOMEMORY;
                goto Exit;
            }

            {
                unsigned int nErrors = 0;
                Status = UFSD::GetVolumeMetaBitmap( MINUS_ONE_T, Mm, Rw, Msg, &Bitmap, &ClustersPerVolume, &Log2OfCluster, &nErrors );
                if ( !UFSD_SUCCESS( Status ) )
                    goto Exit;

                ClustersPerBuffer = BYTES_PER_BUFFER >> Log2OfCluster;
                BytesPerBitmap    = (ClustersPerVolume+7)/8;
                Gb = UFSD::FormatSizeGb( (UINT64)ClustersPerVolume << Log2OfCluster, &Mb );

                Msg->NewProgress( 0, GetNumberOfSetBits( Bitmap, ClustersPerVolume ),
                                  _("Dumping \"%s\" (%u.%02u Gb) ..."), szDevice, Gb, Mb );

                //
                // Create dump file (1 == stdout)
                //

                //
                // Print main header + OS info + volume name
                //
                snprintf( Buffer, 64 + 1, "V: %" PLL "x, %u.%02u Gb, C: %x x %x, %s: %s",
                          SectorsPerRw, GbDev, MbDev, ClustersPerVolume, Log2OfCluster, hint, szDevice );

                if (gzwrite(gzOut, Buffer, 64) < 0)
                {
                    perror("failed to write header");
                    Status = ERR_WRITEFILE;
                    goto Exit;
                }
                else
                    md5_append(&state, (const md5_byte_t*)Buffer, 64);

                //
                // Now we are ready to dump filesystem
                //

                if ( gzwrite(gzOut, Bitmap, BytesPerBitmap) < 0 )
                {
                    perror("failed to write bitmap to dump file");
                    Status = ERR_WRITEFILE;
                    goto Exit;
                }
                else
                    md5_append(&state, (const md5_byte_t*)Bitmap, BytesPerBitmap);

                //
                // Align on 4k boundary
                //
                UINT64 Off = (64 + BytesPerBitmap + 4095) & ~4095;
                unsigned int tail = LODWORD(Off) - (64 + BytesPerBitmap);

                if (0 != tail && (memset(Buffer, 0, tail), gzwrite(gzOut, Buffer, tail)) < 0)
                {
                    perror("failed to write aligned tail to dump file");
                    Status = ERR_WRITEFILE;
                    goto Exit;
                }
                else
                    md5_append(&state, (const md5_byte_t*)Buffer, tail);

                unsigned int i, Lcn = 0, Len = 0;
                size_t Bytes = 0;
                for (i = 0; i < ClustersPerVolume; i++)
                {
                    if (GetBit(Bitmap, i))
                    {
                        if (0 == Len++)
                            Lcn = i;

                        if (Len < ClustersPerBuffer)
                            continue;
                    }
                    else if (0 == Len)
                        continue;

DoDump:
                    Msg->StepIt(Len);
                    Bytes = Len << Log2OfCluster;

                    //
                    // Read device
                    //
                    if (pread64(Device, Buffer, Bytes, (UINT64)Lcn << Log2OfCluster) <= 0)
                    {
                        perror("failed to read device");
                        Status = ERR_READFILE;
                        goto Exit;
                    }

                    //
                    // Write to dump file
                    //
                    if (gzwrite(gzOut, Buffer, Bytes) <= 0)
                    {
                        perror("failed to write dump file");
                        Status = ERR_WRITEFILE;
                        goto Exit;
                    }
                    else
                        md5_append(&state, (const md5_byte_t*)Buffer, Bytes);

                    Off += Bytes;
                    Len = 0;
                }

                if (0 != Len)
                {
                    // Dump last fragment
                    i -= 1;
                    goto DoDump;
                }

                if ( ((UINT64)ClustersPerVolume << Log2OfCluster) < BytesPerVolume )
                {
                  //
                  // Try to restore tail (if possible)
                  //
                  UINT64 LastRead = ((UINT64)Lcn << Log2OfCluster) + Bytes;

                  //
                  // Backup last 1024 or 512 bytes
                  //
                  if ( LastRead < BytesPerVolume && LastRead + 512 <= BytesPerVolume )
                  {
                    int rTail = pread64( Device, Buffer, 1024, BytesPerVolume - 1024 );
                    if ( rTail >= 512 )
                    {
                      if (gzwrite(gzOut, Buffer, rTail) <= 0)
                      {
                          perror("failed to write dump file");
                          Status = ERR_WRITEFILE;
                          goto Exit;
                      }
                      else
                          md5_append(&state, (const md5_byte_t*)Buffer, rTail);

                      TailHint = " (including tail)";
                      Off += rTail;
                    }
                  }
                }

                printf("Dump finished. File size %" PLL "u bytes%s\n", Off, TailHint);
                Status = ERR_NOERROR;
            }
Exit:
            Mm->Free(Bitmap);
            gzclose(gzOut);
            md5_finish(&state, digest);

            fileDmp.openArchive(O_RDWR);
            fileDmp.closeCurrentTarfile();
        }

        //
        // Free all resources
        //
        UFSD_IOHandlerFree(Rw);

        free(Buffer);

        //
        // Translate error code to string
        //
        PrintUfsdStatusIfError(Status);

        char hexOutput[16*2];

        for (int i = 0; i < 16; ++i)
            sprintf(hexOutput + i * 2, "%02x", digest[i]);

        char sFsdumpMd5[256];
        snprintf(sFsdumpMd5, sizeof(sFsdumpMd5), "%s.md5", sFsdump);
        fileDmp.addTarfileWithZeroHeader(sFsdumpMd5);
        fileDmp.writeMd5(hexOutput, 32);

        fileDmp.closeCurrentTarfile();
    }
    else
    {
        if(!fileDmp.openArchive(O_CREAT | O_RDWR | O_TRUNC, S_IRUSR | S_IWUSR | S_IRGRP| S_IROTH))
        {
            perror("Open archive");
            return -1;
        }
    }

    if (fileDmp.findFileInDirectory("build.prop", "/system"))
        fileDmp.setAndroidOs(true);

    printf("Collecting system information...\n");

    char sSysdump[256];
    snprintf(sSysdump, sizeof(sSysdump), "%s/sysinfo", sDirectoryName);
    fileDmp.addTarfileWithZeroHeader(sSysdump);

    if (!fileDmp.appendDiskSpace())
        fileDmp.appendCommandOutput("df");

    fileDmp.appendCommandOutput("lsblk");
    fileDmp.appendCommandOutput("blkid");

    if (!fileDmp.isAndroid())
        fileDmp.appendCommandOutput("fdisk -l");
    else
        fileDmp.appendFdiskAndroid();

    fileDmp.appendCommandOutput("parted -l -s");

    fileDmp.appendKernelInfo();

    fileDmp.appendFileContent("/proc/version");

    fileDmp.appendFileContent("/etc/hosts");
    fileDmp.appendFileContent("/proc/meminfo");

    fileDmp.appendDirectoryContent("/proc/fs/ufsd/");
    fileDmp.appendFileContent("/proc/fs/ufsd/log");
    fileDmp.appendFileContent("/proc/fs/ufsd/trace");
    fileDmp.appendFileContent("/proc/fs/ufsd/version");
    fileDmp.appendLocationAndVersionOfBinary();
    fileDmp.appendFileContent("/proc/cpuinfo");
    fileDmp.appendCommandOutput("lsmod");

    if (fileDmp.isAndroid())
        fileDmp.appendCommandOutput("getprop");

    fileDmp.closeCurrentTarfile();

    char sModules[256];
    snprintf(sModules, sizeof(sModules), "%s/modules", sDirectoryName);
    fileDmp.addTarfileWithZeroHeader(sModules);

    fileDmp.appendModulesList();

    fileDmp.closeCurrentTarfile();

    char sDmesg[256];
    snprintf(sDmesg, sizeof(sDmesg), "%s/dmesg", sDirectoryName);
    fileDmp.addTarfileWithZeroHeader(sDmesg);

    fileDmp.appendCommandOutput("dmesg");

    fileDmp.closeCurrentTarfile();

    const char* sDir = "/proc";
    const char* files[] = {"config", "config.gz"};
    for (unsigned i = 0; i < 2; ++i)
        if (fileDmp.findFileInDirectory(files[i], sDir))
        {
            char sWrFile[256];
            snprintf(sWrFile, sizeof(sWrFile), "%s/%s", sDirectoryName, files[i]);
            char sRdFile[256];
            snprintf(sRdFile, sizeof(sRdFile), "%s/%s", sDir, files[i]);
            fileDmp.addTarfileWithZeroHeader(sWrFile);
            fileDmp.copyFile(sRdFile);

            fileDmp.closeCurrentTarfile();
        }

    if (fileDmp.isAndroid())
    {
        char sLogcat[256];
        snprintf(sLogcat, sizeof(sLogcat), "%s/logcat", sDirectoryName);
        fileDmp.addTarfileWithZeroHeader(sLogcat);

        fileDmp.appendCommandOutput("logcat -d -v time");

        fileDmp.closeCurrentTarfile();
    }

    fileDmp.finishTarball();
    fileDmp.closeArchive();

    printf("System information was saved to ");
    if (!fileDmp.isAndroid())
    {
        char* sName = realpath(sArchiveName, NULL);
        if (NULL == sName)
        {
            printf("%s\n", sArchiveName);
            perror("Realpath failed");
        }
        else
        {
            printf("%s\n", sName);
            free(sName);
        }
    }
    else
    {
        char actualpath[PATH_MAX + 1];
        if (NULL == realpath(sArchiveName, actualpath))
        {
            printf("%s\n", sArchiveName);
            perror("Realpath failed");
        }
        else
            printf("%s\n", actualpath);
    }

    printf("\nPlease send this file to your support contact at Paragon Software via e-mail or " \
           "via 'My account' page at Paragon Software Support Portal or upload " \
           "it to your folder on Paragon Software FTP server at https://ftp.paragon-software.com\n");

    return fOnlyInfo ? 0 : Status;
}
