////////////////////////////////////////////////////////////////
// October of 2004. (C) Paragon Software
// If this code works, it was written by Alexander Mamaev
// If not, I don't know who wrote it.
//
// This file implements CMemoryManager for UFSD library
//
////////////////////////////////////////////////////////////////

#include <string.h>
#include <stdlib.h>
#include <stdio.h>

#if defined(_WIN32)
  #include <windows.h>
#else
  #include <unistd.h>
  #if defined __APPLE__
    #include <sys/sysctl.h>
  #elif defined __QNX__
    #include <sys/syspage.h>
  #else
    #include <sys/sysinfo.h>
  #endif
#endif

#if defined _MSC_VER
  // Let next functions be always inline
  #pragma intrinsic(memcpy, memcmp, memset)
#endif

// Include UFSD specific files
#include "ufsd.h"

#if defined UFSD_TRACE

  #ifdef __GNUC__
    extern "C" void _UFSDTrace( const char* fmt, ... ) __attribute__ ((format (printf, 1, 2)));
  #else
    // This function must be placed somewhere in the client code
    extern "C" void _UFSDTrace( const char* fmt, ... );
  #endif
  #define UFSDTrace(a) _UFSDTrace a

  #define TRACE_ONLY(x) x

#else

  #define UFSDTrace(a) do{}while((void)0,0)
  #define TRACE_ONLY(x)

#endif

struct UFSD_MemoryManager : public UFSD::CMemoryManager
{
#ifdef UFSD_TRACE
    unsigned  m_MemsetCnt, m_MemcpyCnt, m_MemmoveCnt, m_MemcmpCnt,
              m_MallocCnt, m_FreeCnt;
    size_t  m_MaxRequest;

    long m_UsedMem, m_UsedMemMax;

    UFSD_MemoryManager()
      : m_MemsetCnt(0)
      , m_MemcpyCnt(0)
      , m_MemmoveCnt(0)
      , m_MemcmpCnt(0)
      , m_MallocCnt(0)
      , m_FreeCnt(0)
      , m_MaxRequest(0)
      , m_UsedMem(0)
      , m_UsedMemMax(0)
    {
    }

    ~UFSD_MemoryManager()
    {
      unsigned Mb   = (unsigned)(m_UsedMemMax/(1024*1024));
      unsigned Kb   = (unsigned)((m_UsedMemMax%(1024*1024)) / 1024);
      unsigned b    = (unsigned)(m_UsedMemMax%1024);
      unsigned leak = (unsigned)(m_UsedMem > 0? m_UsedMem : 0);

      if ( 0 != Mb )
      {
        UFSDTrace(( "MemoryManager report: Peak usage %u.%03u Mb (%u bytes), leak %u bytes\n",
                    Mb, Kb, (unsigned)m_UsedMemMax, leak ));
      }
      else
      {
        UFSDTrace(( "MemoryManager report: Peak usage %u.%03u Kb (%u bytes), leak %u bytes\n",
                    Kb, b, (unsigned)m_UsedMemMax, leak ));
      }

      UFSDTrace(( "MemoryManager report: MaxRequest %u bytes\n",(unsigned)m_MaxRequest ));
#ifndef UFSD_USE_BUILTIN_MEMCPY
      UFSDTrace(( "MemoryManager statistic:\n"
                  "\tMemset   %u\n"
                  "\tMemcpy   %u\n"
                  "\tMemcmp   %u\n"
                  "\tMemmove  %u\n"
                  "\tMalloc   %u\n"
                  "\tFree     %u\n"
                  , m_MemsetCnt, m_MemcpyCnt, m_MemcmpCnt, m_MemmoveCnt,
                  m_MallocCnt, m_FreeCnt ));
#elif !defined UFSD_USE_BUILTIN_MEMMOVE
      UFSDTrace(( "MemoryManager statistic:\n"
                  "\tMemmove  %u\n"
                  "\tMalloc   %u\n"
                  "\tFree     %u\n"
                  , m_MemmoveCnt, m_MallocCnt, m_FreeCnt ));
#else
      UFSDTrace(( "MemoryManager statistic:\n"
                  "\tMalloc   %u\n"
                  "\tFree     %u\n"
                  , m_MallocCnt, m_FreeCnt ));
#endif
    }
#endif

#ifndef UFSD_USE_BUILTIN_MEMCPY
    virtual void Memset(void * dst, unsigned char bVal, size_t count)
    {
      TRACE_ONLY( m_MemsetCnt += 1; )
      //memset( dst, bVal, count );
      CMemoryManager::Memset( dst, bVal, count );
    }
    virtual void Memcpy( void *dst, const void *src, size_t count )
    {
      TRACE_ONLY( m_MemcpyCnt += 1; )
      //memcpy( dst, src, count );
      CMemoryManager::Memcpy( dst, src, count );
    }
    virtual int Memcmp( const void *dst, const void *src, size_t count)
    {
      TRACE_ONLY( m_MemcmpCnt += 1; )
      //return memcmp( dst, src, count );
      return CMemoryManager::Memcmp( dst, src, count );
    }
#endif

#ifndef UFSD_USE_BUILTIN_MEMMOVE
    virtual void Memmove( void *dst, const void *src, size_t count )
    {
      TRACE_ONLY( m_MemmoveCnt += 1; )
      //memmove( dst, src, count );
      CMemoryManager::Memmove( dst, src, count );
    }
#endif

    // Returns total ram in bytes or 0 if not implemented
    virtual UINT64 GetTotalRam() const
    {
#if defined(_WIN32)
      MEMORYSTATUS  Mem;
      GlobalMemoryStatus( &Mem );
      return Mem.dwTotalPhys;
#elif defined(__APPLE__)
      int mib[2] = {CTL_HW, HW_MEMSIZE};
      UINT64 totalmem;
      size_t length = sizeof(UINT64);
      if ( 0 == sysctl( mib, 2, &totalmem, &length, NULL, 0 ) )
        return totalmem;
      return 0;
#elif defined(__QNX__)
      char *str = SYSPAGE_ENTRY( strings )->data;
      struct asinfo_entry *as = SYSPAGE_ENTRY( asinfo );
      UINT64 total = 0;
      unsigned num;
      for ( num = _syspage_ptr->asinfo.entry_size / sizeof( *as ); num > 0; num-- )
      {
        if ( strcmp( &str[as->name], "ram" ) == 0 )
          total += as->end - as->start + 1;
        as++;
      }
      return total;
#elif !defined(__ANDROID__)
      struct sysinfo si;
      if ( 0 == sysinfo( &si ) )
        return si.totalram;
      return 0;
#else
      #pragma message("default support included")
      return 0;
#endif
    }

    // Returns page size or 0 if not implemented
    virtual unsigned int GetPageSize() const
    {
#if defined(_WIN32)
      SYSTEM_INFO info;
      GetSystemInfo( &info );
      return info.dwPageSize;
#else
      return sysconf(_SC_PAGESIZE);
#endif
    }

    //
    // Header aligned on 16 bytes
    // Align should not be less than 8 bytes!
    // Used when UFSD_TRACE is activated
    //
    struct Hdr{
      size_t size[16/sizeof(size_t)];
    };

    virtual void * Malloc( size_t size )
    {
#ifdef UFSD_TRACE
      m_MallocCnt += 1;
      if ( size > m_MaxRequest )
        m_MaxRequest = size;
      m_UsedMem += size;
      if ( m_UsedMemMax < m_UsedMem )
        m_UsedMemMax = m_UsedMem;
      //
      // Allocate required memory + header
      //
      Hdr* ret = (Hdr*)::malloc( size + sizeof(Hdr) );
      if ( NULL == ret )
        return NULL;
      ret->size[0] = size;
      return ret + 1;
#else
      return ::malloc( size );
#endif
    }
    virtual void Free( void * p )
    {
      if ( NULL == p )
        return;
#ifdef UFSD_TRACE
      m_FreeCnt += 1;
      Hdr* ptr = (Hdr*)(p)-1;
      m_UsedMem -= ptr->size[0];
      ::free( ptr );
#else
      ::free( p );
#endif
    }
};


// The only memory manager
static UFSD_MemoryManager s_MemMngr;

///////////////////////////////////////////////////////////
// UFSD_GetMemoryManager
//
// The API exported from this module.
// It returns the pointer to the static class.
///////////////////////////////////////////////////////////
UFSD::CMemoryManager*
UFSD_GetMemoryManager()
{
  return &s_MemMngr;
}


#ifndef _WIN32
///////////////////////////////////////////////////////////
// operator 'new'
//
//
///////////////////////////////////////////////////////////
void*
operator new(size_t cb) throw() {
  return ::malloc(cb);
}


///////////////////////////////////////////////////////////
// operator 'delete'
//
//
///////////////////////////////////////////////////////////
void
operator delete(void* p) {
  ::free(p);
}
#endif


///////////////////////////////////////////////////////////
// UFSD_HeapAlloc
//
//
///////////////////////////////////////////////////////////
void*
UFSD_HeapAlloc(
    IN unsigned long Size,
    IN int    Zero
    )
{
  void* p = ::malloc( Size );
  if ( NULL != p && Zero )
    memset( p, 0, Size );
  return p;
}


///////////////////////////////////////////////////////////
// UFSD_HeapFree
//
//
///////////////////////////////////////////////////////////
void
UFSD_HeapFree(
    IN void *p
    )
{
  ::free( p );
}
