////////////////////////////////////////////////////////////////
// October of 2004. (C) Paragon Software
// If this code works, it was written by Alexander Mamaev
// If not, I don't know who wrote it.
//
// This file implements IMessage for UFSD library
//
////////////////////////////////////////////////////////////////

#include <stdio.h>
#include <stdarg.h> // va_list
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <assert.h>
#include <signal.h>
#define ASSERT  assert
#ifndef C_ASSERT
  #define C_ASSERT(e) typedef char __C_ASSERT__[(e)?1:-1]
#endif


#ifdef _WIN32
#define vsnprintf _vsnprintf
#define strdup _strdup
#endif


#if defined UFSD_TRACE

  #ifdef __GNUC__
    extern "C" void _UFSDTrace( const char* fmt, ... ) __attribute__ ((format (printf, 1, 2)));
  #else
    // This function must be placed somewhere in the client code
    extern "C" void _UFSDTrace( const char* fmt, ... );
  #endif
  #define UFSDTrace(a) _UFSDTrace a

  #define TRACE_ONLY(x) x

#else

  #define UFSDTrace(a) do{}while((void)0,0)
  #define TRACE_ONLY(x)

#endif

#include "ufsd.h"
#include "gettext.h"

#ifndef UNREFERENCED_PARAMETER
# define UNREFERENCED_PARAMETER(P)   ((void)(P))
#endif

// To compile at maximum warning level we should suppress some warnings...
// Pragma warnings are placed here
#if defined _MSC_VER
  #pragma warning( disable : 4127 ) // conditional expression is constant
#endif


#if defined(__linux__)

#include <locale.h>

#endif

#ifndef UFSDLIB_DOMAIN
# define UFSDLIB_DOMAIN "ufsdlib"
#endif
#ifndef UFSDLIB_LOCALEDIR
# define UFSDLIB_LOCALEDIR "/usr/share/locale"
#endif


static bool s_ufsd_lib_initialized;
static bool s_Break;
static bool s_EnableBreak;

#if defined(__linux__) || defined (__APPLE__) || defined (__QNX__)

///////////////////////////////////////////////////////////
// CtrlC handler
//
//
///////////////////////////////////////////////////////////
void __cdecl CtrlC( int )
{
  s_Break = true;
}


///////////////////////////////////////////////////////////
// ufsd_lib_initialize
//
// Helper function
///////////////////////////////////////////////////////////
static void
ufsd_lib_initialize()
{
  bindtextdomain(UFSDLIB_DOMAIN, UFSDLIB_LOCALEDIR);

  signal( SIGINT, &CtrlC );
  s_ufsd_lib_initialized  = true;
  s_EnableBreak           = true;
}

#else

//
// Don't include <windows.h>
//
// #include <windows.h>

#define CTRL_C_EVENT        0
#define CTRL_BREAK_EVENT    1

// Forward declaration
extern "C"
int __stdcall
SetConsoleCtrlHandler(
    int (__stdcall *)( unsigned long CtrlType ),
    int Add
    );

///////////////////////////////////////////////////////////
// CtrlCHandler
//
//
///////////////////////////////////////////////////////////
static int __stdcall
CtrlCHandler(
    IN unsigned long CtrlType
    )
{
  // only print to console if it's safe
  if ( CTRL_C_EVENT     == CtrlType
    || CTRL_BREAK_EVENT == CtrlType )
  {
    s_Break = true;
  }

  return 1;
}

///////////////////////////////////////////////////////////
// ufsd_lib_initialize
//
// Helper function
///////////////////////////////////////////////////////////
static void
ufsd_lib_initialize()
{
  SetConsoleCtrlHandler( CtrlCHandler, 1 );
  s_ufsd_lib_initialized  = true;
  s_EnableBreak           = true;
}

#endif



#define MAX_CHARS_IN_LINE   79


//=========================================================
//      CUFSD_Message class
//=========================================================
struct CUFSD_Message : public UFSD::IMessage
{
  FILE*     m_File;
  int       m_nBarCnt;
  int       m_LastPercentsPrinted;
  int       m_LastLen;
  UINT64    m_Total;
  UINT64    m_Current;
  char*     m_CurMessage;

  size_t (__stdcall *m_ExtSyncRequest)(
      IN UFSD::IMessage*  This,
      IN size_t           code,
      IN OUT void*        arg
      );

  bool      m_bSkipPercents;


  // Helper function
  int PutMessageV(
      IN UFSD::MsgType  type,
      IN const char*    fmt,
      IN va_list        arglist
      );

  CUFSD_Message( const CUFSD_Message& );
  CUFSD_Message& operator=( const CUFSD_Message& );

  CUFSD_Message()
     : m_nBarCnt(0)
     , m_LastPercentsPrinted(-1)
     , m_LastLen(-1)
     , m_Total(0)
     , m_Current(0)
     , m_CurMessage(NULL)
     , m_ExtSyncRequest(NULL)
  {
  }

  ~CUFSD_Message()
  {
    ASSERT( 0 == m_nBarCnt );
    free( m_CurMessage );
  }

  // Resets all values to initial state
  void Reset();

  //=======================================================
  //           Progress bar related functions
  //=======================================================
  // Open progress bar. Note: can be called many times
  // Returns 0 if error
  virtual int           OpenBar(
      IN const char*  szTitle, ...
      );

  // Close progress bar. Note: can be called many times
  // Returns 0 if error
  virtual int           CloseBar();

  // Starts a new progress bar with given maximum value of position
  virtual int NewProgress(
      IN size_t         Flags,      // see UMSG_FLAGS_XXX
      IN const UINT64&  MaxValue,
      IN const char*    szTitle, ...
      )  __attribute__ ((format (printf, 4, 5)));

  // Gets the upper limits of the range of the progress bar
  virtual UINT64 GetMax() { return m_Total; }

  // Sets the current position for a progress bar
  virtual void  SetPos(
      IN const UINT64& Pos
      );

  // Gets the current position of the progress bar
  virtual UINT64  GetPos() { return m_Current; }

  // Advances the current position for a progress bar control by the step increment
  virtual void  StepIt(
      IN const UINT64& Step = 1
      )
  {
    SetPos( m_Current + Step );
  }

  // Returns current percent value
  virtual size_t GetPercent();


  //=======================================================
  //           Messages functions
  //=======================================================
  // Put formatted message
  // Returns the number of put chars
  virtual int           PutMessage(
      IN UFSD::MsgType  type,
      IN const char*    fmt, ...
      );

  //=======================================================
  //  This "callback" function allows to ask user
  //=======================================================
  size_t MessageBox(
      IN size_t         flags,
      IN const char*    fmt, ...
      );

  // Call synchronous request
  virtual size_t SyncRequest(
      IN size_t     code,
      IN OUT void*  arg
      );

  //=======================================================
  // During long operations user can press "Cancel" button
  // This function returns non 0 in this case
  //=======================================================
  virtual bool          IsBreak(
      IN bool bClear
      );
};



///////////////////////////////////////////////////////////
// CUFSD_Message::Reset
//
// Resets all values to initial state
///////////////////////////////////////////////////////////
void
CUFSD_Message::Reset()
{
  m_LastPercentsPrinted = -1;
  m_Total     = 0;
  m_Current   = 0;
  m_nBarCnt   = 0;
  free( m_CurMessage );
  m_CurMessage = NULL;
}



//=======================================================
//           Progress bar related functions
//=======================================================

///////////////////////////////////////////////////////////
// CUFSD_Message::OpenBar
//
// Open progress bar. Note: can be called many times
// Returns 0 if error
///////////////////////////////////////////////////////////
int
CUFSD_Message::OpenBar(
    IN const char*  szTitle, ...
    )
{
  if( 0 == m_nBarCnt )
    Reset();

  if ( NULL != szTitle )
  {
    va_list arglist;
    va_start( arglist, szTitle );
    PutMessageV( UFSD::MsgTitle, szTitle, arglist );
    va_end( arglist );
  }

  m_nBarCnt++;
  return 1;
}

///////////////////////////////////////////////////////////
// CUFSD_Message::CloseBar
//
// Close progress bar. Note: can be called many times
// Returns 0 if error
///////////////////////////////////////////////////////////
int
CUFSD_Message::CloseBar()
{
  m_nBarCnt--;
  ASSERT( m_nBarCnt >= 0 );
  if ( 0 == m_nBarCnt )
  {
//    for ( int i = 0; i < MAX_CHARS_IN_LINE; i++ )
//      fputc( ' ', m_File );
    fputc( '\n', m_File );
    m_LastLen = 0;
    free( m_CurMessage );
    m_CurMessage  = NULL;
  }
  return 1;
}


///////////////////////////////////////////////////////////
// CUFSD_Message::NewProgress
//
// Starts a new progress bar with given maximum value of position
///////////////////////////////////////////////////////////
int
CUFSD_Message::NewProgress(
    IN size_t         Flags,      // see UMSG_FLAGS_XXX
    IN const UINT64&  MaxValue,
    IN const char*    szTitle, ...
    )
{
  UNREFERENCED_PARAMETER( Flags );
  m_Total   = MaxValue;
  m_Current = 0;

  va_list arglist;
  va_start( arglist, szTitle );
  int Ret = PutMessageV( UFSD::MsgLog, szTitle, arglist );
  va_end( arglist );
  return Ret;
}


///////////////////////////////////////////////////////////
// CUFSD_Message::SetPos
//
// Sets the current position for a progress bar
///////////////////////////////////////////////////////////
void
CUFSD_Message::SetPos(
    IN const UINT64& Pos
    )
{
  m_Current = Pos;

  if ( m_bSkipPercents )
    return;

  int this_percents = (int)GetPercent();
  if ( m_LastPercentsPrinted == this_percents  )
    return;

  // Show percent
  m_LastPercentsPrinted = this_percents;

  int len = NULL == m_CurMessage
        ? fprintf( m_File, "\r%d%%", this_percents)
        : fprintf( m_File, "\r%d%% %s", this_percents, m_CurMessage);

  for ( int i = ++len; i < m_LastLen; i++ )
    fputc( ' ', m_File );

  m_LastLen = len;

  fputc( '\r', m_File );
//  fflush( m_File );
}


///////////////////////////////////////////////////////////
// CUFSD_Message::GetPercent
//
// Returns current percent value
///////////////////////////////////////////////////////////
size_t
CUFSD_Message::GetPercent()
{
  // Do simple check for zero
  if ( 0 == m_Total )
    return 0;
  UINT64 percent = (m_Current * 100) / m_Total;
  unsigned int Ret = LODWORD( percent );
  return Ret > 100? 100 : Ret;
}


//=======================================================
//           Messages functions
//=======================================================

//////////////////////////////////////////////////////////
// CUFSD_Message::PutMessageV
//
// Helper function
// Put formatted message to teletype
// Returns the number of put chars
///////////////////////////////////////////////////////////
int
CUFSD_Message::PutMessageV(
    IN UFSD::MsgType  type,
    IN const char*    fmt,
    IN va_list        arglist
    )
{
  char Buffer[1024];
  unsigned int Ret = vsnprintf( Buffer, ARRSIZE(Buffer),
                                dgettext(UFSDLIB_DOMAIN,fmt),
                                arglist );

  if ( Ret > 0 && Ret < ARRSIZE(Buffer)
    && '\n' == Buffer[Ret-1] )
  {
    Buffer[Ret-1] = 0;
  }

  free( m_CurMessage );
  m_CurMessage = NULL;
  switch( type )
  {
  case UFSD::MsgTitle:
  case UFSD::MsgLog:
    Ret = fprintf( m_File, "\r%s", Buffer );
    UFSDTrace(( "%s\n", Buffer ));
    break;
  case UFSD::MsgCur:
    {
      Ret = 0;
      fputc( '\r', m_File );
      if ( m_LastPercentsPrinted >= 0 )
        Ret += fprintf( m_File, "%d%% ", m_LastPercentsPrinted );

      size_t Len = strlen( Buffer );
      if( Len + Ret >= MAX_CHARS_IN_LINE )
      {
        Buffer[50] = '.';
        Buffer[51] = '.';
        Buffer[52] = '.';
        // Move tail of the string (including last zero)
        memmove( Buffer + 53, Buffer + Len - (MAX_CHARS_IN_LINE - Ret  - 53),
                (MAX_CHARS_IN_LINE - Ret  - 53) + 1 );
        Buffer[MAX_CHARS_IN_LINE-Ret] = 0;
      }

      Ret += fprintf( m_File, "%s", Buffer );
      if ( Ret < MAX_CHARS_IN_LINE )
        m_CurMessage = strdup( Buffer );
    }
    break;
  }

  for ( int i = Ret; i < m_LastLen; i++ )
    fputc( ' ', m_File );

  if ( UFSD::MsgCur == type )
  {
    fputc( '\r', m_File );
    m_LastLen = Ret;
  }
  else
  {
    fputc( '\n', m_File );
    m_LastLen = 0;
  }

//  fflush( m_File );
  return Ret;
}


//////////////////////////////////////////////////////////
// CUFSD_Message::PutMessage
//
// Put formatted message to teletype
// Returns the number of put chars
///////////////////////////////////////////////////////////
int
CUFSD_Message::PutMessage(
    IN UFSD::MsgType  type,
    IN const char*    fmt,
    ...
    )
{
  va_list arglist;
  va_start( arglist, fmt );
  int Ret = PutMessageV( type, fmt, arglist );
  va_end( arglist );
  if ( UFSD::MsgTitle == type )
    m_LastPercentsPrinted = -1;
  return Ret;
}

C_ASSERT( 0 == UMSG_MB_OK );
C_ASSERT( 1 == UMSG_MB_OKCANCEL );
C_ASSERT( 2 == UMSG_MB_ABORTRETRYIGNORE );
C_ASSERT( 3 == UMSG_MB_YESNOCANCEL );
C_ASSERT( 4 == UMSG_MB_YESNO );
C_ASSERT( 5 == UMSG_MB_RETRYCANCEL );

static const struct{
  const char*     szQuest;
  unsigned char   Ans[3];
  unsigned char   Ret[3];
} s_Quests[] = {

  // UMSG_MB_OK
  { "Press any key ", {0,0,0}, {UMSG_IDOK,UMSG_IDOK,UMSG_IDOK} },

  // UMSG_MB_OKCANCEL
  { "(O)k|(C)ancel ", {'O','C',0}, {UMSG_IDOK,UMSG_IDCANCEL,0} },

  // UMSG_MB_ABORTRETRYIGNORE
  { "(A)bort|(R)etry|(I)gnore ", {'A','R','I'}, {UMSG_IDABORT,UMSG_IDRETRY,UMSG_IDIGNORE} },

  // UMSG_MB_YESNOCANCEL
  { "(Y)es|(N)o|(C)ancel ", {'Y','N','C'}, {UMSG_IDYES,UMSG_IDNO,UMSG_IDCANCEL} },

  // UMSG_MB_YESNO
  { "(Y)es|(N)o ", {'Y','N',0}, {UMSG_IDYES,UMSG_IDNO,0} },

  // UMSG_MB_RETRYCANCEL
  { "(R)etry|(C)ancel ", {'R','C',0}, {UMSG_IDRETRY,UMSG_IDCANCEL,0} },

};



///////////////////////////////////////////////////////////
// CUFSD_Message::MessageBox
//
// Show message box
// Return value 0 means that request is not implemented or is not processed
///////////////////////////////////////////////////////////
size_t
CUFSD_Message::MessageBox(
    IN size_t       flags,
    IN const char*  fmt, ...
    )
{
  int Ret = 0;
  char Buffer[1024];
  va_list arglist;
  va_start( arglist, fmt );
  Ret = vsnprintf( Buffer, ARRSIZE(Buffer),
                   dgettext(UFSDLIB_DOMAIN,fmt),
                   arglist );
  va_end( arglist );

  if ( Ret > 0 && (unsigned)Ret < ARRSIZE(Buffer)
    && '\n' == Buffer[Ret-1] )
  {
    Buffer[Ret-1] = 0;
  }

  fprintf( m_File, "\r%s\n", Buffer );
  m_LastLen = 0;

  if ( flags >= ARRSIZE(s_Quests) )
    return 0;

  while(true)
  {
    fprintf( m_File, "%s", s_Quests[flags].szQuest );
    Ret = getchar();
    if ( UMSG_MB_OK == flags )
      return UMSG_IDOK;

    Ret = toupper( Ret );

    for ( size_t i = 0; i < ARRSIZE(s_Quests[flags].Ans) && 0 != s_Quests[flags].Ans[i]; ++i )
    {
      if ( Ret == s_Quests[flags].Ans[i] )
        return s_Quests[flags].Ret[i];
    }
  }
}


///////////////////////////////////////////////////////////
// CUFSD_Message::SyncRequest
//
// Call synchronous request
// Return code 0 means that request is not implemented or not processed
///////////////////////////////////////////////////////////
size_t
CUFSD_Message::SyncRequest(
    IN size_t     code,
    IN OUT void*  arg
    )
{
  if ( NULL != m_ExtSyncRequest )
    return (*m_ExtSyncRequest)(this, code, arg);
  return 0;
}


///////////////////////////////////////////////////////////
// CUFSD_Message::IsBreak
//
// During long operations user can press "Cancel" button
// This function returns non 0 in this case
///////////////////////////////////////////////////////////
bool
CUFSD_Message::IsBreak(
    IN bool bClear
    )
{
  if ( !s_EnableBreak )
    return false;

  bool bRet = s_Break;
  if ( bClear )
    s_Break = false;
  return bRet;
}



// The only Message Manager
static CUFSD_Message  s_UfsdMessage;


///////////////////////////////////////////////////////////
// UFSD_GetMessageService
//
//
// The API exported from this module.
// It returns the pointer to the static class.
//
///////////////////////////////////////////////////////////
UFSD::IMessage*
UFSD_GetMessageService(
    IN FILE* fd,
    IN size_t (__stdcall *ExtSyncRequest)(
      IN UFSD::IMessage*  This,
      IN size_t           code,
      IN OUT void*        arg
      ),
    IN bool bSkipPercents
    )
{
  if ( !s_ufsd_lib_initialized )
    ufsd_lib_initialize();

  s_UfsdMessage.m_File            = fd;
  setbuf(fd, 0); // Disable buffering so that percentage will be seen. It is needed for streams other than stderr (which is unbuffered by default).
  s_UfsdMessage.m_ExtSyncRequest  = ExtSyncRequest;
  s_UfsdMessage.m_bSkipPercents   = bSkipPercents;

  return &s_UfsdMessage;
}


///////////////////////////////////////////////////////////
// EnableBreak
//
// Enable/Disable break
///////////////////////////////////////////////////////////
void
EnableBreak(
    IN bool bEnable
    )
{
  s_EnableBreak = bEnable;
}

