////////////////////////////////////////////////////////////////
// October of 2004. (C) Paragon Software
// If this code works, it was written by Alexander Mamaev
// If not, I don't know who wrote it.
//
// This file implements CStrings for UFSD library
//
// NOTE: (UFSD uses UTF-16 Unicode)
//
////////////////////////////////////////////////////////////////

#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <assert.h>
#include <locale.h>

#ifdef _WIN32
#ifndef WIN32_LEAN_AND_MEAN
#define WIN32_LEAN_AND_MEAN
#endif
#include <windows.h>
#endif

#ifndef C_ASSERT
#define C_ASSERT(e) typedef char __C_ASSERT__ [(e)?1:-1]
#endif


// Include UFSD specific files
#include "ufsd.h"

#if defined _MSC_VER
  // Let next functions be always inline
  #pragma intrinsic( strcmp, strcat, strcpy, strlen )
#endif


#if defined _MSC_VER
#define strcasecmp _stricmp
#define strncasecmp _strnicmp
#endif

typedef unsigned char  __u8;

#ifndef __ANDROID__
typedef char           __s8;
#else
#define __s8 char
#endif

typedef unsigned short __u16;

C_ASSERT( sizeof(__u16) == 2 );
C_ASSERT( sizeof(__u8)  == 1 );


class CUFSD_Strings: public UFSD::CStrings
{
public:
  virtual void strcpy(UFSD::STRType type, void  *strDestination, const void *strSource );
  virtual void strncpy(UFSD::STRType type, void *strDest, const void *strSource, size_t count );
  virtual unsigned toupper( unsigned c );
  virtual int strcmp(UFSD::STRType type, const void *string1, const void *string2 );
  virtual int stricmp(UFSD::STRType type, const void *string1, const void *string2 );
  virtual int strncmp(UFSD::STRType type, const void *string1, const void *string2, size_t count );
  virtual int strnicmp(UFSD::STRType type, const void *string1, const void *string2, size_t count );
  virtual void strcat(UFSD::STRType type, void  *strDestination, const void *strSource );
  virtual const void *strstr(UFSD::STRType type,  const void *string, const void *strCharSet );
  virtual const void *strchr(UFSD::STRType type,  const void *string, unsigned c );
  virtual const void *strrchr(UFSD::STRType type,  const void *string, unsigned c );
  virtual unsigned Convert(UFSD::STRType type, const void * strSrc, size_t SrcLen, void * strDest,size_t MaxOut,size_t* DstLen);
#ifdef WIN32
  virtual unsigned ConvertOEM(UFSD::STRType type, const char* strSrc, size_t src_len, void* strDest, size_t maxout, size_t* dst_len );
  virtual unsigned ConvertToOEM(UFSD::STRType type, const void* strSrc, size_t SrcLen, char *strDest, size_t MaxOut, size_t* DstLen );
#endif
  virtual size_t   strlen(UFSD::STRType type, const void * str);
  virtual unsigned MatchMask(UFSD::STRType type, const void *name, const void *mask, unsigned MetaMatch);
  virtual unsigned MatchnMask(UFSD::STRType type, const void *name, const void *mask, size_t Len, unsigned MetaMatch = 0);
};


///////////////////////////////////////////////////////////
// CUFSD_Strings::strcpy
//
// The strcpy function copies strSource, including the terminating null character,
// to the location specified by strDestination.
// The behavior of strcpy is undefined if the source and destination strings overlap.
// type - string type ASCII or UNICODE
///////////////////////////////////////////////////////////
void
CUFSD_Strings::strcpy(
    IN UFSD::STRType type,
    IN void  *strDestination,
    IN const void *strSource
    )
{
  assert(NULL != strDestination);
  assert(NULL != strSource);

  if ( UFSD::StrASCII == type )
    ::strcpy(reinterpret_cast<char*>(strDestination), reinterpret_cast<const char*>(strSource));
  else if ( UFSD::StrUNICODE == type )
  {
#ifdef _WIN32
    ::wcscpy( (wchar_t*)strDestination, (const wchar_t*)strSource );
#else
    __u16* dst       = reinterpret_cast<__u16*>(strDestination);
    const __u16* src = reinterpret_cast<const __u16*>(strSource);
    while( 0 != (*dst = *src) )
    {
      dst++;
      src++;
    }
#endif
  }
  else
  {
    assert(0);
  }
}



///////////////////////////////////////////////////////////
// CUFSD_Strings::strncpy
//
// The strncpy function copies the initial count characters of strSource
// to strDest. If count is less than or equal to the length of strSource,
// a null character is not appended automatically to the copied string.
// If count is greater than the length of strSource, the destination string
// is padded with null characters up to length count.
// The behavior of strncpy is undefined if the source and destination strings overlap.
// type - string type ASCII or UNICODE
///////////////////////////////////////////////////////////
void
CUFSD_Strings::strncpy(
    IN UFSD::STRType type,
    IN void *strDest,
    IN const void *strSource,
    IN size_t count
    )
{
  assert(NULL != strDest);
  assert(NULL != strSource);

  if ( UFSD::StrASCII == type )
    ::strncpy( reinterpret_cast<char*>(strDest), reinterpret_cast<const char*>(strSource), count);
  else if ( UFSD::StrUNICODE == type )
  {
#ifdef _WIN32
    ::wcsncpy( (wchar_t*)strDest, (const wchar_t*)strSource, count );
#else
    __u16 *dest         = reinterpret_cast<__u16*>(strDest);
    const __u16 *source = reinterpret_cast<const __u16*>(strSource);
    while( count && '\0' != (*dest++ = *source++) )
      count--;

    if( count )
      while (--count)
        *dest++ = '\0';
#endif
  }
  else
  {
    assert(0);
  }
}


///////////////////////////////////////////////////////////
// CUFSD_Strings::toupper
//
// Converts a given lowercase letter to an uppercase letter if possible and appropriate.
///////////////////////////////////////////////////////////
unsigned
CUFSD_Strings::toupper(
    IN unsigned c
    )
{
  return ::toupper(c);
}


///////////////////////////////////////////////////////////
// CUFSD_Strings::strcmp
//
// The strcmp function compares string1 and string2 lexicographically
// and returns a value indicating their relationship.
// type - string type ASCII or UNICODE
// Returns:
//   < 0  - string1 less than string2
//    0   - string1 is identical to string2
//   > 0  - string1 greater than string2
///////////////////////////////////////////////////////////
int
CUFSD_Strings::strcmp(
    IN UFSD::STRType type,
    IN const void *string1,
    IN const void *string2
    )
{
  assert(string1);
  assert(string2);

  if ( UFSD::StrASCII == type )
    return ::strcmp(reinterpret_cast<const char*>(string1), reinterpret_cast<const char*>(string2));
  else if ( UFSD::StrUNICODE == type )
  {
#ifdef _WIN32
    return ::wcscmp( (const wchar_t*)string1, (const wchar_t*)string2 );
#else
    int f,l;
    const __u16* dst = reinterpret_cast<const __u16*>(string1);
    const __u16* src = reinterpret_cast<const __u16*>(string2);

    do {
      f = *(dst++);
      l = *(src++);
    } while ( f && (f == l) );
    return f - l;
#endif
  }
  assert(0);
  return -1;
}


///////////////////////////////////////////////////////////
// CUFSD_Strings::stricmp
//
// The strcmp function lexicographically compares lowercase versions of string1 and string2
// and returns a value indicating their relationship.
// type - string type ASCII or UNICODE
// Returns:
//   < 0  - string1 less than string2
//    0   - string1 is identical to string2
//   > 0  - string1 greater than string2
///////////////////////////////////////////////////////////
int
CUFSD_Strings::stricmp(
    IN UFSD::STRType type,
    IN const void *string1,
    IN const void *string2
    )
{
  assert(string1);
  assert(string2);

  if ( UFSD::StrASCII == type )
    return ::strcasecmp(reinterpret_cast<const char*>(string1), reinterpret_cast<const char*>(string2));
  else if ( UFSD::StrUNICODE == type )
  {
#ifdef _WIN32
    return ::wcsicmp( (const wchar_t*)string1, (const wchar_t*)string2 );
#else
    int f,l;
    const __u16* dst = reinterpret_cast<const __u16*>(string1);
    const __u16* src = reinterpret_cast<const __u16*>(string2);

    do {
      f = ::toupper(*(dst++));
      l = ::toupper(*(src++));
    } while ( f && (f == l) );
    return f - l;
#endif
  }
  assert(0);
  return -1;
}


///////////////////////////////////////////////////////////
// CUFSD_Strings::strnicmp
//
// The strnicmp function lexicographically compares, at most, the first count
// characters of string1 and string2. The comparison is performed without regard to case;
// strnicmp is a case-insensitive version of strncmp.
// The comparison ends if a terminating null character is reached in either string
// before count characters are compared. If the strings are equal when a terminating
// null character is reached in either string before count characters are compared,
// the shorter string is lesser.
// Two strings containing characters located between 'Z' and 'a' in the ASCII table
// ('[', '\', ']', '^', '_', and '`') compare differently, depending on their case.
// For example, the two strings "ABCDE" and "ABCD^" compare one way if the comparison
// is lowercase ("abcde" > "abcd^") and the other way ("ABCDE" < "ABCD^") if it is uppercase.
//  type - string type ASCII or UNICODE
// Returns:
//   < 0  - string1 substring less than string2 substring
//    0   - string1 substring identical to string2 substring
//   > 0  - string1 substring greater than string2 substring
///////////////////////////////////////////////////////////
int
CUFSD_Strings::strnicmp(
    IN UFSD::STRType type,
    IN const void *string1,
    IN const void *string2,
    IN size_t count
    )
{
  assert(string1);
  assert(string2);

  if ( UFSD::StrASCII == type )
    return ::strncasecmp(reinterpret_cast<const char*>(string1), reinterpret_cast<const char*>(string2), count);
  else if ( UFSD::StrUNICODE == type )
  {
    if ( 0 == count )
      return 0;
#ifdef _WIN32
    return wcsnicmp( (const wchar_t*)string1, (const wchar_t*)string2, count );
#else

    int f,l;
    const __u16* dst = reinterpret_cast<const __u16*>(string1);
    const __u16* src = reinterpret_cast<const __u16*>(string2);
    do {
      f = ::toupper(*(dst++));
      l = ::toupper(*(src++));
    } while ( --count && f && (f == l) );
    return f - l;
#endif
  }
  assert(0);
  return -1;
}


///////////////////////////////////////////////////////////
// CUFSD_Strings::strncmp
//
// The strncmp function lexicographically compares, at most, the first count
// characters in string1 and string2 and returns a value indicating the relationship
// between the substrings
//type - string type ASCII or UNICODE
// Returns:
//   < 0  - string1 substring less than string2 substring
//    0   - string1 substring identical to string2 substring
//   > 0  - string1 substring greater than string2 substring
///////////////////////////////////////////////////////////
int
CUFSD_Strings::strncmp(
    IN UFSD::STRType type,
    IN const void *string1,
    IN const void *string2,
    IN size_t count
    )
{
  assert(string1);
  assert(string2);

  if ( UFSD::StrASCII == type )
    return ::strncmp(reinterpret_cast<const char*>(string1), reinterpret_cast<const char*>(string2), count);
  else if ( UFSD::StrUNICODE == type )
    return ::memcmp(string1, string2, count*sizeof(__u16));
  assert(0);
  return -1;
}


///////////////////////////////////////////////////////////
// CUFSD_Strings::strcat
//
// The strcat function appends strSource to strDestination and terminates the resulting
// string with a null character. The initial character of strSource overwrites
// the terminating null character of strDestination.
// The behavior of strcat is undefined if the source and destination strings overlap.
// type - string type ASCII or UNICODE
///////////////////////////////////////////////////////////
void
CUFSD_Strings::strcat(
    IN UFSD::STRType type,
    IN void  *strDestination,
    IN const void *strSource
    )
{
  assert(NULL != strDestination);
  assert(NULL != strSource);

  if ( UFSD::StrASCII == type )
    ::strcat(reinterpret_cast<char*>(strDestination), reinterpret_cast<const char*>(strSource));
  else if ( UFSD::StrUNICODE == type )
    ::memcpy( reinterpret_cast<__u16*>(strDestination) + this->strlen( UFSD::StrUNICODE, strDestination ),
              strSource, sizeof(__u16)*this->strlen( UFSD::StrUNICODE, strSource ) + sizeof(__u16) );
  else
  {
    assert(0);
  }
}


///////////////////////////////////////////////////////////
// CUFSD_Strings::strstr
//
// The strstr function returns a pointer to the first occurrence of strSearch in string.
// The search does not include terminating null characters.
// type - string type ASCII or UNICODE
// Returns:
//  NULL - if no string found
//  pointer to the substring if it is found
///////////////////////////////////////////////////////////
const void*
CUFSD_Strings::strstr(
    IN UFSD::STRType type,
    IN const void *string,
    IN const void *strCharSet
    )
{
  if ( UFSD::StrASCII == type )
    return ::strstr(reinterpret_cast<const char*>(string), reinterpret_cast<const char *>(strCharSet));
  else if ( UFSD::StrUNICODE == type )
  {
    if ( !*(__u16*)strCharSet )
      return (void*)string;

    for ( const __u16* cp = reinterpret_cast<const __u16*>(string);
          0 != *cp;
          cp++ )
    {
      const __u16* s1 = cp;
      const __u16* s2 = reinterpret_cast<const __u16*>(strCharSet);

      while( *s1 && *s2 && !(*s1-*s2) )
        s1++, s2++;

      if( !*s2 )
        return (void*)cp;
    }

    return NULL;
  }
  assert(0);
  return NULL;
}



///////////////////////////////////////////////////////////
// CUFSD_Strings::strchr
//
// The strchr function finds the first occurrence of c in string, or it returns NULL
// if c is not found.
// The null-terminating character is included in the search.
// type - string type ASCII or UNICODE
///////////////////////////////////////////////////////////
const void*
CUFSD_Strings::strchr(
    IN UFSD::STRType type,
    IN const void *string,
    IN unsigned c
    )
{
  if ( UFSD::StrASCII == type )
    return ::strchr(reinterpret_cast<const char *>(string), (int)(c));
  else if ( UFSD::StrUNICODE == type )
  {
#ifdef _WIN32
    return ::wcschr( (const wchar_t*)string, (wchar_t)c );
#else
    for ( const __u16* str = reinterpret_cast<const __u16*>(string); 0 != *str; str++ )
    {
      if( *str == c )
        return (void*)str;
    }
    return NULL;
#endif
  }
  assert(0);
  return NULL;
}



///////////////////////////////////////////////////////////
// CUFSD_Strings::strrchr
//
// The strrchr function finds the last occurrence of c in string.
// The search includes the terminating null character.
// type - string type ASCII or UNICODE
// Returns a pointer to the last occurrence of c in string, or NULL if c is not found.
///////////////////////////////////////////////////////////
const void*
CUFSD_Strings::strrchr(
    IN UFSD::STRType type,
    IN const void *string,
    IN unsigned c
    )
{
  if ( UFSD::StrASCII == type )
    return ::strrchr(reinterpret_cast<const char *>(string), (int)(c));
  else if ( UFSD::StrUNICODE == type )
  {
#ifdef _WIN32
    return ::wcsrchr( (const wchar_t*) string, (wchar_t)c );
#else
    for ( const __u16* str = reinterpret_cast<const __u16*>(string) + this->strlen( UFSD::StrUNICODE, string ) - 1;
          str >= string;
          str-- )
    {
      if( *str == c )
        return (void*)str;
    }
    return NULL;
#endif
  }
  assert(0);
  return NULL;
}


//
// See file /linux/fs/nls/nls_base.c
//

/*
 * Sample implementation from Unicode home page.
 * http://www.stonehand.com/unicode/standard/fss-utf.html
 */
struct utf8_table {
  int     cmask;
  int     cval;
  int     shift;
  long    lmask;
  long    lval;
};

static struct utf8_table utf8_table[] =
{
    {0x80,  0x00,   0*6,    0x7F,           0,         /* 1 byte sequence */},
    {0xE0,  0xC0,   1*6,    0x7FF,          0x80,      /* 2 byte sequence */},
    {0xF0,  0xE0,   2*6,    0xFFFF,         0x800,     /* 3 byte sequence */},
    {0xF8,  0xF0,   3*6,    0x1FFFFF,       0x10000,   /* 4 byte sequence */},
    {0xFC,  0xF8,   4*6,    0x3FFFFFF,      0x200000,  /* 5 byte sequence */},
    {0xFE,  0xFC,   5*6,    0x7FFFFFFF,     0x4000000, /* 6 byte sequence */},
    {0,                   /* end of table    */}
};


///////////////////////////////////////////////////////////
// utf8_mbtowc
//
// Convert a multibyte character to a corresponding wide character
// n - Number of bytes to check (in 's')
// If the object that 's' points to forms a valid multibyte
// character, utf8_mbtowc returns the length in bytes of the multibyte character.
// If the object that 's' points to is a wide-character null character (L'\0'),
// the function returns 1.
// If the object that 's' points to does not form a valid multibyte character
// within the first 'n' characters, it returns 1.
///////////////////////////////////////////////////////////
static int
utf8_mbtowc(__u16 *p, const __u8 *s, int n)
{
  int nc = 0;
  int c0 = *s;
  long l = c0;
  for ( const struct utf8_table *t = utf8_table; t->cmask; t++) {
    nc += 1;
    if ((c0 & t->cmask) == t->cval) {
      l &= t->lmask;
      if (l < t->lval)
        return -1;
      *p = (__u16)l;
      return nc;
    }
    if (n <= nc)
      return -1;
    s += 1;
    int c = (*s ^ 0x80) & 0xFF;
    if (c & 0xC0)
      return -1;
    l = (l << 6) | c;
  }
  return -1;
}


///////////////////////////////////////////////////////////
// utf8_mbstowcs
//
// Convert a multibyte string to a corresponding wide string
// m - maximum characters in 'pwcs'
// n - maximum number of chars in 's'
///////////////////////////////////////////////////////////
static int
utf8_mbstowcs(__u16 *pwcs, int m, const __s8 *s, int n)
{
  __u16 *op;
  const __u8 *ip;
  int size;

  op = pwcs;
  ip = (const __u8*)s;
  while ( n > 0 && *ip && m > 0) {
    if (*ip & 0x80) {
      size = utf8_mbtowc(op, ip, n);
      if (-1 == size) {
        /* Ignore character and move on */
        ip += 1;
        n  -= 1;
      } else {
        op += 1;
        m  -= 1;
        ip += size;
        n  -= size;
      }
    } else {
      *op++ = *ip++;
      n -= 1;
      m -= 1;
    }
  }
  return (int)(op - pwcs);
}


///////////////////////////////////////////////////////////
// utf8_wctomb
//
// Converts a wide character to the corresponding multibyte character
// If utf8_wctomb converts the wide character to a multibyte character,
// it returns the number of bytes (which is never greater than MB_CUR_MAX)
// in the wide character.
// If 'wc' is the wide-character null character (L'\0'), utf8_wctomb returns 1.
// If the conversion is not possible in the utf8 locale, utf8_wctomb returns 1.
// n - maximum number of bytes in 's'
///////////////////////////////////////////////////////////
static int
utf8_wctomb(__u8 *s, int n, __u16 wc )
{
  long l = wc;
  for (const struct utf8_table *t = utf8_table; t->cmask; t++) {
    if (l <= t->lmask) {
      int c = t->shift;
      __u8* s0 = s;
      *s++ = (__u8)(t->cval | (l >> c));
      while( c > 0 && --n > 0 ) {
        c -= 6;
        *s++ = (__u8)(0x80 | ((l >> c) & 0x3F));
      }
      return (int)(s - s0);
    }
  }
  return -1;
}


///////////////////////////////////////////////////////////
// utf8_wcstombs
//
// Converts a wide string to the corresponding multibyte string
// n - maximum number of chars in 's'
// m - maximum characters in 'pwcs'
///////////////////////////////////////////////////////////
int
utf8_wcstombs(__s8 *s, int n, const __u16 *pwcs, int m)
{
  const __u16 *ip = pwcs;
  __u8* op = (__u8*)s;

  while (m-- > 0 && *ip && n > 0) {
    if (*ip > 0x7f) {
      int size = utf8_wctomb(op, n, *ip);
      if (size == -1) {
        /* Ignore character and move on */
      } else {
//        __u16 tmp;
//        assert( size == utf8_mbtowc(&tmp, op, size) && tmp == *ip );
        op += size;
        n  -= size;
      }
    } else {
      *op++ = (__u8) *ip;
      n -= 1;
    }
    ip++;
  }
  return (int)(op - (__u8*)s);
}


//#define CHECK_CONVERT "Use this define to check string conversation UNICODE->MBCS->UNCIODE"

///////////////////////////////////////////////////////////
// CUFSD_Strings::Convert
//
//Convert string strSrc from UNICODE to ASCII and vice versa
//Result is in strDest.
//  srcLen - the length of source string in characters
//            if (long)SrcLen < 0 then zero terminated string is assumed
//  MaxOut - number of characters in output buffer
//  DstLen - the length of destination string in characters
// Parameter type defines type of result string:
// if type - StrASCII, then result is in UNICODE format.
// if type - StrUNICODE, then result is in ASCII format.
//Returns:
//  0 - if success
///////////////////////////////////////////////////////////
unsigned
CUFSD_Strings::Convert(
    IN  UFSD::STRType type,
    IN  const void*   strSrc,
    IN  size_t        SrcLen,
    OUT void*         strDest,
    IN  size_t        MaxOut,
    OUT size_t*       DstLen
    )
{
#if 0
  unsigned Ret  = 0;
  size_t dLen   = 0;

  if ( (long)SrcLen < 0 )
    SrcLen = 0x7fffffff; //INT_MAX;

  if ( SrcLen > MaxOut )
    SrcLen = MaxOut;

  if ( UFSD::StrASCII == type )
  {
    const char* aStr  = reinterpret_cast<const char*>(strSrc);
    __u16* wDst       = reinterpret_cast<__u16*>(strDest);
    while( dLen < SrcLen )
    {
      dLen += 1;
      if ( 0 == (*wDst++ = *aStr++) )
        break;
    }
  }
  else if ( UFSD::StrUNICODE == type )
  {
    const __u16* wStr   = reinterpret_cast<const __u16*>(strSrc);
    char* aDst          = reinterpret_cast<char*>(strDest);
    while( dLen < SrcLen )
    {
      dLen += 1;
      if ( 0 == (*aDst++ = (char)(*wStr++)) )
        break;
    }
  }
  else
  {
    assert(0);
    Ret = ~0u;
  }

  if ( NULL != DstLen )
    *DstLen = dLen;
  return Ret;

#else

  size_t tmp;
  if ( (long)SrcLen < 0 )
    SrcLen = 0x7fffffff; //INT_MAX;
  if ( NULL == DstLen )
    DstLen = &tmp;
  if ( UFSD::StrASCII == type )
  {
    // MBCS->UNICODE
    int WChars = utf8_mbstowcs( (__u16*)strDest, (int)MaxOut, (const char*)strSrc, (int)SrcLen );
    if ( WChars > 0 )
    {
      *DstLen = WChars;
      if ( *DstLen < MaxOut )
        reinterpret_cast<__u16*>(strDest)[*DstLen] = 0;
#if defined CHECK_CONVERT & !defined NDEBUG
      int Len    = (*DstLen*4 + 1);
      char* aStr = (char*)malloc( Len );
      if ( NULL != aStr )
      {
        // UNICODE->MBCS
        if ( 0x7fffffff == SrcLen )
          SrcLen = ::strlen( (const char*)strSrc );
        int SrcLen2 = utf8_wcstombs( aStr, Len, (const __u16*)strDest, *DstLen );
        assert( SrcLen == SrcLen2 && 0 == memcmp( aStr, strSrc, SrcLen2 ) );
        free( aStr );
      }
#endif
      return 0;
    }
  }
  else if ( UFSD::StrUNICODE == type )
  {
    // UNICODE->MBCS
    int Chars = utf8_wcstombs( (char*)strDest, (int)MaxOut, (const __u16*)strSrc, (int)SrcLen );
    if ( Chars > 0 )
    {
      *DstLen = Chars;
      if ( *DstLen < MaxOut )
        reinterpret_cast<char*>(strDest)[*DstLen] = 0;
#if defined CHECK_CONVERT & !defined NDEBUG
      int Len    = sizeof(__u16)*(1 + *DstLen);
      __u16* wStr = (__u16*)malloc( Len );
      if ( NULL != wStr )
      {
        // MBCS->UNICODE
        if ( 0x7fffffff == SrcLen )
          SrcLen = ::wcslen( (const wchar_t*)strSrc );
        int SrcLen2 = utf8_mbstowcs( wStr, Len, (const char*)strDest, (int)*DstLen );
        assert( SrcLen == SrcLen2 && 0 == memcmp( wStr, strSrc, SrcLen2*2 ) );
        free( wStr );
      }
#endif
      return 0;
    }
  }

  *DstLen = 0;
  assert(0);
  return ~0U;
#endif
}


///////////////////////////////////////////////////////////
// CUFSD_Strings::strlen
//
// Retrieve string length - the number of characters until zero
///////////////////////////////////////////////////////////
size_t
CUFSD_Strings::strlen(
    IN UFSD::STRType type,
    IN const void * str
    )
{
  assert(NULL != str);

  if ( UFSD::StrASCII == type )
    return ::strlen(reinterpret_cast<const char*>(str));
  else if ( UFSD::StrUNICODE == type )
  {
    const __u16* wStr = reinterpret_cast<const __u16*>(str);
    size_t Len = 0;
    while( 0 != *wStr++ )
      Len++;
    return Len;
  }

  assert(0);
  return 0;
}


///////////////////////////////////////////////////////////
// CUFSD_Strings::MatchMask
//
// Compare filename and mask
// type       - string type ASCII or UNICODE
// name       - zero terminated name
// mask       - zero terminated mask
// MetaMatch  - "inverted" sign. ( If MetaMatch is set then MatchMask is equivalent str(i)cmp )
// Returns 0 if name matches the mask
///////////////////////////////////////////////////////////
unsigned
CUFSD_Strings::MatchMask(
    IN UFSD::STRType type,
    IN const void *name,
    IN const void *mask,
    IN unsigned MetaMatch
    )
{
  return MatchnMask(type, name, mask, this->strlen(type, name), MetaMatch);
}


///////////////////////////////////////////////////////////
// fnmatch
//
// len  - length of name
///////////////////////////////////////////////////////////
template <class T>
int fnmatch( const T* Mask, const T* Name, size_t Len )
{
  for(;;)
  {
    if ( T('*') == *Mask )
    {
      while ( T('.') != *Mask && 0 != *Mask )
        Mask++;
      while ( T('.') != *Name && 0 != Len )
      {
        Len--;
        Name++;
      }
    }

    if ( 0 == *Mask )
      return 0 == Len? 0 : 1;

    if ( 0 == Len && T('.') == *Mask )
    {
      Mask++;
      continue;
    }

//    if ( toupper(*Mask) == toupper(*Name)
    if ( (*Mask == *Name)
      || (T('?') == *Mask && 0 != Len) )
    {
      Mask++;
      Name++;
      Len--;
    }
    else
      return 1; // not matched
  }
}



///////////////////////////////////////////////////////////
// CUFSD_Strings::MatchnMask
//
// Compare filename and mask not zero terminating!
// type       - string type ASCII or UNICODE
// name       - string of length "len"
// len        - length of name
// mask       - zero terminated mask
// MetaMatch  - "inverted" sign. ( If MetaMatch is set then MatchMask is equivalent strn(i)cmp )
// Returns 0 if name matches the mask
///////////////////////////////////////////////////////////
unsigned
CUFSD_Strings::MatchnMask(
    IN UFSD::STRType  type,
    IN const void     *name,
    IN const void     *mask,
    IN size_t         len,
    IN unsigned       MetaMatch
    )
{
  if (MetaMatch)
  {
    if ( 0 == this->strnicmp(type, name, mask, len)
      && ( (UFSD::StrASCII == type && 0 == static_cast<const char*>(mask)[len])
        || (UFSD::StrUNICODE == type && 0 == static_cast<const __u16*>(mask)[len]) ) )
    {
      return 0;
    }
    return ERR_NOFILEEXISTS;
  }

  if ( UFSD::StrASCII == type )
  {
    if ( '*' == static_cast<const char*>(mask)[0]
      && 0   == static_cast<const char*>(mask)[1] )
    {
      return 0;
    }
  }
  else if ( UFSD::StrUNICODE == type )
  {
    if ( '*' == static_cast<const __u16*>(mask)[0]
      && 0   == static_cast<const __u16*>(mask)[1] )
    {
      return 0;
    }
  }

  if ( UFSD::StrASCII == type )
    return fnmatch( (const char*)mask, (const char*)name, len );
  else if ( UFSD::StrUNICODE == type )
    return fnmatch( (const __u16*)mask, (const __u16*)name, len );

  if ( len != this->strlen(type, mask) )
    return ERR_NOFILEEXISTS;

  return MatchnMask( type, name, mask, len,1);
}


#ifdef WIN32
///////////////////////////////////////////////////////////
// CUFSD_Strings::ConvertOEM
//
//
///////////////////////////////////////////////////////////
unsigned
CUFSD_Strings::ConvertOEM(
    IN  UFSD::STRType type,
    IN  const char*   strSrc,
    IN  size_t        src_len,  // in bytes
    OUT void*         strDest,
    IN  size_t        maxout,   // in symbols
    OUT size_t*       dst_len
    )
{
  int ret;

  if ( (long)src_len < 0 )
    src_len = this->strlen( UFSD::StrASCII, strSrc );
  if ( NULL != dst_len )
    *dst_len = 0;

  if ( UFSD::StrUNICODE == type )
  {
    // Source OEM, target UNICODE
    ret = MultiByteToWideChar( CP_OEMCP, 0, (LPCSTR)strSrc, (int)src_len, (WCHAR*)strDest, (int)maxout );
    if ( ret > 0 && (size_t)ret < maxout )
      ((WCHAR*)strDest)[ret] = 0;
  }
  else
  {
    // Source OEM target ASCII
    // Do: OEM -> UNICODE -> ASCII
    ret = MultiByteToWideChar( CP_OEMCP, 0, (LPCSTR)strSrc, (int)src_len, NULL, 0 );
    if ( ret <= 0 )
      return ERR_BADPARAMS;
    WCHAR* Mem = (WCHAR*)malloc( ret * sizeof(WCHAR) );
    if ( NULL == Mem )
      return ERR_NOMEMORY;
    // OEM->UNICODE
    ret = MultiByteToWideChar( CP_OEMCP, 0, (LPCSTR)strSrc, (int)src_len,  Mem, ret );
    // UNICODE->ASCII
    ret = WideCharToMultiByte( CP_ACP, 0, Mem, ret, (LPSTR)strDest, (int)maxout, NULL, NULL );
    free( Mem );
    if ( ret > 0 && (size_t)ret < maxout )
      ((char*)strDest)[ret] = 0;
  }

  if ( ret > 0 )
  {
    if ( NULL != dst_len )
      *dst_len = ret;
    return 0;
  }

  return ERR_BADPARAMS;
}


static WCHAR (WINAPI * s_RtlUpcaseUnicodeChar)( IN WCHAR  SourceCharacter );


///////////////////////////////////////////////////////////
// CUFSD_Strings::ConvertToOEM
//
//
///////////////////////////////////////////////////////////
unsigned
CUFSD_Strings::ConvertToOEM(
    IN  UFSD::STRType type,
    IN  const void*   strSrc,
    IN  size_t        src_len, // in symbols
    OUT char*         strDest,
    IN  size_t        maxout,  // in bytes
    OUT size_t*       dst_len
    )
{
  int ret;
  WCHAR* Mem;
  if ( (long)src_len < 0 )
    src_len = this->strlen( type, strSrc );
  if ( NULL != dst_len )
    *dst_len = 0;

  if ( UFSD::StrASCII == type )
  {
    // Source ASCII, Target OEM
    // ASCII -> UNICODE -> OEM
    ret = MultiByteToWideChar( CP_ACP, 0, (LPCSTR)strSrc, (int)src_len, NULL, 0 );
    if ( ret <= 0 )
      return ret;
    Mem = (WCHAR*)malloc( ret * sizeof(WCHAR) );
    if ( NULL == Mem )
      return ERR_NOMEMORY;
    // ASCII -> UNICODE
    src_len = MultiByteToWideChar( CP_ACP, 0, (LPCSTR)strSrc, (int)src_len,  Mem, ret );
  }
  else
  {
    //
    // Copy in buffer to upcase string
    //
    Mem = (WCHAR*)malloc( src_len * sizeof(WCHAR) );
    if ( NULL == Mem )
      return ERR_NOMEMORY;
    memcpy( Mem, strSrc, src_len * sizeof(WCHAR) );
  }

  //
  // Do upcase UNICODE
  //
  if ( NULL != s_RtlUpcaseUnicodeChar )
  {
    for ( size_t i = 0; i < src_len; i++ )
      Mem[i] = s_RtlUpcaseUnicodeChar( Mem[i] );
  }

  //
  // UNICODE -> OEM
  //
  ret = WideCharToMultiByte( CP_OEMCP, 0, Mem, (int)src_len, strDest, (int)maxout, NULL, NULL );

  free( Mem );

  if ( (size_t)ret < maxout )
    strDest[ret] = 0;

  if ( NULL != dst_len )
    *dst_len = ret;
  return 0;
}
#endif


// The only string manager
static CUFSD_Strings s_StrMngr;


///////////////////////////////////////////////////////////
// UFSD_GetStringsService
//
// The API exported from this module.
// It returns the pointer to the static class.
///////////////////////////////////////////////////////////
UFSD::CStrings*
UFSD_GetStringsService()
{
#ifdef WIN32
  *(FARPROC*)&s_RtlUpcaseUnicodeChar = GetProcAddress( GetModuleHandleW(L"ntdll.dll"), "RtlUpcaseUnicodeChar" );
  setlocale( LC_ALL, "" );
#else
//  setlocale( LC_ALL, "UTF-8" );
#endif
  return &s_StrMngr;
}
