////////////////////////////////////////////////////////////////
// March of 2006. (C) Paragon Software
// If this code works, it was written by Alexander Mamaev
// If not, I don't know who wrote it.
//
// This file contains the code that:
// - parses command line
// - calls UFSD library
//
////////////////////////////////////////////////////////////////

#include <stdio.h>
#include <stdlib.h> // exit
#include <string.h> // memset
#include <assert.h>

//
// Include main ufsd header
//
#include "ufsd.h"

//
// All UFSD classes and functions are placed in UFSD namespace
//
using namespace UFSD;

#ifndef NLS_CAT_NAME
# define NLS_CAT_NAME "wipe"
#endif


//
// This file contains declarations of useful UFSD-client functions
//
#include "funcs.h"
#include "gettext.h"

#ifdef _WIN32
//---------------------------------------------------------
// To compile windows utility
//---------------------------------------------------------

#else
//---------------------------------------------------------
// To compile linux utility
//---------------------------------------------------------
#include <linux/hdreg.h>
#include <sys/ioctl.h>
#include <locale.h>

//
// When this file is compiled as a Linux utility
// then some useful functions (GetAbsPath,ParsePath) that are shared for
// many projects are placed in file misc.cpp
//
# include "misc.h"

//---------------------------------------------------------
// End of Linux specific
//---------------------------------------------------------
#endif // _WIN32

///////////////////////////////////////////////////////////
// OnAbout
//
//
///////////////////////////////////////////////////////////
static void
OnAbout(void)
{
  fprintf( stdout, _("Zeroes free space (unused clusters and tails of files/directories) on NTFS/FAT volume\n") );
}


///////////////////////////////////////////////////////////
// OnUsage
//
// print a list of the parameters to the program
///////////////////////////////////////////////////////////
static void
OnUsage()
{
  OnAbout();
  fprintf( stdout, _(
"Usage: " NLS_CAT_NAME " [options] device\n"
"  -c                Wipe unused clusters.\n"
"  -t                Wipe tails of files/directories.\n"
"  --help            Display this help\n"
"  --trace           Turn on UFSD trace\n"
"  --verbose         Explain what is being done\n"
"  --version         Show version and exit\n") );
#ifdef _WIN32
  fprintf( stdout,
_("E.g. " NLS_CAT_NAME " -c -t g:\n"
  "E.g. " NLS_CAT_NAME " -c -t \\\\?\\Volume{d7c5ce70-3220-11db-b0fb-005056c00008}\n") );
#else
  fprintf( stdout, _("E.g. " NLS_CAT_NAME " -c -t /dev/hdb1\n") );
#endif
}


//
// Operation options
// It will be filled from user' parameter string.
//
struct wipe_options {
  unsigned
     clusters:1,
     tail:1,
     quiet:1,
     nopercents:1,
     verbose:1;
};

///////////////////////////////////////////////////////////
// parse_options
//
// Full trusted function
// Returns 0 if error
///////////////////////////////////////////////////////////
static bool
parse_options(
    IN int argc,
    IN const char *argv[],
    OUT struct wipe_options* opts,
    OUT const char** szDevice
    )
{
  int i;
  const char* cmd = NULL;

  *szDevice = NULL;

  if ( argc <= 1 )
    return false;

  for ( i = 1; i < argc; i++ )
  {
    //
    // Make sure arg is in proper format.
    //
    cmd = argv[i];

    if ( '-' != cmd[0] )
    {
      // The first "non options argument" is a device name
      if ( NULL != *szDevice )
        goto Exit;
      *szDevice = cmd;
    }

    //
    // Which arg is it?
    //
    else switch ( cmd[1] )
    {
    case 'c':
      opts->clusters = 1;
      break;

    case 't':
      opts->tail = 1;
      break;

    case '-':
      if ( 0 == strcmp( "version", cmd + 2 ) )
      {
        fprintf( stdout, VER_STRING );
        exit(1);
      }
      else if ( 0 == strcmp( "trace", cmd + 2 ) )
      {
        EnableTrace();
        break;
      }
      else if ( 0 == strcmp( "verbose", cmd + 2 ) )
      {
        opts->verbose = 1;
        break;
      }
      else if ( 0 == strcmp( "help", cmd + 2 ) )
      {
        // Force to call OnUsage
        return false;
      }
      else if ( 0 == strcmp( "nopercents", cmd + 2 ) )
      {
        opts->nopercents = 1;
        break;
      }

      //
      // !!!No break here!!!
      //

    default:
      goto Exit;  // double break
    }
  }

  return true;

Exit:
  fprintf( stderr, _("Error in argument: \"%s\"\n"), cmd );
  return false; // Force to call OnUsage
}


///////////////////////////////////////////////////////////
// main
//
//
///////////////////////////////////////////////////////////
int __cdecl
main(
    IN int argc,
    IN const char* argv[]
    )
{
  unsigned Status       = ERR_BADPARAMS;
  const char* szDevice  = NULL;
  CRWBlock* Rw          = NULL;
  wipe_options opts;
  int Device = -1;

#ifdef _WIN32
  char szVolume[64];
#else
  setlocale(LC_MESSAGES, "");
  setlocale(LC_CTYPE, "");
  bindtextdomain(NLS_CAT_NAME, LOCALEDIR);
  textdomain(NLS_CAT_NAME);
#endif

  memset( &opts, 0, sizeof(opts) );

  if ( argc < 1 )
    return -1;  // Impossible

  const char* szErr = CheckUFSDEndian();
  if ( NULL != szErr )
  {
    fprintf( stderr, "%s\n", szErr );
    return -1;
  }

  // Parse command line
  // And exit if there are errors
  if ( !parse_options( argc, argv, &opts, &szDevice ) )
  {
    OnUsage();
    return -1;
  }

  if ( NULL == szDevice )
  {
    fprintf( stderr, _("Device is not supplied\n") );
    return -2;
  }

#ifdef _WIN32

  if ( ':' == szDevice[1] )
    _snprintf( szVolume, ARRSIZE(szVolume), "\\\\.\\%s", szDevice );
  else
    _snprintf( szVolume, ARRSIZE(szVolume), szDevice );
  szVolume[ARRSIZE(szVolume)-1] = 0;
  szDevice = szVolume;

#else

  //
  // Check for mounted devices
  //
  const char* szMount = GetMount( szDevice, opts.verbose );
  if ( NULL != szMount )
  {
    fprintf( stderr, _("\nThe device \"%s\" is mounted as \"%s\"\n"
                       "Operation impossible\n\n"),
                     szDevice, szMount );
    return -3;
  }

#endif


  //
  // Init libufsd.so.
  // Without such initialization library is not able to show messages and errors
  //
  InitUfsdLibrary();

  //
  // Allocate BlockDevice
  //
  Status = UFSD_IOHandlerCreate( szDevice, false, &Rw, &Device, false, opts.verbose );

  if ( ERR_NOERROR == Status )
  {
    assert(NULL != Rw);
    CUFSD ufsd(UFSD_GetMemoryManager());

    //
    // Init UFSD
    //
    Status = ufsd.Init( UFSD_GetStringsService(), Rw, (size_t)-1, UFSD_GetTimeService() );
    if ( ERR_NOERROR == Status )
    {
      //
      // Call UFSD code
      //
      UFSD_FS_WIPE  Wipe = {0};

      Wipe.IMsg     = UFSD_GetMessageService( stdout, NULL, opts.nopercents );
      Wipe.WipeFunc = NULL;

      if ( opts.clusters )
        Wipe.Options |= UFSD_FS_WIPE_CLST;
      if ( opts.tail )
        Wipe.Options |= UFSD_FS_WIPE_TAIL;

      if ( 0 == Wipe.Options )
        fprintf( stderr, _("Neither options (\"-c\" \"-t\") are set\n") );

      Status = ufsd.FsIoControl( IOCTL_FS_WIPE, &Wipe, sizeof(Wipe) );

      if ( ERR_NOERROR == Status )
        fprintf( stdout, _("OK\n") );

    }
  }

  //
  // Free all resources
  //
  UFSD_IOHandlerFree( Rw );

  //
  // Translate error code to string
  //
  PrintUfsdStatusIfError( Status );

  // Exit to host OS
  return Status;
}
